{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff HyperBase                  }
{                                                       }
{         Streaming support                             }
{                                                       }
{         Copyright (c) 1997,99 Vladimir Gaitanoff      }
{                                                       }
{*******************************************************}

{$I HB.INC}
{$D-,L- }

unit hbStream;

interface
uses Windows, ActiveX, ComObj, hbIntf, Classes, vg2Classes, hbXML;

type
  TPacketData = (pdFldDesc, pdAttr, pdFldVal, pdFldValDelayed, pdEofNestedRows);

  TDSDataPacketFldDescList = array of TDSDataPacketFldDesc;

{ TDSWriter }

  TDSWriter = class(TComObject, IDSWriter, IDSWriter5)
  private
    FFieldDescs: TDSDataPacketFldDescList;
    FStates: array of Word;
    FParentCols, FChildNums: array of Integer;
    FColumns: Word;
    FVersion: DWord;
    FStream: TMemoryStream;
    FXMLWriter: TXMLWriter;
    FXMLMode: DWord;
    FXMLMetaDataTag, FXMLRowDataTag: Integer;
    FXMLColumn: Word;
    procedure Clear;
    procedure WriteHeader;
    procedure WritePacketData(PacketData: TPacketData);
    procedure WriteBuffer(const Buff; Count: Integer);
  protected

    function Init_Sequential(   { Initialze by sequentially adding columns }
        Version: DWord;
        Columns: Word
    ): DBResult; stdcall;

    function Init(
        Version: DWord;
        Columns: Word;
        FieldDesc: PDSDataPacketFldDesc
    ): DBResult; stdcall;

    function AddAttribute(      { Add an optional parameter }
        AttrArea: TPcktAttrArea;
        Attr: PChar;
        AttrType: DWord;
        Len: DWord;
        Value: Pointer
    ): DBResult; stdcall;

    function GetDataPacket(     { Return pointer to the finished 'safearray' }
    var SA: PSafeArray
    ): DBResult; stdcall;

    function GetSize(           { Get the size of the safearray. }
    var DataPacketSize: DWord
    ): DBResult; stdcall;

    function PutField(          { Add this field to the data stream. }
        FldStatus: TPcktFldStatus;
        Len: DWord;
        Src: Pointer
    ): DBResult; stdcall;

    function AddColumnDesc(     { Add a column descriptor. }
      const FldDes: TDSDataPacketFldDesc
    ): DBResult; stdcall;

    { Reset all data (in order to create partial data).  Datapackets will not
      contain meta-info. Last created datapacket must be released by caller. }
    function Reset: DBResult; stdcall;

    { Return number of fielddescriptors, including embedded tables etc. }

    function GetColumnCount(var Count: DWord): DBResult; stdcall;

    function GetColumnDescs(    { Return all fielddescriptors }
        FieldDes: PDSDataPacketFldDesc
    ): DBResult; stdcall;

    function GetErrorString(        { Retrieve error string }
        iErrCode  : DBResult;
        pString   : PChar
    ): DBResult; stdcall;

   { Signals last row for a nested table, in case rowcount was not known
     in advance }
   function EndOfNestedRows: DBResult; stdcall;

    { IDSWriter5 }

    function SetXMLMode(iMode: Integer): DBResult; stdcall;

    function GetXMLMode: DWord; stdcall;

  public
    procedure Initialize; override;
    destructor Destroy; override;
  end;

  IDSReader = interface (IUnknown)
    ['{9E8D2F96-591C-11D0-BF52-0020AF32BD65}']
    function ReadDataPacket(Writer: IDSWriter; SA: PSafeArray): DBResult; stdcall;
  end;

{ TDSReader }
  TDSReader = class(TComObject, IDSReader)
  private
    procedure ReadBinaryPacket(Stream: TStream; Writer: IDSWriter);
    procedure ReadXMLPacket(Stream: TStream; Writer: IDSWriter);
  protected
    function ReadDataPacket(Writer: IDSWriter; SA: PSafeArray): DBResult; stdcall;
  end;

const
  SDataPacket     = 'DATAPACKET';
  SMetaData       = 'METADATA';
  SFields         = 'FIELDS';
  SField          = 'FIELD';
  SParams         = 'PARAMS';
  SRowData        = 'ROWDATA';
  SRow            = 'ROW';
  SRowState       = 'RowState';
  SVersion        = 'Version';
  SAttrName       = 'attrname';
  SFieldType      = 'fieldtype';
  SUniExt         = '.uni';

  STrue           = 'TRUE';
  SFalse          = 'FALSE';

implementation
uses Consts, SysUtils, vg2SysUtils, vg2BCDUtils, hbErrors, hbUtils, hbTypes;

const
  pktSign         = 'HyperBase data stream'#13#10#0;
  XMLSign         = '<?xml version="1.0" standalone="yes"?>';

  XMLFieldTypes: array [fldZSTRING..fldTABLE] of string = (
    'string', 'date', 'bin.hex', 'boolean', 'i2', 'i4', 'r8',
    'fixed', 'bin.hex', 'time', 'dateTime', 'ui2', 'ui4', '',
    'bin.hex', '', '', '', '', '', '', '' { Ref }, 'nested');

function GetFieldTypeStr(iFieldType: Integer): string;
var
  FldType, Len: Integer;
begin
  FldType := FieldTypeMap[(iFieldType and dsTypeBitsMask) shr dsSizeBitsLen];
  case FldType of
    fldUNKNOWN:
      HBError(DBERR_NOTSUPPORTED);
    fldINT32:
      begin
        Len := (iFieldType and dsSizeBitsMask);
        if Len = SizeOf(SmallInt) then
          FldType := fldINT16 else
        if Len = SizeOf(Int64) then
          HBError(DBERR_NOXMLDATATYPE);
      end;
    fldUINT32:
      begin
        Len := (iFieldType and dsSizeBitsMask);
        if Len = SizeOf(SmallInt) then
          FldType := fldUINT16 else
        if Len = SizeOf(Int64) then
          HBError(DBERR_NOXMLDATATYPE);
      end;
    fldFLOAT:
      ;
    fldBCD:
      ;
    fldZSTRING, fldUNICODE:
      ;
    fldBLOB:
      ;
    fldDATE, fldBOOL, fldTIME, fldTIMESTAMP:
      ;
    fldADT, fldARRAY:
      begin
        HBError(DBERR_NOXMLDATATYPE);
      end;
    fldTABLE:
      ;
  else
    HBError(DBERR_NIY);
  end;
  if FldType <> fldUNICODE then
    Result := XMLFieldTypes[FldType] else
    Result := XMLFieldTypes[fldZSTRING] + SUniExt;
end;

function GetXMLString(ParamType, ParamLen: DWord; Source: Pointer; var Value: string): Boolean;
const
  BooleanMap: array[Boolean] of string = (SFalse, STrue);
var
  TS: TTimeStamp;
begin
  Result := True;
  case (ParamType and dsTypeBitsMask) shr dsSizeBitsLen of
    dsfldINT,
    dsfldUINT:
    begin
      case ParamLen of
        1: Value := IntToStr(Byte(Source^));
        2: Value := IntToStr(SmallInt(Source^));
        4: Value := IntToStr(Integer(Source^));
      end;
    end;
    dsfldBOOL:
      Value := BooleanMap[WordBool(Source^)];
    dsfldFLOATIEEE:
      Value := FloatToStr(Double(Source^));
    dsfldBCD:
      Value := FloatToStr(Currency(Source^));
    dsfldDATE:
      begin
        TS.Date := PDateTimeRec(Source).Date;
        TS.Time := 0;
        Value := HTMLEncodeDate(TimeStampToDateTime(TS));
      end;
    dsfldTIME:
      begin
        TS.Date := DateDelta;
        TS.Time := PDateTimeRec(Source).Time;
        Value := HTMLEncodeDate(TimeStampToDateTime(TS));
      end;
    dsfldTIMESTAMP:
      begin
        TS := MSecsToTimeStamp(TDateTimeRec(Source^).DateTime);
        Value := HTMLEncodeDate(TimeStampToDateTime(TS));
      end;
    dsfldZSTRING:
    begin
      // ???
      //SetString(Value, PChar(Source) + SizeOf(Word), ParamLen - SizeOf(Word) - 1);
      Value := PChar(Source);
    end;
    dsfldUNICODE:
    begin
      Value := PWideChar(Source);
    end;
    dsfldBYTES:
    begin
      SetString(Value, PChar(Source), ParamLen);
    end;
  else
    Result := False;
  end;
end;

function StrToBool(const Value: string): Boolean;
begin
  if SameText(Value, STrue) then
    Result := True
  else if SameText(Value, SFalse) then
    Result := False
  else
    raise EConvertError.Create(SInvalidString);
end;

{ TDSWriter }
procedure TDSWriter.Initialize;
begin
  inherited;
  FStream := TMemoryStream.Create;
  FXMLWriter := TXMLWriter.Create(FStream);
end;

destructor TDSWriter.Destroy;
begin
  Clear;
  FStream.Free;
  FXMLWriter.Free;
  inherited;
end;

procedure TDSWriter.WriteHeader;
begin
  if FXMLMode = 0 then
  begin
    FStream.WriteBuffer(pktSign, Length(pktSign));
    FStream.WriteBuffer(FVersion, SizeOf(Integer));
    FStream.WriteBuffer(FColumns, SizeOf(Word));
  end else begin
    FStream.Write(XMLSign[1], Length(XMLSign));
    FXMLWriter.BeginWriteNode(SDataPacket);
    FXMLWriter.WriteAttribute(SVersion, Format('%d.0', [PACKETVERSION_2]));
  end;
end;

procedure TDSWriter.Clear;
begin
  FFieldDescs := nil;
  FParentCols := nil;
  FChildNums  := nil;
  FStates     := nil;

  FStream.Size := 0;
  FXMLMetaDataTag := 0;
  FXMLRowDataTag := 0;
  FXMLColumn := 0;
  FXMLWriter.Clear;
end;

procedure TDSWriter.WritePacketData(PacketData: TPacketData);
begin
  FStream.WriteBuffer(PacketData, SizeOf(TPacketData));
end;

procedure TDSWriter.WriteBuffer(const Buff; Count: Integer);
begin
  FStream.WriteBuffer(Count, SizeOf(Integer));
  FStream.WriteBuffer(Buff, Count);
end;

function TDSWriter.AddColumnDesc(const FldDes: TDSDataPacketFldDesc): DBResult;
var
  FieldsStarted: Boolean;
begin
  try
    SetLength(FFieldDescs, Length(FFieldDescs) + 1);
    Move(FldDes, FFieldDescs[High(FFieldDescs)], SizeOf(TDSDataPacketFldDesc));
    if FXMLMode = 0 then
    begin
      WritePacketData(pdFldDesc);
      FStream.WriteBuffer(FldDes, SizeOf(FldDes));
    end else begin
      FieldsStarted := FXMLMetaDataTag = 0;
      if FieldsStarted then
      begin
        FXMLMetaDataTag := FXMLWriter.BeginWriteNode(SMetaData);
        FXMLWriter.BeginWriteNode(SFields);
      end;
      if FXMLWriter.StackTop <> SFields then
        FXMLWriter.EndWriteNode;
      FXMLWriter.BeginWriteNode(SField);
      FXMLWriter.WriteAttribute(SAttrName, FldDes.szFieldName);
      FXMLWriter.WriteAttribute(SFieldType, GetFieldTypeStr(FldDes.iFieldType));
      if (FldDes.iFieldType and dsTypeBitsMask) shr dsSizeBitsLen = dsfldEMBEDDEDTBL then
        FXMLWriter.BeginWriteNode(SFields);
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSWriter.GetColumnDescs(FieldDes: PDSDataPacketFldDesc): DBResult;
begin
  Move(Pointer(FFieldDescs)^, FieldDes^, Length(FFieldDescs) * SizeOf(TDSDataPacketFldDesc));
  Result := DBERR_NONE;
end;

function TDSWriter.GetColumnCount(var Count: DWord): DBResult;
begin
  Count := Length(FFieldDescs);
  Result := DBERR_NONE;
end;

function TDSWriter.Init(Version: DWord; Columns: Word;
  FieldDesc: PDSDataPacketFldDesc): DBResult;
var
  I: Integer;
begin
  Init_Sequential(Version, Columns);
  for I := 0 to Columns - 1 do
    AddColumnDesc(TDSDataPacketFldDescList(FieldDesc)[I]);
  Result := DBERR_NONE;
end;

function TDSWriter.Init_Sequential(Version: DWord; Columns: Word): DBResult;
begin
  Clear;
  FVersion := Version;
  FColumns := Columns;
  WriteHeader;
  Result := DBERR_NONE;
end;

function TDSWriter.PutField(FldStatus: TPcktFldStatus; Len: DWord; Src: Pointer): DBResult;

  procedure InitParentCols(var StartNo: Integer; ChildNo, ParentNo: Integer);
  var
    I, SaveStartNo: Integer;
    FldDes: PDSDataPacketFldDesc;
  begin
    FParentCols[StartNo] := ParentNo;
    FChildNums[StartNo] := ChildNo;
    SaveStartNo := StartNo;
    Inc(StartNo);
    FldDes := @FFieldDescs[SaveStartNo];
    if (FldDes.iFieldType and dsTypeBitsMask) shr dsSizeBitsLen = dsfldEMBEDDEDTBL then
      for I := 0 to FldDes.iFieldType and dsSizeBitsMask - 1 do
        InitParentCols(StartNo, I, SaveStartNo);
  end;

var
  Col, ParentCol, ChildNo: Integer;
  FldDes: PDSDataPacketFldDesc;
  Value: string;
begin
  Result := DBERR_NONE;
  try
    if FXMLMODE = 0 then
    begin
      if Len and dsDELAYEDBIT = 0 then
      begin
        WritePacketData(pdFldVal);
        FStream.WriteBuffer(FldStatus, SizeOf(FldStatus));
        WriteBuffer(Src^, Len);
      end else begin
        WritePacketData(pdFldValDelayed);
        FStream.WriteBuffer(FldStatus, SizeOf(FldStatus));
        WriteBuffer(Src^, SizeOf(Integer));
      end;
    end else begin
      if FXMLRowDataTag = 0 then
      begin
        if FXMLMetaDataTag > 0 then
          FXMLWriter.FlushTo(FXMLMetaDataTag);
        FXMLRowDataTag := FXMLWriter.BeginWriteNode(SRowData);
      end;

      // Initialize parent columns references
      if not Assigned(FParentCols) then
      begin
        SetLength(FParentCols, Length(FFieldDescs));
        SetLength(FChildNums, Length(FFieldDescs));
        SetLength(FStates, Length(FFieldDescs));
        Col := 0; ChildNo := 0;
        while Col <= High(FFieldDescs) do
        begin
          InitParentCols(Col, ChildNo, -1);
          Inc(ChildNo);
        end;
      end;

      ParentCol := FParentCols[FXMLColumn];
      if (ParentCol + 1) = FXMLColumn then
      begin
        if Hi(FStates[FXMLColumn]) > 0 then
        begin
          if ParentCol < 0 then
            FXMLWriter.BeginWriteNode(SRow) else
            FXMLWriter.BeginWriteNode(SRow + FFieldDescs[ParentCol].szFieldName);
          if Lo(FStates[FXMLColumn]) <> 0 then
            FXMLWriter.WriteAttribute(SRowState, IntToStr(Lo(FStates[FXMLColumn])));
        end else begin
          FStates[FXMLColumn] := PByte(Src)^ or $0100;
          Exit;
        end;
      end;

      FldDes := @FFieldDescs[FXMLColumn];

      if (FldDes.iFieldType and dsTypeBitsMask) shr dsSizeBitsLen = dsfldEMBEDDEDTBL then
      begin
        FXMLWriter.BeginWriteNode(FldDes.szFieldName);
        Inc(FXMLColumn);
        if (FldStatus <> fldIsChanged) or (Len and dsDELAYEDBIT <> 0) then
          EndOfNestedRows; 
      end else begin
        if (Len and dsDELAYEDBIT = 0) and (FldStatus = fldIsChanged) and GetXMLString(FldDes.iFieldType, Len, Src, Value) then
          FXMLWriter.WriteAttribute(FldDes.szFieldName, Value);
        FXMLColumn := (FXMLColumn + 1);
        if ParentCol = -1 then
        begin
          if FXMLColumn > High(FFieldDescs) then
          begin
            FXMLColumn := 0;
            FStates[0] := 0;
            FXMLWriter.EndWriteNode; // Close ROW tag or EMBEDDEDTABLE tag
          end;
        end else begin
          if (FXMLColumn > High(FFieldDescs)) or (FChildNums[FXMLColumn] = FFieldDescs[ParentCol].iFieldType and dsSizeBitsMask) then
          begin
            FXMLColumn := ParentCol + 1;
            FStates[FXMLColumn] := 0;
            FXMLWriter.EndWriteNode; // Close ROW tag or EMBEDDEDTABLE tag
          end;
        end;
      end;
    end;
  except
    Result := HandleExceptions;
  end;
end;

function TDSWriter.AddAttribute(AttrArea: TPcktAttrArea; Attr: PChar;
  AttrType, Len: DWord; Value: Pointer): DBResult;
const
  B: Byte = 0;
var
  AttrStr: string;
begin
  if FXMLMode = 0 then
  begin
    WritePacketData(pdAttr);
    FStream.WriteBuffer(AttrArea, SizeOf(AttrArea));
    FStream.WriteBuffer(AttrType, SizeOf(AttrType));
    if Assigned(Attr) then
      WriteBuffer(Attr^, StrLen(Attr) + 1) else
      WriteBuffer(B, 1);
    WriteBuffer(Value^, Len);
  end else begin
    if AttrArea = fldAttrArea then
    begin
      if GetXMLString(AttrType, Len, Value, AttrStr) then
        FXMLWriter.WriteAttribute(Attr, AttrStr); // Attribute of FIELD tag
    end else begin
      if FXMLWriter.Stack[FXMLWriter.Stack.Count - 2] = SFields then
      begin
        FXMLWriter.EndWriteNode; // Close FIELD tag
        FXMLWriter.EndWriteNode; // Close FIELDS tag
      end;

      if not Assigned(Attr) then
      begin
        FXMLWriter.EndWriteNode; // Close PARAMS tag
                                 // All attributes of packet area are written
      end else begin
        if FXMLWriter.StackTop <> SParams then
          FXMLWriter.BeginWriteNode(SParams);
        if GetXMLString(AttrType, Len, Value, AttrStr) then
          FXMLWriter.WriteAttribute(Attr, AttrStr);
      end;
    end;
  end;
  Result := DBERR_NONE;
end;

function TDSWriter.EndOfNestedRows: DBResult;
var
  NextCol, ParentCol: Integer;
begin
  if FXMLMode = 0 then
  begin
    WritePacketData(pdEofNestedRows)
  end else begin
    FXMLWriter.EndWriteNode; // Close DataSet tag
    ParentCol := FParentCols[FXMLColumn];
    NextCol := FindInteger(FParentCols[ParentCol], FParentCols[ParentCol + 1], High(FParentCols) - ParentCol);
    if NextCol >= 0 then
    begin
      FXMLColumn := ParentCol + 1 + NextCol;
      FStates[ParentCol + 1] := 0;
    end else begin
      FXMLColumn := 0;
      FStates[0] := 0;
      FXMLWriter.EndWriteNode; // Close ROW tag
    end;
  end;
  Result := DBERR_NONE;
end;

function TDSWriter.GetXMLMode: DWord;
begin
  Result := FXMLMode;
end;

function TDSWriter.SetXMLMode(iMode: Integer): DBResult;
begin
  FXMLMode := iMode and xmlON;
  Result := DBERR_NONE;
end;

function TDSWriter.Reset: DBResult;
begin
  Clear;
  WriteHeader;
  Result := DBERR_NONE;
end;

function TDSWriter.GetSize(var DataPacketSize: DWord): DBResult;
begin
  FXMLWriter.Flush;
  DataPacketSize := FStream.Size;
  Result := DBERR_NONE;
end;

function TDSWriter.GetDataPacket(var SA: PSafeArray): DBResult;
begin
  FXMLWriter.Flush;
  CreateDataPacket(FStream.Memory^, FStream.Size, SA);
  Result := DBERR_NONE;
end;

function TDSWriter.GetErrorString(iErrCode: DBResult; pString: PChar): DBResult;
begin
  try
    GetErrorMessage(iErrCode, pString);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

{ TDSReader }
procedure TDSReader.ReadBinaryPacket(Stream: TStream; Writer: IDSWriter);

  procedure ReadHeader;
  var
    Columns: Word;
    Version: Integer;
  begin
    Stream.ReadBuffer(Version, SizeOf(Integer));
    Stream.ReadBuffer(Columns, SizeOf(Word));
    Check(Writer.Init_Sequential(Version, Columns));
  end;

  function NextData: TPacketData;
  begin
    Stream.ReadBuffer(Result, SizeOf(TPacketData));
  end;

  procedure ReadBuffer(var P: Pointer; ASize: DWord);
  begin
    with Stream do
    begin
      if DWord(Position) + ASize > DWord(Size) then
        raise EReadError.Create(SReadError);
      P := PChar(TCustomMemoryStream(Stream).Memory) + Position;
      Stream.Seek(ASize, soFromCurrent);
    end;
  end;

  procedure ReadBufferSize(var P: Pointer; var ASize: DWord);
  begin
    Stream.ReadBuffer(ASize, SizeOf(Integer));
    ReadBuffer(P, ASize);
  end;

  procedure ReadFldDesc;
  var
    FldDes: PDSDataPacketFldDesc;
  begin
    ReadBuffer(Pointer(FldDes), SizeOf(TDSDataPacketFldDesc));
    Check(Writer.AddColumnDesc(FldDes^));
  end;

  procedure ReadAttr;
  var
    AttrArea: TPcktAttrArea;
    AttrType: DWord;
    Attr: Pointer;
    AttrLen: DWord;
    Value: Pointer;
    Len: DWord;
  begin
    Stream.ReadBuffer(AttrArea, SizeOf(AttrArea));
    Stream.ReadBuffer(AttrType, SizeOf(AttrType));
    ReadBufferSize(Attr, AttrLen);
    ReadBufferSize(Value, Len);
    Writer.AddAttribute(AttrArea, Attr, AttrType, Len, Value);
  end;

  procedure ReadFldVal(DELAYBIT: DWord);
  var
    FldStatus: TPcktFldStatus;
    Buff: Pointer;
    Size: DWord;
  begin
    Stream.ReadBuffer(FldStatus, SizeOf(FldStatus));
    ReadBufferSize(Buff, Size);
    Check(Writer.PutField(FldStatus, Size or DELAYBIT, Buff));
  end;

begin
  ReadHeader;
  while Stream.Position < Stream.Size do
    case NextData of
      pdFldDesc:
        ReadFldDesc;
      pdAttr:
        ReadAttr;
      pdFldVal:
        ReadFldVal(0);
      pdFldValDelayed:
        ReadFldVal(dsDELAYEDBIT);
      pdEofNestedRows:
        Writer.EndOfNestedRows;
    else
      HBError(DBERR_DATAPACKETMISMATCH);
    end;
end;

type
{ TXMLNode }
  TXMLNode = class
  private
    FName: string;
    FAttrs: TStrings;
  public
    constructor Create(const AName: string);
    destructor Destroy; override;
    property Attrs: TStrings read FAttrs;
    property Name: string read FName write FName;
  end;

{ TXMLField }
  TXMLField = class(TXMLNode)
  private
    FAttrInfos: TAttrInfos;
    FParent: TXMLField;
    FParams: TStrings;
    FFLDDes: DSFLDDesc;
    FPacketFldDesc: TDSDataPacketFldDesc;
    FFields: TList;
    FRows: TList;
  protected
    procedure InsertField(AField: TXMLField);
  public
    constructor Create(const AName: string; AParent: TXMLField);
    destructor Destroy; override;
    procedure ClearRows;
    procedure DecodeFldDesc;
    procedure DecodePacketFldDesc;
    procedure AddRow(Row: TStrings);
    property AttrInfos: TAttrInfos read FAttrInfos;
    property FLDDes: DSFLDDesc read FFLDDes;
    property PacketFldDesc: TDSDataPacketFldDesc read FPacketFldDesc;
    property Parent: TXMLField read FParent;
    property Params: TStrings read FParams;
    property Rows: TList read FRows;
  end;

{ TRowStrings }
  TRowStrings = class(TStringList)
  private
    FTag: Integer;
  public
    property Tag: Integer read FTag write FTag;
  end;

{ TXMLNode }
constructor TXMLNode.Create(const AName: string);
begin
  FName := AName;
  FAttrs := TStringList.Create;
end;

destructor TXMLNode.Destroy;
begin
  FAttrs.Free;
  inherited;
end;

{ TXMLField }
constructor TXMLField.Create(const AName: string; AParent: TXMLField);
begin
  inherited Create(AName);
  FParams := TStringList.Create;
  if Assigned(AParent) then AParent.InsertField(Self);
end;

destructor TXMLField.Destroy;
begin
  FParams.Free;
  ListClear(FFields);
  ListDestroyObjects(FRows, True);
  inherited;
end;

procedure TXMLField.ClearRows;
begin
  ListDestroyObjects(FRows, True);
end;

procedure TXMLField.AddRow(Row: TStrings);
begin
  ListAdd(FRows, Row);
end;

procedure TXMLField.InsertField(AField: TXMLField);
begin
  ListAdd(FFields, AField);
  AField.FParent := Self;
end;

procedure TXMLField.DecodeFldDesc;
var
  I, iFldType: Integer;
  FieldName, FieldType: string;
begin
  FieldName := FAttrs.Values[SAttrName];
  if (FieldName = '') then
    HBError(DBERR_DATAPACKETMISMATCH);
  StrCopy(FFLDDes.szName, PChar(FieldName));

  FieldType := FAttrs.Values[SFieldType];

  iFldType := fldUNKNOWN;
  if FieldType <> '' then
  begin
    if (SameText(FieldType, 'string.uni')) then
      iFldType := fldUNICODE
    else
      for I := 0 to High(XMLFieldTypes) do
        if SameText(FieldType, XMLFieldTypes[I]) then
        begin
          iFldType := I;
          Break;
        end;
  end;

  case iFldType of
    fldINT16, fldUINT16:
      FFLDDes.iUnits1 := SizeOf(SmallInt);
    fldINT32, fldUINT32:
      FFLDDes.iUnits1 := SizeOf(Integer);
    fldUNKNOWN:
      HBError(DBERR_DATAPACKETMISMATCH);
  end;

  FFLDDes.iFldType := iFldType;
end;

procedure TXMLField.DecodePacketFldDesc;
begin
  if Attrs.IndexOfName(szWIDTH) >= 0 then
    FFLDDes.iUnits1 := StrToInt(Attrs.Values[szWIDTH]);

  if Attrs.IndexOfName(szDECIMALS) >= 0 then
    FFLDDes.iUnits2 := StrToInt(Attrs.Values[szDECIMALS]);

  if Attrs.IndexOfName(szSERVERCALC) >= 0 then
    FFLDDes.bCalculated := StrToBool(Attrs.Values[szSERVERCALC]);

  if Attrs.IndexOfName(szSUBTYPE) >= 0 then
    FFldDes.iFldSubType := GetFieldSubType(FFLDDes.iFldType, Attrs.Values[szSUBTYPE]);

  if FFLDDes.iFldType = fldTABLE then
    FFLDDes.iUnits1 := ListCount(FFields) + 1;
  FldDescToPacketDS(FFLDDes, 1, FPacketFldDesc, FAttrInfos);
end;

procedure TDSReader.ReadXMLPacket(Stream: TStream; Writer: IDSWriter);

var
  Parser: TXMLParser;
  Fields: TList;
  Params: TStrings;

  function FindField(Parent: TXMLField; const FieldName: string): TXMLField;
  var
    I: Integer;
  begin
    for I := 0 to Fields.Count - 1 do
    begin
      Result := Fields[I];
      if (Result.Parent = Parent) and SameText(Result.FLDDes.szName, FieldName) then Exit;
    end;
    Result := nil;
  end;

  procedure CheckToken(Token: TXMLToken; const TokenStr: string = '');
  begin
    if (Parser.Token <> Token) or (TokenStr <> '') and not SameText(Parser.TokenStr, TokenStr) then
      HBError(DBERR_DATAPACKETMISMATCH);
  end;

  procedure CheckNextToken(Token: TXMLToken; const TokenStr: string = '');
  begin
    Parser.NextToken;
    CheckToken(Token, TokenStr);
  end;

  procedure CheckNextTag(const Tag: string);
  begin
    CheckNextToken(toTagOpen);
    CheckNextToken(toSimbol, Tag);
  end;

  procedure CheckWaitForToken(Token: TXMLToken; const TokenStr: string = '');
  begin
    Parser.WaitForToken(Token);
    CheckToken(Token, TokenStr);
  end;

  procedure CheckWaitForTag(const Tag: string);
  begin
    CheckWaitForToken(toTagOpen);
    CheckNextToken(toSimbol, Tag);
  end;

  procedure CheckWaitForTagEnd;
  begin
    CheckWaitForToken(toTagEnd);
    CheckWaitForToken(toTagClose);
  end;

  procedure ReadHeader;
  begin
    CheckWaitForTag('?xml');
    CheckWaitForToken(toTagClose);
    Check(Writer.Init_Sequential(PACKETVERSION_2, 0));
  end;

  function HasAttribute: Boolean;
  var
    SavePos: Integer;
  begin
    SavePos := Stream.Position;
    Result := (Parser.NextToken = toSimbol);
    Stream.Position := SavePos;
  end;

  function HasTag: Boolean;
  var
    SavePos: Integer;
  begin
    SavePos := Stream.Position;
    Result := (Parser.NextToken = toTagOpen) and (Parser.NextToken = toSimbol);
    Stream.Position := SavePos;
  end;

  procedure ReadAttributes(Attrs: TStrings);
  var
    AttrName, AttrVal: string;
  begin
    Attrs.Clear;
    while HasAttribute do
    begin
      CheckNextToken(toSimbol);
      AttrName := Parser.TokenStr;
      CheckNextToken(toEquality);
      CheckNextToken(toQuotedString);
      AttrVal := Parser.TokenStr;
      Attrs.Values[AttrName] := Copy(AttrVal, 2, Length(AttrVal) - 2); // Strip quotes
    end;
  end;

  procedure ReadFields(Parent: TXMLField); forward;

  procedure ReadField(Parent: TXMLField);
  var
    Field: TXMLField;
  begin
    CheckNextTag(SField);
    Field := TXMLField.Create(SField, Parent);
    try
      ReadAttributes(Field.Attrs);
      Fields.Add(Field);
    except
      Field.Free;
      raise;
    end;

    Field.DecodeFldDesc;
    if Field.FLDDes.iFldType = fldTABLE then
    begin
      CheckNextToken(toTagClose);
      ReadFields(Field);
    end;
    CheckWaitForTagEnd;
  end;

  procedure ReadFields(Parent: TXMLField);

    function HasField: Boolean;
    var
      SavePos: Integer;
    begin
      SavePos := Stream.Position;
      Result := (Parser.NextToken = toTagOpen) and (Parser.NextToken = toSimbol) and
        SameText(Parser.TokenStr, SField);
      Stream.Position := SavePos;
    end;

  var
    SavePos: Integer;
  begin
    CheckNextTag(SFields);     //<FIELDS>
    CheckNextToken(toTagClose);
    while HasField do ReadField(Parent);
    CheckNextToken(toTagOpen); //</FIELDS>
    CheckWaitForTagEnd;

    SavePos := Stream.Position;
    if (Parser.NextToken = toTagOpen) and (Parser.NextToken = toSimbol) and SameText(Parser.TokenStr, SParams) then
    begin
      if Assigned(Parent) then
        ReadAttributes(Parent.Params) else
        ReadAttributes(Params);
      CheckWaitForTagEnd;
    end else
      Stream.Position := SavePos;
  end;

  procedure WriteParams(Attrs: TStrings);

    procedure WriteAttrBool(const AttrName: string);
    var
      Value: Boolean;
    begin
      if Attrs.IndexOfName(AttrName) >= 0 then
      begin
        Value := StrToBool(Attrs.Values[AttrName]);
        Check(Writer.AddAttribute(pcktAttrArea, PChar(AttrName), (dsfldBOOL shl dsSizeBitsLen) or 1, SizeOf(Value), @Value));
      end;
    end;

    procedure WriteAttrInteger(const AttrName: string);
    var
      Value: Integer;
    begin
      if Attrs.IndexOfName(AttrName) >= 0 then
      begin
        Value := StrToInt(Attrs.Values[AttrName]);
        Check(Writer.AddAttribute(pcktAttrArea, PChar(AttrName), (dsfldINT shl dsSizeBitsLen) or 4, SizeOf(Value), @Value));
      end;
    end;

    procedure WriteAttrString(const AttrName: string);
    var
      Value: string;
    begin
      if Attrs.IndexOfName(AttrName) >= 0 then
      begin
        Value := HTMLDecode(Attrs.Values[AttrName]);
        Check(Writer.AddAttribute(pcktAttrArea, PChar(AttrName), (dsfldZSTRING shl dsSizeBitsLen), Length(Value), PChar(Value)));
      end;
    end;

  begin
    WriteAttrBool(szDATASET_DELTA);
    WriteAttrBool(szREADONLY);
    WriteAttrBool(szCONSTRAINTS);
    WriteAttrInteger(szAUTOINCVALUE);
    WriteAttrInteger(szMD_SEMANTICS);
    WriteAttrInteger(szLCID);
    WriteAttrString(szCHANGE_LOG);
  end;

  procedure WriteFields(Parent: TXMLField);
  var
    I: Integer;
    Field: TXMLField;
  begin
    for I := 0 to Fields.Count - 1 do
    begin
      Field := Fields.List^[I];
      if Field.Parent = Parent then
      begin
        Field.DecodePacketFldDesc;
        Check(Writer.AddColumnDesc(Field.PacketFldDesc));
        WriteAttributes(Writer, Field.AttrInfos, 1);
        if Field.FLDDes.iFldType = fldTABLE then
        begin
          WriteFields(Field);
          Check(Writer.AddAttribute(pcktAttrArea, nil, 0, 0, nil));
        end;
      end;
    end;
    if Assigned(Parent) then
      WriteParams(Parent.Params) else
      WriteParams(Params);
  end;

  procedure ReadMetaData;
  begin
    CheckNextTag(SMetaData);
    CheckNextToken(toTagClose);
    ReadFields(nil);
    WriteFields(nil);
    CheckWaitForTagEnd;
  end;

  function GetIntValue(Row: TStrings; const AttrName: string): Integer;
  begin
    Result := StrToInt(Row.Values[AttrName]);
  end;

  function GetStrValue(Row: TStrings; const AttrName: string): string;
  begin
    Result := HTMLDecode(Row.Values[AttrName]);
  end;

  function GetBoolValue(Row: TStrings; const AttrName: string): Boolean;
  begin
    Result := StrToBool(Row.Values[AttrName]);
  end;

  function GetDateValue(Row: TStrings; const AttrName: string): TDateTime;
  begin
    Result := HTMLDecodeDate(Row.Values[AttrName]);
  end;

  procedure WriteDataSet(Field: TXMLField; ParentRow: Integer); forward;

  procedure WriteField(Field: TXMLField; ParentRow: Integer; Row: TStrings);
  var
    AttrValue: string;
    I: Integer;
    W: WideString;
    B: WordBool;
    D: Double;
    TS: TTimeStamp;
    DateData: Double;
    C: Currency;
    BCD: FMTBCD;
  begin
    with Field, FLDDes do
    begin
      if iFldType <> fldTABLE then
      begin
        if Row.IndexOfName(szName) >= 0 then
        begin
          AttrValue := HTMLDecode(Row.Values[szName]);
          case FldDes.iFldType of
            fldZSTRING, fldBLOB:
              Check(Writer.PutField(fldIsChanged, Length(AttrValue), PChar(AttrValue)));
            fldUNICODE:
              begin
                W := AttrValue;
                Check(Writer.PutField(fldIsChanged, Length(W) * 2, PChar(W)));
              end;
            fldDATE:
              begin
                TS := DateTimeToTimeStamp(HTMLDecodeDate(AttrValue));
                Check(Writer.PutField(fldIsChanged, SizeOf(Integer), @TS.Date));
              end;
            fldTIME:
              begin
                TS := DateTimeToTimeStamp(HTMLDecodeDate(AttrValue));
                Check(Writer.PutField(fldIsChanged, SizeOf(Integer), @TS.Time));
              end;
            fldTIMESTAMP:
              begin
                TS := DateTimeToTimeStamp(HTMLDecodeDate(AttrValue));
                DateData := TimeStampToMSecs(TS);
                Check(Writer.PutField(fldIsChanged, SizeOf(Double), @DateData));
              end;
            fldBOOL:
              begin
                B := StrToBool(AttrValue);
                Check(Writer.PutField(fldIsChanged, SizeOf(WordBool), @B));
              end;
            fldINT16, fldINT32, fldUINT16, fldUINT32:
              begin
                I := StrToInt(AttrValue);
                Check(Writer.PutField(fldIsChanged, iUnits1, @I));
              end;
            fldFLOAT:
              begin
                D := StrToFloat(AttrValue);
                Check(Writer.PutField(fldIsChanged, SizeOf(Double), @D));
              end;
            fldBCD:
              begin
                C := StrToFloat(AttrValue);
                CurrToFMTBCD(C, BCD, iUnits1, iUnits2);
                Check(Writer.PutField(fldIsChanged, SizeOf(FMTBCD), @BCD));
              end;
          else
            Check(Writer.PutField(fldIsNull, 0, nil));
          end;
        end else
          Check(Writer.PutField(fldIsNull, 0, nil));
      end else
        WriteDataSet(Field, ParentRow);
    end;
  end;

  procedure WriteRow(Parent: TXMLField; ParentRow: Integer; Row: TStrings); forward;

  procedure WriteDataSet(Field: TXMLField; ParentRow: Integer);
  var
    I, Count: Integer;
    Row: TRowStrings;
    Recs: TList;
  begin
    with Field do
    begin
      Recs := TList.Create;
      try
        for I := 0 to ListCount(Rows) - 1 do
        begin
          Row := Rows.List^[I];
          if Row.Tag = ParentRow then
            Recs.Add(Row);
        end;

        Count := Recs.Count;
        if Count = 0 then Count := -1;
        Check(Writer.PutField(fldIsChanged, SizeOf(DWord), @Count));
        for I := 0 to Recs.Count - 1 do
          WriteRow(Field, I, Recs.List^[I]);
        Check(Writer.EndOfNestedRows);
      finally
        Recs.Free;
      end;
    end;
  end;

  procedure WriteRow(Parent: TXMLField; ParentRow: Integer; Row: TStrings);
  var
    I: Integer;
    Field: TXMLField;
    RowState: Integer;
  begin
    if Row.IndexOfName(SRowState) >= 0 then
      RowState := GetIntValue(Row, SRowState) else
      RowState := 0;

    Check(Writer.PutField(fldIsChanged, SizeOf(DSAttr), @RowState));
    for I := 0 to Fields.Count - 1 do
    begin
      Field := Fields.List^[I];
      if Field.Parent = Parent then
        WriteField(Field, ParentRow, Row);
    end;

    if not Assigned(Parent) then
      for I := 0 to Fields.Count - 1 do
      begin
        Field := Fields.List^[I];
        if Field.FLDDes.iFldType = fldTABLE then
          Field.ClearRows;
      end;
  end;

  procedure ReadRows(Parent: TXMLField); forward;

  procedure ReadRow(Parent: TXMLField);
  var
    Row, RowRef: TRowStrings;
    Field: TXMLField;
  begin
    CheckNextToken(toTagOpen);
    CheckNextToken(toSimbol);
    if SameText(Parser.TokenStr, SRow) or Assigned(Parent) and SameText(Parser.TokenStr, SRow + Parent.FLDDes.szName) then
    begin
      Row := TRowStrings.Create;
      if Assigned(Parent) and Assigned(Parent.Parent) then
        Row.Tag := ListCount(Parent.Parent.Rows) - 1 else
        Row.Tag := -1;
      try
        RowRef := Row;
        if Assigned(Parent) then
        begin
          Parent.AddRow(Row);
          Row := nil;
        end;

        ReadAttributes(RowRef);
        case Parser.NextToken of
          toTagEnd:
            begin
              CheckNextToken(toTagClose);
            end;
          toTagClose:
            begin
              while HasTag do
              begin
                Field := FindField(Parent, Parser.TokenStr);
                if Assigned(Field) and (Field.FLDDes.iFldType = fldTABLE) then
                begin
                  ReadRows(Field);
                end else
                  Break;
              end;
              CheckWaitForTagEnd;
            end;
        else
          HBError(DBERR_DATAPACKETMISMATCH);
        end;

        if not Assigned(Parent) then WriteRow(nil, -1, RowRef);

      finally
        Row.Free;
      end;
    end else
      HBError(DBERR_DATAPACKETMISMATCH);
  end;

  procedure ReadRows(Parent: TXMLField);
  var
    RowDataTag: string;
  begin
    if Assigned(Parent) then
      RowDataTag := Parent.FLDDes.szName else
      RowDataTag := SRowData;
    CheckNextTag(RowDataTag);
    case Parser.NextToken of
      toTagClose:
        begin
          while HasTag do ReadRow(Parent);
          CheckWaitForTagEnd;
        end;
      toTagEnd:
        CheckNextToken(toTagClose);               // No more rows available
    else
      HBError(DBERR_DATAPACKETMISMATCH);
    end;
  end;

  procedure ReadDataPacket;
  begin
    CheckNextTag(SDataPacket);
    CheckWaitForToken(toTagClose);
    ReadMetaData;
    ReadRows(nil);
    CheckWaitForTagEnd;
  end;

begin
  Parser := TXMLParser.Create(Stream);
  with Parser do
  try
    ReadHeader;
    Fields := TList.Create;
    Params := TStringList.Create;
    try
      try
        ReadDataPacket;
      except
        on EConvertError do
          HBError(DBERR_DATAPACKETMISMATCH);
      end;
    finally
      Params.Free;
      ListDestroyObjects(Fields, True);
    end;
  finally
    Parser.Free;
  end;
end;

function TDSReader.ReadDataPacket(Writer: IDSWriter; SA: PSafeArray): DBResult;

var
  Size: Integer;
  Data: Pointer;
  Stream: TReadingMemoryStream;

  function ReadSignature: Boolean;
  var
    Sign: array [0..Length(pktSign) - 1] of Char;
  begin
    ZeroMemory(@Sign, SizeOf(Sign));
    Stream.Read(Sign, Length(pktSign));
    Result := StrComp(Sign, pktSign) = 0;
  end;

begin
  try
    Stream := TReadingMemoryStream.Create;
    try
      Size := DataPacketSize(SA);
      OleCheck(SafeArrayAccessData(SA, Data));
      try
        Stream.SetPointer(Data, Size);
        if ReadSignature then
        begin
          ReadBinaryPacket(Stream, Writer);
        end else begin
          Stream.Position := 0;
          ReadXMLPacket(Stream, Writer);
        end;
      finally
        SafeArrayUnaccessData(SA);
      end;
    finally
      Stream.Free;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

end.



