{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff HyperBase                  }
{                                                       }
{         Database Core                                 }
{                                                       }
{         Copyright (c) 1997,99 Vladimir Gaitanoff      }
{                                                       }
{*******************************************************}

{$I HB.INC}
{$D-,L-}

unit hbCore;

interface

uses Windows, ActiveX, hbIntf, hbTypes, Classes, ComObj, vg2CanExpr;

type
  TDSCursor = class;

{ TDSBase }

  TDSBase = class(TComObject, IDSBase, IDSBase5, IDSWriter)
  private
    FFields: DSFLDDescList;
    FIndexes: DSIDXDescList;
    FCursor, FMDCursor, FAICursor: IDSCursor;
    FLastUsedIndex: Integer;

    FName: string;
    FRecBufSize: Integer;
    FRecords: TList;
    FCursors: TList;
    FChangeCount: Integer;
    FChanges: TList;
    FChangesDS, FChangesVL: TList;
    FBlobs: TList;

    FLogChanges: Boolean;
    FReadOnly, FStreamedReadOnly: Boolean;
    FIsDelta: Boolean;
    FNoMetaData: Boolean;
    FDSIsPartial: Boolean;
    FLCID: Integer;
    FBlobsInDelta: DWord;
    FMDOptions: DWord;
    FXMLStreamMode: DWord;

    FAutoIncDisabled: Boolean;
    FAutoIncCounter: Integer;

    FConstrDisabled: Boolean;

    FRecNoCounter: Integer;

    FNotifyDisableCount: Integer;

    { Calculated fields }
    FCiClientData: DWord;         { Client data }
    FCpfCalc     : pfDSCalcField; { Callback function, NULL to remove }

    { Attributes }
    FAttrInfos: TAttrInfos;

    { Reconcile }
    FRiClientData: DWord;
    FpfReconcile: pfDSReconcile;
    FResponses: TList;
    FRCursor: IDSCursor;

    { Constraints }
    FConstrErrorMsg: string;

    { Embedded datasets }
    FParentDS: TDSBase;
    FParentFieldNo: DWord;
    FChilds: DSBASEList;
    FLinkFieldNo: DWord;
    FLinkFieldCreated: Boolean;

    { Writer }
    FWHasMetaData, FWHasChildren: Boolean;
    FWDataSet: TDSBase;
    FWDataSetFieldNo: Integer;
    FWDataSetAttrInfos: PAttrInfos;
    FWStates: TList;
    FWRecNos: TList;
    FWStateReached: Boolean;
    FWChanges: TChangeInfos;
    FWFldDatas: TMemoryStream;

    FWFldDescs: TFldInfos;
    FWAttrInfos: TAttrInfos;

    FRecsWritten: TList;

    procedure AddFieldInternal(ParentID: Word; pFldDes: pDSFLDDesc;
      CreateEmbedded: Boolean);
    procedure GetFieldDescList(var AFields: DSFLDDescList; Recurse, Response: Boolean);

    { Streaming: reading }
    procedure WriterStart;
    procedure WriterCreateMetaData;
    procedure WriterApply;
    procedure WriterCreateRecord(Recs: TList; Index: Integer);
    procedure WriterFree;

    { Streaming: writing }
    procedure WriteColumn(const Writer: IDSWriter; var DescNo: Integer; DoWrite: Boolean);
    procedure WriteMetaData(const Writer: IDSWriter; Root: Boolean);
    procedure WriteChangeLog(const Writer: IDSWriter);

    procedure WriteDataSet(const Writer: IDSWriter; pRecBufM: PChar);
    procedure WriteField(const Writer: IDSWriter; pRecBuf: PChar; FieldNo: DWord);
    procedure WriteRecord(const Writer: IDSWriter; pRecBuf: PChar);

    procedure Clear;

    procedure CursorNeeded(var ACursor: IDSCursor);

    { Change log }
    procedure AddChange(Change: PChangeInfo);
    procedure DelChange(Change: PChangeInfo);
    function HasChangeEntry(RecNo: Integer): Boolean;
    procedure AddChangeDS(DataSet: TDSBase; Change: PChangeInfo);
    procedure DelChangeDS(DataSet: TDSBase; Change: PChangeInfo);
    procedure LogChange(RecAttr: DSAttr; OldSeqNo, NewSeqNo: Integer);
    procedure ClearChangeLog;
    function FindChangeEntry(var Start: Integer; NewSeqNo: Integer): PChangeInfo;
    procedure DestroyChangeEntry(var Change: PChangeInfo; Decrement: Boolean; DeltaRecs: TList = nil);
    procedure ClearParentChangeEntry(ParentSeqNo: Integer);
    procedure UndoChangeEntry(Change: PChangeInfo; var RecNo: Integer; Commited: Boolean);
    function InternalUndoLastChange(var RecNo: Integer; Commited: Boolean): DBResult;
    function UndoLastChange(var RecNo: Integer; Commited: Boolean): DBResult;

    procedure InsertCursor(ACursor: TDSCursor);
    procedure RemoveCursor(ACursor: TDSCursor);

    procedure DisableNotify;
    procedure EnableNotify;
    procedure NotifyCursors(Event: Integer; Data1, Data2, Data3: Integer);

    function InitRecord(pRecBuf: PChar): DBResult;
    procedure AllocRecord(var P: PChar);
    procedure FreeRecord(var P: PChar);
    procedure AssignFields(pRecDst, pRecSrc: PChar; ChangesOnly: Boolean = False; SkipLinks: Boolean = False);
    procedure CopyRecord(Source, Dest: PChar);

    procedure RemoveRecord(RecNo: DWord; pRecBuf: PChar = nil; DeltaRecs: TList = nil);
    function BlankNotChanged(pRecBuf, pRecOrg: PChar; CheckOnly: Boolean): Boolean;

    procedure CheckNotReadOnly;
    procedure CheckBlobFetched(pRecBuf: PChar; iFldNo: DWord);

    function RevertRecord(RecNo: Integer): DBResult;

    { Constraints }
    function GetMaxAutoInc(FieldNo: DWord): Integer;
    procedure CheckRequired(pRecBuf: PChar; AutoInc: Boolean);
    procedure CheckConstraints(pRecBuf, ConstrType: PChar; Defaults: Boolean);
    procedure CheckUnique(pRecBuf, pRecOrg: PChar);

    procedure CalculateFields(pRecBuf: PChar);
    procedure InternalInsertRecord(pRecBuf: PChar; var NewRecNo: Integer);
    function InsertRecord(pRecBuf: PChar; var NewRecNo: Integer; CommitTables: Boolean = True): DBResult;
    function ModifyRecord(RecNo: DWord; pRecBuf: PChar; var NewRecNo: Integer; CommitTables: Boolean = True; Detail: Boolean = True): DBResult;
    function DeleteRecord(RecNo: DWord): DBResult;

    function GetRecord(RecNo: DWord; pRecBuf: PChar): DBResult;
    procedure GetRecordOrg(RecNo: DWord; var OrgRecNo: DWord; pRecBuf: PChar = nil);
    function GetRecordPtr(RecNo: Integer): PChar;

    function GetField(pRecBuf: Pointer; iFieldNo: DWord;
      pFldBuf: Pointer; var bBlank: Bool): DBResult;

    procedure PutFieldBuff(pRecBuf: Pointer; iFieldNo: DWord; pFldBuf: Pointer);

    function PutFieldInternal(pRecBuf: Pointer; iFieldNo: DWord;
      pFldBuf: Pointer): DBResult;
    function PutField(pRecBuf: Pointer; iFieldNo: DWord;
      pFldBuf: Pointer): DBResult;

    procedure PutBlankParent(pRecBuf: PChar; iFldNo: DWord);
    procedure PutBlankValues(pRecBuf: Pointer; iFldNo, iBlankValue: DWord);
    procedure PutBlankInternal(pRecBuf: Pointer; iFldNo, iBlankValue: DWord);

    procedure CommitRecord(pRecBuf: PChar; CommitTables: Boolean = True);

    function GetBlobLen(pRecBuf: Pointer; iFieldNo: DWord;
      var iLength: DWord): DBResult;
    function GetBlob(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
      pBuf: Pointer; var iLength: DWord): DBResult;
    function PutBlobInternal(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
      pBuf: Pointer; iLength: DWord): DBResult;
    function PutBlob(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
      pBuf: Pointer; iLength: DWord): DBResult;

    procedure CheckInactiveIndex(szName: PChar);
    function FindIndex(pszName: PChar): Integer;
    procedure GetIndexDesc(IndexNo: DWord; p1: PDSIDXDesc);
    procedure UpdateLinkedIndex(var IdxDesc: DSIDXDesc);

    { Comparing records }
    function FldCmpEx(iFldType: DWord; pFld1, pFld2: Pointer; iUnits1,
      iUnits2: DWord; CaseIns: Bool): Integer;
    function CmpRecsEx(IndexDS: pDSIDXDesc; iFields, iPartLen: Integer;
      pRec1, pRec2: PChar): Integer;

    procedure CalcKeySize(var IndexDS: DSIDXDesc);
    function ExtractKey(pRecBuf, pKeyBuf: PChar; const IndexDS: DSIDXDesc): DBResult;

    { Delta & Reconcile }
    procedure GetDeltaRecords(pRecBufM: PChar; Recs: TList);
    procedure InitDelta(Delta: TDSBase; pRecBufM, pRecBufM2: PChar; BlobsInDelta: DWord);
    procedure MergeRecord(Attr: DSAttr; RecNo: DWord; pNewRec: PChar; DeltaRecs: TList);
    procedure ClearChanges(Attr: DSAttr; RecNo: DWord; pRecOrg: PChar = nil; DeltaRecs: TList = nil; RollingBack: Boolean = False);
    function GetResponses(DeltaDS, ErrorDS: TDSBase; FieldNums: TList;
      iClientData: DWord; pfReconcile_MD: pfDSReconcile_MD): Boolean;
    procedure MergeRecords(DeltaDS: TDSBase);

    function ReconcileCallback(iRslt: Integer; iUpdateKind: DSAttr;
      iResAction: dsCBRType; iErrCode: Integer; pErrMessage, pErrContext: PChar;
      pRecUpd, pRecOrg, pRecConflict: Pointer; iLevels: Integer;
      piFieldIDs: PInteger): dsCBRType; stdcall;

      { Refresh record }
    procedure AppendChildren(pRecBuf: PChar; Source: TDSBase);

    function GetRowRequestPacket(   { Get packet describing the curent 'path',
                                      for delayed fetching/refreshing }
        RecNo           : DWord;
        IndexDS         : pDSIDXDesc;
        bKeysOnly       : Bool;     { Only include unique keys in packet }
        bFetchAllBlobs  : Bool;     { fetch all blobs for 'current'record }
        bFetchAllDetails: Bool;     { fetch all details for 'current' record }
        bInclMetaData   : Bool;     { Include metadata in packet }
        out Packet      : PSafeArray{ returns datapacket with row description }
    ): DBResult; stdcall;
    
    { Embedded datasets }

    procedure CreateEmbeddedDS(Index: Integer);
    function GetEmbeddedDSBase(iFieldNo: DWord): TDSBase;
    function GetEmbeddedDSBaseByDescNo(DescNo: Integer): TDSBase;
    function GetParentSeqNo(pRecBuf: PChar): DWord;
  protected
    { IDSBase }

    function Create(                { Create empty dataset }
        iFields  : DWord;           { Number of fields }
        pFldDes  : pDSFLDDesc;      { Array of field descriptors }
        pszName  : PChar            { Name (optional) }
    ): DBResult; overload; stdcall;

    function AddField(              { Add a field to the dataset }
        pFldDes  : pDSFLDDesc       { Field descriptor }
    ): DBResult; stdcall;

    function AppendData(            { Appends data packet to dataset. }
        Packet   : PSafeArray;      { Data packet }
        bEof     : Bool             { If True, this is last packet }
    ): DBResult; stdcall;

    function GetOptParameter(       { Returns optional parameter (unknown to dataset) }
        iNo      : DWord;           { Number 1..iOptAttr }
        iFldNo   : DWord;           { 0 if not field attribute }
    var ppName   : Pointer;         { returns ptr to name }
    var piType   : DWord;           { returns type }
    var piLen    : DWord;           { returns length }
    var ppValue  : Pointer          { returns ptr to value }
    ): DBResult; stdcall;

    function AddOptParameter(       { Adds optional parameter to dataset }
        iFldNo   : DWord;           { 0 if not field attribute }
        pszAttr  : PChar;           { ptr to name }
        iType    : DWord;           { type }
        iLen     : DWord;           { length }
        pValue   : Pointer          { ptr to value }
    ): DBResult; stdcall;

    function GetProps(              { Get dataset properties }
    var Prop : DSProps
    ): DBResult; stdcall;

    function GetFieldDescs(         { Get field descriptors }
        Fields  : pDSFLDDesc        { Array of fields descriptors (output) }
    ): DBResult; stdcall;

    function GetIndexDescs(         { Get index descriptors }
        p1: PDSIDXDesc              { Array of index descriptors (output) }
    ): DBResult; stdcall;

    function GetDelta(              { Extract delta from dataset }
    out DsDelta: IDSBase            { Delta in a dataset }
    ): DBResult; stdcall;

    function StreamDS(              { Create data packet from the dataset }
    out Packet  : PSafeArray        { Return data packet }
    ): DBResult; stdcall;

    function AcceptChanges: DBResult; stdcall; { Accept all current changes }

    function PutBlank(              { Put blank value }
        pRecBuf      : Pointer;     { RecBuf OR }
        iRecNo       : DWord;       { iRecNo }
        iFldNo       : DWord;
        iBlankValue  : DWord
    ): DBResult; stdcall;

    function CreateIndex(           { Create, and add an index }
    const IdxDesc  : DSIDXDesc
    ): DBResult; stdcall;

    function RemoveIndex(           { Remove index of given name }
        pszName  : PChar
    ): DBResult; stdcall;

    function GetErrorString(        { Retrieve error string }
        iErrCode  : DBResult;
        pString   : PChar
    ): DBResult; stdcall;

    function FldCmp(                { Compare field values returns 0 if equal }
        iFldType  : DWord;          { Fieldtype }
        pFld1     : Pointer;        { NULL if blank }
        pFld2     : Pointer;        { NULL if blank }
        iUnits1   : DWord;
        iUnits2   : DWord
    ): Integer; stdcall;

    function GetProp(               { Get property }
        eProp       : DSProp;
        piPropValue : Pointer
    ): DBResult; stdcall;

    function SetProp(               { Set property }
        eProp      : DSProp;
        iPropValue : DWord
    ): DBResult; stdcall;

    function SetFieldCalculation(   { Register fieldcalculation on this field }
        iClientData  : DWord;       { Client data }
        pfCalc       : pfDSCalcField { Callback function, NULL to remove }
    ): DBResult; stdcall;

    function Reconcile(             { Reconcile update errors }
        DeltaPacket   : PSafeArray; { Delta data packet }
        ErrorPacket   : PSafeArray; { NULL if all changes accepted }
        iClientData   : DWord;
        pfReconcile   : pfDSReconcile { Callback-fn (called for each error) }
    ): DBResult; stdcall;

    { Place Holders for aggregate functions that were moved into DSCursor }
    function Reserved1(Input: Pointer): DBResult; stdcall;
    function Reserved2(Input: Pointer): DBResult; stdcall;
    function Reserved3(Input: Pointer): DBResult; stdcall;

    function Refresh(               { Refresh dataset }
        NewPacket    : PSafeArray;  { New updated packet }
        iClientData  : DWord;       { Client data }
        pfReconcile  : pfDSReconcile { Callback for resolving conflicts }
    ): DBResult; stdcall;

    function Reset: DBResult; overload; stdcall; { Remove all data from dataset }

    function RollBack(              { Rollback changes to this savepoint }
        iSavePoint  : SAVEPOINT
    ): DBResult; stdcall;

    function GetEmbeddedDS(
        iFieldID  : DWord;          { FieldID of embedded table (0 : get the first one) }
    out DsDet     : IDSBase         { Returns the ds of the embedded table }
    ): DBResult; stdcall;

    function MakeEmbedded(
        DsDet             : IDSBase;  { Embed this dataset }
        iFieldsLink       : DWord;
        piFieldsM         : PDWord;   { Fields in Master }
        piFieldsD         : PDWord;   { Fields in Detail }
        pMasterFieldName  : PChar;    { Name of new link field in master, NULL if using default name }
        pDetailFieldName  : PChar     { Name of new link field in detail, NULL if using defaultname }
    ): DBResult; stdcall;

    function RefreshRecords(            { Refresh specific records }
        NewDataPacket   : PSafeArray;   { Datapacket containing refreshed records }
        iRecNo          : DWord;        { Refresh this specific record (0 if more than one.Unique key req.) }
        iClientData     : DWord;
        pfReconcile     : pfDSReconcile { (NULL) Callback for resolving conflicts }
    ): DBResult; stdcall;

    function ReleaseBlobs(          { Release all uncommitted blobs }
        iBlobId  : DWord            { 0: all uncommitted blobs, otherwise : specific blob }
    ): DBResult; stdcall;

    function Clone(                 { Clones the structure of the dsbase, including details if any }
         iPType : DWord;            { 0:normal-ds, 1:delta-ds, 2:error-ds }
         bRecurse : Bool;           { TRUE:create complete tree-structure }
         bCloneOptParams: Bool;     { TRUE:clone all optional parameters (normal only) }
         var DataSet: IDSBase       { Returned dsbase }
    ): DBResult; stdcall;

    function Reconcile_MD(
        pDsRec          : IDSBase;           { Ds for info }
        pDeltaPacket    : PSafeArray;        { Delta pickle }
        pErrorPacket    : PSafeArray;        { NULL if all changes accepted }
        iClientData     : DWord;
        pfReconcile_MD  : pfDSReconcile_MD   { Callback-fn (called for each error) }
    ): DBResult; stdcall;

    { IDSWriter }

    function Init_Sequential(   { Initialze by sequentially adding columns }
        Version: DWord;
        Columns: Word
    ): DBResult; stdcall;

    function Init(
        Version: DWord;
        Columns: Word;
        FieldDesc: PDSDataPacketFldDesc
    ): DBResult; stdcall;

    function AddAttribute(      { Add an optional parameter }
        AAttrArea: TPcktAttrArea;
        AAttr: PChar;
        AAttrType: DWord;
        ALen: DWord;
        AValue: Pointer
    ): DBResult; stdcall;

    function GetDataPacket(     { Return pointer to the finished 'safearray' }
    var SA: PSafeArray
    ): DBResult; stdcall;

    function GetSize(           { Get the size of the safearray. }
    var DataPacketSize: DWord
    ): DBResult; stdcall;

    function IDSWriter.PutField = WriterPutField;          { Add this field to the data stream. }

    function WriterPutField(          { Add this field to the data stream. }
        AFldStatus: TPcktFldStatus;
        ALen: DWord;
        ASrc: Pointer
    ): DBResult; stdcall;

    function AddColumnDesc(     { Add a column descriptor. }
      const AFldDes: TDSDataPacketFldDesc
    ): DBResult; stdcall;

    { Reset all data (in order to create partial data).  Datapackets will not
      contain meta-info. Last created datapacket must be released by caller. }
    function IDSWriter.Reset = WriterReset;

    function WriterReset: DBResult; stdcall;

    { Return number of fielddescriptors, including embedded tables etc. }

    function GetColumnCount(var Count: DWord): DBResult; stdcall;

    function GetColumnDescs(    { Return all fielddescriptors }
        FieldDes: PDSDataPacketFldDesc
    ): DBResult; stdcall;

    function IDSWriter.GetErrorString = GetErrorString;

   { Signals last row for a nested table, in case rowcount was not known
     in advance }
   function EndOfNestedRows: DBResult; stdcall;


  public
    procedure Initialize; override;
    destructor Destroy; override;
  end;

{ TDSCursor }

  TDSCursor = class(TComObject, IDSCursor, IDSCursor5)
  private
    FIndexDS: DSIDXDesc;
    FDSBase: TDSBase;
    FIsDelta: Boolean;
    FIDSBase: IDSBase;
    FFields: DSFLDDescList;
    FRecNums: TList;
    FRecPos: Integer;
    FInUndo: Boolean;
    FDefaultOrder, FChangeIndex: Boolean;
    FCrack: Boolean;

    { Range }
    FKeyFields: DWord;
    FKeyRec1, FKeyRec2: PChar;
    FKeyRec1Incl, FKeyRec2Incl: Boolean;

    { Filters }
    FExprEval: Tvg2ExprEvaluator;
    FFClientData: DWord;       { Client supplied data }
    FFFilter: pfDSFilter;      { ptr to filter function }

    { Change callback }
    FCNClientData: DWord;
    FCNCallBack: pfCHANGECallBack;

    { Aggregates }
    FAggs, FAggsValid: TList;

    { Linked cursors }
    FMaster: TDSCursor;
    FMasterRecBuf: PChar;
    FDetails: TList;
    FFieldsM: DWordList;
    FFieldsD: DWordList;

    { Records and their order }
    function CurrentSeqNo: DWord;
    procedure SetRecPos(Value: Integer);

    function AcceptsRecord(pRecBuf: PChar): Boolean;

    { Records and order }
    procedure CheckIndexed;
    procedure CheckIndexWritable;

    function InternalMoveToKey(SearchCond: DBSearchCond; iFields,
      iPartLen: DWord; pKeyBuf, pRecBuf: Pointer; var RecNo: Integer): DBResult;
    function InsertRecNo(RecNo: Integer): Integer;
    procedure RemoveRecNo(RecNo: Integer);

    procedure Clear;
    procedure Rebuild;
    procedure Sort;

    procedure CheckCursorPos;
    procedure UpdateCursorPos;

    procedure RangeDropped;
    procedure NotifyChanged;
    procedure Notify(const Event, Data1, Data2, Data3: Integer);

    { Comparing records }
    function CmpRecsIndexEx(pRec1, pRec2: PChar; iFields: DWord): Integer;
    function CmpRecsIndex(pRec1, pRec2: PChar): Integer;
    function CmpRecsIndexDef(pRec1, pRec2: PChar): Integer;

    function GetCurrentRecordPtr(var pRecBuf: PChar): DBResult;

    { Aggregates }
    procedure GetAggBounds(iFields: Integer; RecNo: DWord; var R1, R2: DWord);
    procedure InvalidateAggs;

    { Linked cursors }
    procedure CheckFetched;
    procedure InsertCursor(ACursor: TDSCursor);
    procedure RemoveCursor(ACursor: TDSCursor);
    procedure ResyncDetails;

    property RecPos: Integer read FRecPos write SetRecPos;
  protected

    function InitCursor(            { Associate Cursor with a DataSet }
       DataSet: IDSBase
    ): DBResult; stdcall;

    function CloneCursor(           { Clone cursor from cursor }
       Cursor: IDSCursor
    ): DBResult; stdcall;

    function GetCursorProps(        { Get cursor properties }
    var p1: DSProps
    ): DBResult; stdcall;

    function GetIndexDescs(         { Get index descriptors }
        bCurrentOnly  : Bool;       { Only return 'current' indexdesc, otherwise all }
    var IdxDesc       : DSIDXDesc
    ): DBResult; stdcall;

    function GetFieldDescs(         { Get field descriptors }
        p1  : pDSFLDDesc
    ): DBResult; stdcall;

    function GetCurrentRecord(      { Return record at current cursorposition }
        pRecBuf  : Pointer
    ): DBResult; stdcall;

    function GetRecordBlock(        { Return block of records }
        piRecs   : PDWord;
        pRecBuf  : Pointer
    ): DBResult; stdcall;

    function GetCurrentBookMark(    { Get bookmark for current position }
        pBookMark  : Pointer
    ): DBResult; stdcall;

    function GetSequenceNumber(     { Get Sequence number of current position }
    var iSeq  : DWord
    ): DBResult; stdcall;

    function GetRecordAttribute(    { Get record attribute of current position }
    var Attr  : DSAttr
    ): DBResult; stdcall;

    function GetRecordCount(        { Number of records in active view }
    var iRecs  : Integer
    ): DBResult; stdcall;

    function MoveToBOF: DBResult; stdcall; { Set to beginning of table (BOF) }

    function MoveToEOF: DBResult; stdcall; { Set to end of table (EOF) }

    function MoveRelative(i: Integer): DBResult; stdcall;

    function MoveToSeqNo(i: DWord): DBResult; stdcall;

    function MoveToBookMark(        { Goto bookmark }
      pBookMark: Pointer
    ): DBResult; stdcall;

    function MoveToKey(             { Goto key }
        SearchCond  : DBSearchCond;
        iFields     : DWord;
        iPartLen    : DWord;
        pRecBuf     : Pointer
    ): DBResult; stdcall;

    function CompareBookMarks(      { Compare two bookmark (positions) -1, 0, 1 }
        pBookMark1  : Pointer;
        pBookMark2  : Pointer;
    var iCmp        : Integer
    ): DBResult; stdcall;

    function ExtractKey(            { Extract key from record }
        pRecBuf  : Pointer;
        pKeyBuf  : Pointer
    ): DBResult; stdcall;

    function GetRecordForKey(       { Return (first) record with given key }
        iFields   : DWord;
        iPartLen  : DWord;
        pKey      : Pointer;
        pRecBuf   : Pointer
    ): DBResult; stdcall;

    function GetField(              { Extract field value from record buffer }
        pRecBuf   : Pointer;
        iFieldNo  : DWord;
        pFldBuf   : Pointer;
    var bBlank    : Bool            { Returns TRUE/FALSE if blank }
    ): DBResult; stdcall;

    function PutField(              { Put field value into record buffer }
        pRecBuf   : Pointer;
        iFieldNo  : DWord;
        pFldBuf   : Pointer         { If NULL, adds a blank value }
    ): DBResult; stdcall;

    { Blob functions }

    function GetBlobLen(            { Return length of blob }
        pRecBuf   : Pointer;
        iFieldNo  : DWord;
    var iLength   : DWord
    ): DBResult; stdcall;

    function GetBlob(               { Return blob }
        pRecBuf   : Pointer;
        iFieldNo  : DWord;
        iOffSet   : DWord;          { Starting position }
        pBuf      : Pointer;
    var iLength   : DWord          { No of bytes to be read/ returns number read }
    ): DBResult; stdcall;

    function PutBlob(               { Write blob data }
        pRecBuf   : Pointer;
        iFieldNo  : DWord;
        iOffSet   : DWord;          { Starting position }
        pBuf      : Pointer;
        iLength   : DWord
    ): DBResult; stdcall;

    function InitRecord(            { Initialize record buffer (for insertion) }
        pRecBuf  : Pointer
    ): DBResult; stdcall;

    function DeleteRecord: DBResult; stdcall; { Delete current record }

    function ModifyRecord(          { Modify current record }
        pRecBuf  : Pointer
    ): DBResult; stdcall;

    function InsertRecord(          { Insert new record }
        pRecBuf  : Pointer
    ): DBResult; stdcall;

    function UndoLastChange(        { Undo last update }
        bFollowChange  : Bool
    ): DBResult; stdcall;

    function AddFilter(             { Add a canexpr-filter to this cursor }
        pcanExpr  : Pointer;        { Can expr }
        iLen      : DWord;          { Length of canexpr }
    var hFilter   : hDSFilter
    ): DBResult; stdcall;

    function DropFilter(            { Drop a filter }
       hFilter  : hDSFilter
    ): DBResult; stdcall;

    function SetRange(              { Set a range on a cursor }
        iFields    : DWord;
        pKey1      : Pointer;
        bKey1Incl  : Bool;
        pKey2      : Pointer;
        bKey2Incl  : Bool
    ): DBResult; stdcall;

    function DropRange: DBResult; stdcall; { Remove current range }

    function SortOnFields(          { Sort on fields }
        iFields    : DWord;
        piFields   : PDWord;        { NULL -> all fields }
        pDescending: PBool;         { NULL -> all ascending }
        pCaseInsensitive: PBool     { NULL -> all case-sensitive }
    ): DBResult; stdcall;

    function UseIndexOrder(         { Switch to index order }
        pszName  : PChar
    ): DBResult; stdcall;

    function SetNotifyCallBack(     { Called when posting changes to dataset }
        iClientData  : DWord;
        pfCallBack   : pfCHANGECallBack   { Call back fn being registered }
    ): DBResult; stdcall;

    function AddFilterCallBack(     { Add a canexpr-filter to this cursor }
        iClientData  : DWord;     { Client supplied data }
        pfFilter     : pfDSFilter;  { ptr to filter function }
    var hFilter      : hDSFilter
    ): DBResult; stdcall;

    function VerifyField(           { Verify if field value is valid }
        iFieldNo  : DWord;
        pFldBuf   : Pointer
    ): DBResult; stdcall;

    function GetProp(               { Get property }
        eProp        : CURProp;
        piPropValue  : Pointer
    ): DBResult; stdcall;

    function RevertRecord: DBResult; stdcall; { Restore current record }

    function LocateWithFilter(
        pCanExpr  : Pointer;        { NULL -> use previous }
        iLen      : DWord           { Length of canexpr }
    ): DBResult; stdcall;

    function AddAggregate(
        iFlds     : DWord;          { Defines grouping  (0 if global) }
        iCanLen   : DWord;          { Length of canexpr (0 if grouping only) }
        pCanExpr  : Pointer;        { Canexpression for aggregate }
    var hAgg      : hDSAggregate    { returns aggregate handle }
    ): DBResult; stdcall;

    function DropAggregate(
        hAgg  : hDSAggregate
    ): DBResult; stdcall;

    function GetAggregateValue(
        hAgg     : hDSAggregate;
        pValue   : Pointer;
    var bBlank   : Bool
    ): DBResult; stdcall;

    function GetAggregateDesc(
        hAgg   : hDSAggregate;
    var Desc   : DSFLDDesc
    ): DBResult; stdcall;

    function MoveToNextSubGroup(
        iFields  : DWord
    ): DBResult; stdcall;

    function GetSubGroupState(
        iFields        : DWord;
    var iSubGroupState : GROUPSTATE
    ): DBResult; stdcall;

    function LinkCursors(
        iFieldsM   : DWord;
        piFieldsM  : PDWord;         { Fields in Master }
        piFieldsD  : PDWord;         { Fields in Detail }
        hCurDet    : IDSCursor       { Detail cursor to link }
    ): DBResult; stdcall;

    function ResyncWithMaster: DBResult; stdcall; { If this is a detail, reset range }

    function SetProp(               { Set property }
        eProp       : CURProp;      { Property to set }
        iPropValue  : DWord         { Property value (or pointer to value) }
    ): DBResult; stdcall;

   function GetRecordNumber(  { Return record number of current record, if any }
        var iRecNo: DWord
    ): DBResult; stdcall;


    function GetRowRequestPacket(   { Get packet describing the curent 'path',
                                      for delayed fetching/refreshing }
        bKeysOnly       : Bool;     { Only include unique keys in packet }
        bFetchAllBlobs  : Bool;     { fetch all blobs for 'current'record }
        bFetchAllDetails: Bool;     { fetch all details for 'current' record }
        bInclMetaData   : Bool;     { Include metadata in packet }
        out Packet      : PSafeArray{ returns datapacket with row description }
    ): DBResult; stdcall;

   function RefreshRecord( { Refresh details/blobs for this record, and all
                             'current' records above, if any }
        Packet  : PSafeArray        { New updated pickle }
    ): DBResult; stdcall;

  public
    procedure Initialize; override;
    destructor Destroy; override;
  end;

const
  Counter: Integer = 0;

implementation
uses SysUtils, hbCLSID, hbConsts, hbErrors, hbUtils, hbStream, vg2SysUtils, vg2BCDUtils;

function BaseFromIDS(Base: IDSBase): TDSBase;
begin
  Base.GetProp(DSProp(-1), @Result);
end;

function CursorFromIDS(Cursor: IDSCursor): TDSCursor;
begin
  Cursor.GetProp(CURProp(-1), @Result);
end;

function CreateDSBase: IDSBase;
begin
  Result := TDSBase.Create as IDSBase;
end;

function CreateDSBaseCursor(DSBase: IDSBase; var Cursor: IDSCursor): DBResult;
begin
  Cursor := TDSCursor.Create as IDSCursor;
  Result := Cursor.InitCursor(DSBase);
end;

function CreateDSCloneCursor(SourceCursor: IDSCursor; var Cursor: IDSCursor): DBResult;
begin
  Cursor := TDSCursor.Create as IDSCursor;
  Result := Cursor.CloneCursor(SourceCursor);
end;

function CreateDSWriter: IDSWriter;
begin
  Result := TDSWriter.Create as IDSWriter;
end;

function CompareInt64(const Value1, Value2: Int64): Integer;
asm
      MOV       ECX,DWORD PTR [Value1]
      MOV       EDX,DWORD PTR [Value1 + 4]
      XOR       EAX,EAX
@C1:  CMP       EDX,DWORD PTR [Value2]
      JG        @@G
      JL        @@L
@C2:  CMP       ECX,DWORD PTR [Value2 + 4]
      JB        @@L
      JE        @@Q
@@G:  INC       EAX
      JMP       @@Q
@@L:  DEC       EAX
@@Q:
end;

function CompareUInt64(const Value1, Value2: Int64): Integer;
asm
      MOV       ECX,DWORD PTR [Value1]
      MOV       EDX,DWORD PTR [Value1 + 4]
      XOR       EAX,EAX
@C1:  CMP       EDX,DWORD PTR [Value2]
      JA        @@G
      JB        @@L
@C2:  CMP       ECX,DWORD PTR [Value2 + 4]
      JB        @@L
      JE        @@Q
@@G:  INC       EAX
      JMP       @@Q
@@L:  DEC       EAX
@@Q:
end;

type
  TDSExprEvaluator = class(Tvg2ExprEvaluator)
  private
    FDSBase: TDSBase;
    FRecBuf: PChar;
    FiFldType: Integer;
    FiFldLen : Integer;
  protected
    function DoGetFieldValue(FieldNo: Word; const FieldName: string;
      var FldType, FldLen: Integer): Variant; override;
    procedure SetRecBuf(pRecBuf: PChar);
  public
    function EvaluteRec(pRecBuf: PChar): Boolean;
    property DSBase: TDSBase read FDSBase write FDSBase;
    property iFldType: Integer read FiFldType;
    property iFldLen: Integer read FiFldLen;
  end;

{ TDSExprEvaluator }
procedure TDSExprEvaluator.SetRecBuf(pRecBuf: PChar);
begin
  FRecBuf := pRecBuf;
end;

function TDSExprEvaluator.EvaluteRec(pRecBuf: PChar): Boolean;
begin
  SetRecBuf(pRecBuf);
  try
    Result := Evaluate(FiFldType, FiFldLen)
  except
    on ENotSupported do
      Result := False
    else
      Result := False;
  end;
end;

function TDSExprEvaluator.DoGetFieldValue(FieldNo: Word; const FieldName: string;
  var FldType, FldLen: Integer): Variant;
var
  S: string;
  W: WideString;
  TS: TTimeStamp;
begin
  with FDSBase.FFields[FieldNo - 1] do
  begin
    FldType := iFldType;
    FldLen  := iFldLen - ord((iFldType = fldZSTRING));

    if not Assigned(FRecBuf) or ((FRecBuf + iNullOffsInRec)^ <> Char(NOT_BLANK)) then
    begin
      Result := Null;
      Exit;
    end;

    case iFldType of
      fldZSTRING, fldBYTES, fldVARBYTES:
        Result := StrPas(FRecBuf + iFldOffsInRec);
      fldUNICODE:
        begin
          W := SysAllocStringLen(PWideChar(FRecBuf + iFldOffsInRec + 2), PWord(FRecBuf + iFldOffsInRec)^);
          Result := W;
          SysFreeString(PWideChar(W));
        end;
      fldINT16:
        Result := PSmallInt(FRecBuf + iFldOffsInRec)^ ;
      fldUINT16:
        Result := PWord(FRecBuf + iFldOffsInRec)^ ;
      fldINT32:
        Result := PInteger(FRecBuf + iFldOffsInRec)^ ;
      fldUINT32:
        Result := Integer(PDWord(FRecBuf + iFldOffsInRec)^);
      fldBOOL:
        Result := PWordBool(FRecBuf + iFldOffsInRec)^ ;
      fldFLOAT:
        Result := PDouble(FRecBuf + iFldOffsInRec)^ ;
      fldBCD:
        begin
          SetString(S, PChar(FRecBuf + iFldOffsInRec), SizeOf(FMTBCD));
          Result := S;
        end;
      fldDATE:
        begin
          TS.Date := PDateTimeRec(FRecBuf + iFldOffsInRec).Date;
          TS.Time := 0;
          Result := TimeStampToDateTime(TS);
          FldType := fldTIMESTAMP;
          FldLen := SizeOf(TDateTime);
        end;
      fldTIME:
        begin
          TS.Date := DateDelta;
          TS.Time := PDateTimeRec(FRecBuf + iFldOffsInRec).Time;
          Result := TimeStampToDateTime(TS);
          FldType := fldTIMESTAMP;
          FldLen := SizeOf(TDateTime);
        end;
      fldTIMESTAMP:
        begin
          Result := PDateTimeRec(FRecBuf + iFldOffsInRec).DateTime;
        end;
      fldINT64, fldUINT64:
        Result := Integer(PInt64(FRecBuf + iFldOffsInRec)^);
      fldBLOB, fldCURSOR, fldADT, fldArray, fldTABLE, fldREF:
        Result := Null;
    else
      HBError(DBERR_NOTSUPPORTED);
    end;
  end;
end;

{ TDSBase }
procedure TDSBase.Initialize;
var
  Index: DSIDXDesc;
begin
  inherited;
  FRecords := TList.Create;
  FCursors := TList.Create;
  FBlobs   := TList.Create;
  FChanges := TList.Create;
  FChangesDS := TList.Create;
  FChangesVL := TList.Create;

  ZeroMem(@Index, SizeOf(Index));
  StrCopy(Index.szName, szDEFAULT_ORDER);
  Index.bUnique := True;
  Check(CreateIndex(Index));

  StrCopy(Index.szName, szCHANGEINDEX);
  Index.bUnique := False;
  Check(CreateIndex(Index));
  Clear;
end;

destructor TDSBase.Destroy;
begin
  FCursor := nil;
  FMDCursor := nil;
  FAICursor := nil;
  Clear;
  FIndexes := nil;
  FRecords.Free;
  FCursors.Free;
  FBlobs.Free;
  FChanges.Free;
  FChangesDS.Free;
  FChangesVL.Free;
  inherited;
end;

procedure TDSBase.Clear;
var
  I: Integer;
begin
  FCursor := nil;
  FMDCursor := nil;
  FAICursor := nil;

  for I := FCursors.Count - 1 downto 0 do
    RemoveCursor(FCursors[I]);

  FreeAttrInfos(FAttrInfos);

  Reset;

  FName       := '';

  FRecBufSize := Reserved;

  FLogChanges := True;
  FReadOnly   := False;
  FStreamedReadOnly := False;

  FLCID       := LOCALE_USER_DEFAULT;

  FFields     := nil;
  FWFldDescs  := nil;

  FLinkFieldNo:= 0;
  FLinkFieldCreated := False;

  for I := 0 to High(FChilds) do
    FChilds[I] := nil;

  FChilds := nil;
  SetLength(FIndexes, 2);
end;

procedure TDSBase.CursorNeeded(var ACursor: IDSCursor);
begin
  if not Assigned(ACursor) then
    Check(CreateDSBaseCursor(Self as IDSBase, ACursor));
end;

function TDSBase.AcceptChanges: DBResult;
var
  I: Integer;
  P: PDeltaRec;
  Attr: DSAttr;
  DeltaRecs: TList;
begin
  try
    DeltaRecs := TList.Create;
    try

      GetDeltaRecords(nil, DeltaRecs);

      for I := 0 to High(FChilds) do
         FChilds[I].AcceptChanges;

      I := 0;
      DisableNotify;
      try
        while I < DeltaRecs.Count do
        begin
          P := DeltaRecs[I];
          Attr := P.RecAttr;
          {
          if P.RecAttr = dsRecUnmodified then
            Attr := dsRecModified else
            Attr := P.RecAttr;
          }
          if Attr = dsRecModified then
          begin
            MergeRecord(Attr, PDeltaRec(DeltaRecs[I - 1]).RecNo, P.pRecBuf, DeltaRecs);
          end else if (Attr = dsRecNew) or (Attr = dsRecDeleted) then
          begin
            MergeRecord(P.RecAttr, P.RecNo, nil, DeltaRecs);
          end;
          Inc(I);
        end;
      finally
        EnableNotify;
      end;
      NotifyCursors(ceNotify, 0, 0, 0);
    finally
      ListFreeMem(DeltaRecs, True);
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.Clone(iPType: DWord; bRecurse, bCloneOptParams: Bool; var DataSet: IDSBase): DBResult;

  procedure CloneOptParams(Source, Dest: TDSBase; Response: Boolean);
  var
    I: Integer;
  begin
    with Dest do
    begin
      for I := 0 to High(Source.FAttrInfos) do
        with Source.FAttrInfos[I] do
          AddOptAttribute(FAttrInfos, iFieldNo + DWord(ord(Response)) * 6, Attr, AttrType, Len, Value);

      for I := 0 to High(FChilds) do
        CloneOptParams(BaseFromIDS(Source.FChilds[I]), BaseFromIDS(FChilds[I]), Response);
    end;
  end;

var
  DSBase: TDSBase;
  TmpFields: DSFLDDescList;

  procedure AddField(Index: Integer; pName: PChar; FieldType: Integer; Len: DWord);
  begin
    StrCopy(TmpFields[Index].szName, pName);
    TmpFields[Index].iFldType := FieldType;
    TmpFields[Index].iUnits1 := Len;
  end;

var
  Response: Boolean;
begin
  try
    DSBase := TDSBase.Create;
    DataSet := DSBase as IDSBase;
    Response := FIsDelta and (iPType = 2);

    GetFieldDescList(TmpFields, bRecurse, Response);
    Check(DataSet.Create(Length(TmpFields), Pointer(TmpFields), nil));

    if bCloneOptParams then
      CloneOptParams(Self, DSBase, Response);

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.CalcKeySize(var IndexDS: DSIDXDesc);
var
  I: Integer;
begin
  if not DefaultIndex(IndexDS.szName) then
    with IndexDS do
    begin
      iKeyLen := 0;
      for I := 0 to iFields - 1 do
        Inc(iKeyLen, FFields[iKeyFields[I] - 1].iFldLen);
    end;
end;

function TDSBase.ExtractKey(pRecBuf, pKeyBuf: PChar; const IndexDS: DSIDXDesc): DBResult;
var
  I: Integer;
  bBlank: Bool;
begin
  InitRecord(pKeyBuf);
  CopyRecord(pRecBuf, pKeyBuf);
  with IndexDS do
  begin
    for I := 0 to iFields - 1 do
      with FFields[iKeyFields[I]] do
      begin
        GetField(pRecBuf, I + 1, (pKeyBuf + iFldOffsInRec), bBlank);
        PutBlankInternal(pKeyBuf, I + 1, DWord(bBlank));
      end;
  end;
  Result := DBERR_NONE;
end;

function TDSBase.Create(iFields: DWord; pFldDes: pDSFLDDesc; pszName: PChar): DBResult;

  procedure DoAddField(BaseDS: TDSBase; ParentID: DWord; var DescNo: DWord;
    AddFldDesc: Boolean);
  var
    I, StartDescNo, Parent: DWord;
    FldDes: pDSFLDDesc;
    FldDesc: DSFLDDesc;
    Child: TDSBase;
  begin
    FldDes := @DSFLDDescList(pFldDes)[DescNo];
    BaseDS.AddFieldInternal(ParentID, FldDes, True);
    Inc(DescNo);
    with FldDes^ do
      if iFldType = fldADT then
      begin
        Parent := High(BaseDS.FFields) + 1;
        for I := 1 to iUnits1 do
          DoAddField(BaseDS, Parent, DescNo, AddFldDesc)
      end else if iFldType = fldARRAY then
      begin
        Parent := High(BaseDS.FFields) + 1;
        StartDescNo := DescNo;
        for I := 1 to iUnits1 do
        begin
          DescNo := StartDescNo;
          DoAddField(BaseDS, Parent, DescNo, (I = 1) and AddFldDesc);
        end;
      end else if iFldType = fldTABLE then
      begin
        with BaseDS do
          Child := BaseFromIDS(FChilds[High(FChilds)]);
        for I := 1 to iUnits1 do
          DoAddField(Child, 0, DescNo, True);
        if Child.FLinkFieldNo = 0 then
        begin
          InitLinkField(FldDesc);
          Child.AddFieldInternal(0, @FldDesc, False);
        end;
      end;
  end;

var
  I: DWord;
begin
  try
    Clear;
    if Assigned(pszName) then
      FName := StrPas(pszName);

    I := 0;
    while I < iFields do
      DoAddField(Self, 0, I, True);

    Result := DBERR_NONE;
  except
    Clear;
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.GetFieldDescList(var AFields: DSFLDDescList; Recurse, Response: Boolean);

  procedure DoAddFieldDesc(var FieldNo: Integer; AddFldDesc: Boolean);
  var
    I: Integer;
    Child: TDSBase;
  begin
    if AddFldDesc then
    begin
      AddFieldDesc(AFields, FFields[FieldNo]);
      if Response then
        with AFields[High(AFields)] do
          iFldAttr := iFldAttr and not (fldAttrREQUIRED or fldAttrREADONLY);
    end;

    Inc(FieldNo);
    with FFields[FieldNo - 1] do
      case iFldType of
        fldADT:
          begin
            for I := 1 to iUnits1 do
              DoAddFieldDesc(FieldNo, AddFldDesc);
          end;
        fldARRAY:
          begin
            for I := 1 to iUnits1 do
              DoAddFieldDesc(FieldNo, (I = 1) and AddFldDesc);
          end;
        fldTABLE:
          begin
            if Recurse and AddFldDesc then
            begin
              Child := GetEmbeddedDSBase(FieldNo);
              Inc(AFields[High(AFields)].iUnits1, ord(Response) * 6);
              Child.GetFieldDescList(AFields, Recurse, Response);
            end;
          end;
      end;
  end;

  procedure AddField(Index: Integer; pName: PChar; FieldType: Integer; Len: DWord);
  begin
    StrCopy(AFields[Index].szName, pName);
    AFields[Index].iFldType := FieldType;
    AFields[Index].iUnits1 := Len;
  end;

var
  FieldNo, Start: Integer;
begin
  if Response then
  begin
    Start := Length(AFields);
    SetLength(AFields, Start + 6);
    AddField(Start + 0, 'ERROR_RECORDNO', fldINT32,   0);
    AddField(Start + 1, 'ERROR_RESPONSE', fldINT32,   0);
    AddField(Start + 2, 'ERROR_MESSAGE',  fldZSTRING, MAXMSGLEN + 1);
    AddField(Start + 3, 'ERROR_CONTEXT',  fldZSTRING, MAXMSGLEN + 1);
    AddField(Start + 4, 'ERROR_CATEGORY', fldINT32,   0);
    AddField(Start + 5, 'ERROR_CODE',     fldINT32,   0);
  end;

  FieldNo := 0;
  while FieldNo <= High(FFields) do
    DoAddFieldDesc(FieldNo, True);
end;

procedure TDSBase.AddFieldInternal(ParentID: Word; pFldDes: pDSFLDDesc;
  CreateEmbedded: Boolean);
var
  I, J: Integer;
begin
  if Assigned(pFldDes) then
  begin
    AddFieldDesc(FFields, pFldDes^);

    FFields[High(FFields)].iFieldIDParent := ParentID;

    case pFldDes^.iFldType of
      fldUNKNOWN:
        HBError(DBERR_NOTSUPPORTED);
      fldZSTRING:
        FFields[High(FFields)].iFldLen := FFields[High(FFields)].iUnits1 + 1;
      fldUNICODE:
        FFields[High(FFields)].iFldLen := FFields[High(FFields)].iUnits1 * 2 + 2;
      fldINT16, fldUINT16:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(SmallInt);
          FFields[High(FFields)].iUnits1 := SizeOf(SmallInt);
        end;
      fldINT32, fldUINT32:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(Integer);
          FFields[High(FFields)].iUnits1 := SizeOf(Integer);
        end;
      fldBOOL:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(WordBool);
          FFields[High(FFields)].iUnits1 := SizeOf(WordBool);
        end;
      fldFLOAT:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(Double);
          FFields[High(FFields)].iUnits1 := SizeOf(Double);
        end;
      fldBCD:
         FFields[High(FFields)].iFldLen := SizeOf(FMTBCD);
      fldDATE, fldTIME:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(Integer);
          FFields[High(FFields)].iUnits1 := SizeOf(Integer);
        end;
      fldTIMESTAMP:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(Double);
          FFields[High(FFields)].iUnits1 := SizeOf(Double);
        end;
      fldBYTES, fldVARBYTES:
        FFields[High(FFields)].iFldLen := FFields[High(FFields)].iUnits1;
      fldBLOB:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(TBlobData);
        end;
      fldCURSOR:
        HBError(DBERR_NOTSUPPORTED);
      fldINT64, fldUINT64:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(Int64);
          FFields[High(FFields)].iUnits1 := SizeOf(Int64);
        end;
      fldADT, fldArray:
        begin
          FFields[High(FFields)].iFldLen := 0;
        end;
      fldTABLE:
        begin
          FFields[High(FFields)].iFldLen := SizeOf(DWord);
          if CreateEmbedded then
            CreateEmbeddedDS(High(FFields));
        end;
    else
      HBError(DBERR_NOTSUPPORTED);
    end;

    if (pFldDes.iFldAttr and fldAttrLink) <> 0 then
      FLinkFieldNo := Length(FFields);

    if High(FFields) = 0 then
    begin
      FFields[0].iNullOffsInRec := Reserved;
      FFields[0].iFldOffsInRec  := Reserved + SizeOf(Char);
    end else begin
      I := High(FFields) - 1;
      J := FFields[I].iFldOffsInRec + FFields[I].iFldLen;
      FFields[High(FFields)].iNullOffsInRec := J;
      FFields[High(FFields)].iFldOffsInRec  := J + SizeOf(Char);
    end;
    Inc(FRecBufSize, SizeOf(Char) + FFields[High(FFields)].iFldLen);
  end else
    Clear;
end;

function TDSBase.AddField(pFldDes: pDSFLDDesc): DBResult;
begin
  try
    AddFieldInternal(0, pFldDes, True);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.CheckInactiveIndex(szName: PChar);
var
  I: Integer;
begin
  if DefaultIndex(szName) then
    HBError(DBERR_ACTIVEINDEX)
  else begin
    for I := 0 to FCursors.Count - 1 do
      if (StrIComp(TDSCursor(FCursors[I]).FIndexDS.szName, szName) = 0) then
        HBError(DBERR_ACTIVEINDEX);
  end;
end;

procedure TDSBase.GetIndexDesc(IndexNo: DWord; p1: PDSIDXDesc);
begin
  Move(FIndexes[IndexNo], p1^, SizeOf(DSIDXDesc));
end;

function TDSBase.FindIndex(pszName: PChar): Integer;
var
  Tmp: string;
begin
  Tmp := StrPas(pszName);
  for Result := 0 to Length(FIndexes) - 1 do
    if CompareText(Tmp, FIndexes[Result].szName) = 0 then Exit;
  Result := -1;
end;

procedure TDSBase.UpdateLinkedIndex(var IdxDesc: DSIDXDesc);
begin
  with FIndexes[High(FIndexes)] do
    if (iFields > 0) and (iKeyFields[0] <> Integer(FLinkFieldNo)) then
    begin
      Move(iKeyFields[0], iKeyFields[1], iFields * SizeOf(iKeyFields[0]));
      iKeyFields[0] := FLinkFieldNo;
      Inc(iFields);
    end;
end;

function TDSBase.CreateIndex(const IdxDesc: DSIDXDesc): DBResult;
var
  I: Integer;
begin
  try
    if IdxDesc.szName <> nil then
    begin
      I := FindIndex(IdxDesc.szName);
      if I < 0 then
      begin
        SetLength(FIndexes, Length(FIndexes) + 1);
        Move(IdxDesc, FIndexes[High(FIndexes)], SizeOf(DSIDXDesc));
        if Assigned(FParentDS) then
          UpdateLinkedIndex(FIndexes[High(FIndexes)]);
        CalcKeySize(FIndexes[High(FIndexes)]);
        Result := DBERR_NONE;
      end else
        Result := DBERR_NAMENOTUNIQUE;
    end else
      Result := DBERR_INDEXNAMEREQUIRED;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.FldCmpEx(iFldType: DWord; pFld1, pFld2: Pointer; iUnits1,
  iUnits2: DWord; CaseIns: Bool): Integer;

const
  Cases: array[Boolean] of Integer = (SORT_STRINGSORT, SORT_STRINGSORT or NORM_IGNORECASE);

var
  I1, I2: Integer;
  W1, W2: DWord;
  I641, I642: Int64;
  B1, B2: WordBool;
  D1, D2: Double;
  F1, F2: Double;
begin
  if (pFld1 <> nil) and (pFld2 = nil) then
  begin
    Result := 1;
    Exit;
  end else if (pFld1 = nil) and (pFld2 <> nil) then
  begin
    Result := -1;
    Exit;
  end else if (pFld1 = nil) and (pFld2 = nil) then
  begin
    Result := 0;
    Exit;
  end;

  case iFldType of
    fldZSTRING, fldBYTES, fldVARBYTES:
      begin
        Result := CompareString(FLCID, Cases[CaseIns], pFld1, -1, pFld2, -1) - 2;
      end;
    fldUNICODE:
      begin
        Result := CompareStringW(FLCID, Cases[CaseIns],
          PWideChar(PChar(pFld1) + 2), PWord(pFld1)^,
          PWideChar(PChar(pFld2) + 2), PWord(pFld2)^) - 2;
      end;
    fldINT16:
      begin
        I1 := PSmallint(pFld1)^;
        I2 := PSmallint(pFld2)^;
        if I1 > I2 then Result :=  1 else
        if I1 < I2 then Result := -1 else Result := 0;
      end;
    fldUINT16:
      begin
        W1 := PWord(pFld1)^;
        W2 := PWord(pFld2)^;
        if W1 > W2 then Result :=  1 else
        if W1 < W2 then Result := -1 else Result := 0;
      end;
    fldINT32:
      begin
        I1 := PInteger(pFld1)^;
        I2 := PInteger(pFld2)^;
        if I1 > I2 then Result :=  1 else
        if I1 < I2 then Result := -1 else Result := 0;
      end;
    fldUINT32:
      begin
        W1 := PDWord(pFld1)^;
        W2 := PDWord(pFld2)^;
        if W1 > W2 then Result :=  1 else
        if W1 < W2 then Result := -1 else Result := 0;
      end;
    fldBOOL:
      begin
        B1 := PWordBool(pFld1)^;
        B2 := PWordBool(pFld2)^;
        if B1 > B2 then Result :=  1 else
        if B1 < B2 then Result := -1 else Result := 0;
      end;
    fldFLOAT:
      begin
        F1 := PDouble(pFld1)^;
        F2 := PDouble(pFld2)^;
        if F1 > F2 then Result :=  1 else
        if F1 < F2 then Result := -1 else Result := 0;
      end;
    fldBCD:
      begin
        Result := CompareFMTBCD(pFMTBcd(pFld1)^, pFMTBcd(pFld2)^);
      end;
    fldDATE:
      begin
        W1 := PDateTimeRec(pFld1)^.Date;
        W2 := PDateTimeRec(pFld2)^.Date;
        if W1 > W2 then Result :=  1 else
        if W1 < W2 then Result := -1 else Result := 0;
      end;
    fldTIME:
      begin
        W1 := PDateTimeRec(pFld1)^.Time;
        W2 := PDateTimeRec(pFld2)^.Time;
        if W1 > W2 then Result :=  1 else
        if W1 < W2 then Result := -1 else Result := 0;
      end;
    fldTIMESTAMP:
      begin
        D1 := PDateTimeRec(pFld1)^.DateTime;
        D2 := PDateTimeRec(pFld2)^.DateTime;
        if D1 > D2 then Result :=  1 else
        if D1 < D2 then Result := -1 else Result := 0;
      end;
    fldINT64:
      begin
        I641 := PInt64(pFld1)^;
        I642 := PInt64(pFld2)^;
        Result := CompareInt64(I641, I642);
      end;
    fldUINT64:
      begin
        I641 := PInt64(pFld1)^;
        I642 := PInt64(pFld2)^;
        Result := CompareUInt64(I641, I642);
      end;
    else
      Result := 0;
  end;
end;

function TDSBase.FldCmp(iFldType: DWord; pFld1, pFld2: Pointer; iUnits1,
  iUnits2: DWord): Integer;
begin
  Result := FldCmpEx(iFldType, pFld1, pFld2, iUnits1, iUnits2, False)
end;

function TDSBase.CmpRecsEx(IndexDS: pDSIDXDesc; iFields, iPartLen: Integer; pRec1, pRec2: PChar): Integer;
var
  I, I1, I2: Integer;
  pFld1, pFld2: PChar;
begin
  Result := 0;
  if pRec1 = pRec2 then Exit;

  if not Assigned(IndexDS) then
  begin
    I1 := GetRecordRecNo(pRec1);
    I2 := GetRecordRecNo(pRec2);
    if I1 > I2 then Result :=  1 else
    if I1 < I2 then Result := -1 else Result := 0;
    Exit;
  end;

  if (iFields = 0) or (iFields > IndexDS.iFields) then
    iFields := IndexDS.iFields;

  for I := 0 to iFields - 1 do
  begin
    with FFields[IndexDS.iKeyFields[I] - 1] do
    begin
      if (pRec1 + iNullOffsInRec)^ = Char(NOT_BLANK) then
        pFld1 := pRec1 + iFldOffsInRec else
        pFld1 := nil;

      if (pRec2 + iNullOffsInRec)^ = Char(NOT_BLANK) then
        pFld2 := pRec2 + iFldOffsInRec else
        pFld2 := nil;

      if iPartLen = 0 then
        Result := FldCmpEx(iFldType, pFld1, pFld2, iFldLen, iFldLen, IndexDS.bCaseInsensitive[I]) else
        Result := FldCmpEx(iFldType, pFld1, pFld2, iPartLen, iPartLen, IndexDS.bCaseInsensitive[I]);
    end;

    if Result <> 0 then
    begin
      if IndexDS.bDescending[I] then
        Result := - Result;
      Break;
    end;
  end;
end;

procedure TDSBase.CreateEmbeddedDS(Index: Integer);
var
  DSBase: IDSBase;
  DSBaseA: TDSBase;
begin
  DSBase  := CreateDSBase;
  DSBaseA := BaseFromIDS(DSBase);
  SetLength(FChilds, Length(FChilds) + 1);
  FChilds[High(FChilds)] := DSBase;
  DSBaseA.FParentDS := Self;
  DSBaseA.FParentFieldNo := Index + 1;
end;

function TDSBase.GetEmbeddedDS(iFieldID: DWord; out DsDet: IDSBase): DBResult;
var
  I, J: DWord;
begin
  try
    J := 0;
    for I := 0 to High(FFields) do
      if FFields[I].iFldType = fldTABLE then
      begin
        if I + 1 = iFieldID then
        begin
          DsDet := FChilds[J];
          Break;
        end;
        Inc(J);
      end else if (FFields[I].iFldAttr and fldAttrLink) <> 0 then
        Dec(iFieldID);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetEmbeddedDSBase(iFieldNo: DWord): TDSBase;
var
  DSBase: IDSBase;
begin
  Check(GetEmbeddedDS(iFieldNo, DsBase));
  Result := BaseFromIDS(DSBase);
end;

function TDSBase.GetEmbeddedDSBaseByDescNo(DescNo: Integer): TDSBase;
var
  I, J: Integer;
  DsDet: IDSBase;
begin
  J := 0;
  for I := 0 to High(FWFldDescs) do
  begin
    if PacketFieldType(FWFldDescs[I].FldDes.iFieldType) = dsfldEMBEDDEDTBL then
    begin
      if I = DescNo then
      begin
        DsDet := FChilds[J];
        Break;
      end;
      Inc(J);
    end;
  end;
  Result := BaseFromIDS(DsDet);
end;

function TDSBase.MakeEmbedded(DsDet: IDSBase; iFieldsLink: DWord;
  piFieldsM, piFieldsD: PDWord; pMasterFieldName,
  pDetailFieldName: PChar): DBResult;
begin
  Result := DBERR_NIY;
end;

function TDSBase.GetParentSeqNo(pRecBuf: PChar): DWord;
var
  bBlank: Bool;
  I: Integer;
  RecNo: Integer;
begin
  if Assigned(FParentDS) then
  begin
    GetField(pRecBuf, FLinkFieldNo, @RecNo, bBlank);
    with FParentDS.FRecords do
      for I := Count - 1 downto 0 do
      begin
        pRecBuf := List^[I];
        if GetRecordRecNo(pRecBuf) = RecNo then
        begin
          Result := I;
          Exit;
        end;
      end;
  end;
  Result := DWord(-1);
end;

function TDSBase.GetErrorString(iErrCode: DBResult; pString: PChar): DBResult;
begin
  try
    if (iErrCode = DBERR_CONSTRAINTFAILED) then
      StrCopy(pString, PChar(FConstrErrorMsg))
    else
      GetErrorMessage(iErrCode, pString);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetIndexDescs(p1: PDSIDXDesc): DBResult;
begin
  Move(Pointer(FIndexes)^, p1^, (High(FIndexes) + 1) * SizeOf(DSIDXDesc));
  Result := DBERR_NONE;
end;

function TDSBase.AddOptParameter(iFldNo: DWord; pszAttr: PChar; iType,
  iLen: DWord; pValue: Pointer): DBResult;
var
  Value: PChar;
begin
  try
    // String parameters begin from the length word.
    if (((iType and dsTypeBitsMask) shr dsSizeBitsLen) = dsfldZSTRING) then
    begin
      GetMem(Value, iLen + SizeOf(Word));
      try
        PWord(Value)^ := iLen;
        StrCopy(Value + SizeOf(Word), pValue);
        hbUtils.AddOptAttribute(FAttrInfos, iFldNo, pszAttr, iType, iLen + SizeOf(Word), Value);
      finally
        FreeMem(Value);
      end;
    end else begin
      if iType and (dsfldBYTES shl dsSizeBitsLen) <> 0 then
      begin
        GetMem(Value, iLen + SizeOf(Integer));
        try
          PInteger(Value)^ := iLen;
          Move(pValue^, (Value + SizeOf(Integer))^, iLen);
          hbUtils.AddOptAttribute(FAttrInfos, iFldNo, pszAttr, iType, iLen + SizeOf(Integer), Value);
        finally
          FreeMem(Value);
        end;
      end else
        hbUtils.AddOptAttribute(FAttrInfos, iFldNo, pszAttr, iType, iLen, pValue);
    end;

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetOptParameter(iNo, iFldNo: DWord; var ppName: Pointer;
  var piType, piLen: DWord; var ppValue: Pointer): DBResult;
begin
  Result := hbUtils.GetOptAttribute(FAttrInfos, iNo, iFldNo, ppName, piType, piLen, ppValue);
  // String parameters begin from the length word.
  if (((piType and dsTypeBitsMask) shr dsSizeBitsLen) = dsfldZSTRING) then
  begin
    Inc(PChar(ppValue), SizeOf(Word));
    Dec(piLen, SizeOf(Word));
  end;
end;

function TDSBase.GetProp(eProp: DSProp; piPropValue: Pointer): DBResult;
begin
  case eProp of
    dspropLOGCHANGES:
      PBool(piPropValue)^ := FLogChanges;
    dspropREADONLY:
      PBool(piPropValue)^ := FReadOnly;
    dspropNOOFCHANGES:
      PInteger(piPropValue)^ := FChangeCount;//FChangesAll.Count div 2;
    dspropCONSTRAINTS_DISABLED:
      PBool(piPropValue)^ := FConstrDisabled;
    dspropDSISPARTIAL:
      PBool(piPropValue)^ := FDSIsPartial;
    dspropRECORDSINDS:
      PInteger(piPropValue)^ := FRecords.Count;
    dspropAUTOINC_DISABLED:
      PBool(piPropValue)^ := FAutoIncDisabled;
    dspropISDELTA:
      PBool(piPropValue)^ := FIsDelta;
    dspropDONTINCLMETADATA:
      PBool(piPropValue)^ := FNoMetaData;
    dspropINCLBLOBSINDELTA:
      pDWord(piPropValue)^ := FBlobsInDelta;
    dspropGETSAVEPOINT:
      PInteger(piPropValue)^ := FChangesDS.Count;
    dspropCOMPRESSARRAYS:
      ;
    dspropMD_SEMANTICS:
      PDWord(piPropValue)^ := FMDOptions;
    dspropFIELD_FULLNAME:
      ;
    dspropFIELDID_FORNAME:
      ;
    dspropFIELDID_FORPARENT:
      ;
    dspropCHANGEINDEX_VIEW:
      ;
    dspropGETUNIQUEINDEX:
      ;
    dspropREMOTE_UPDATEMODE:
      ;
    dspropXML_STREAMMODE:
      PDWord(piPropValue)^ := FXMLStreamMode;
    else if eProp = DSProp(-1) then
      TDSBase(piPropValue^) := Self;
  end;
  Result := DBERR_NONE;
end;

function TDSBase.GetProps(var Prop: DSProps): DBResult;
var
  TmpFields: DSFLDDescList;
begin
  try
    ZeroMem(@Prop, SizeOf(DSProps));
    with Prop do
    begin
      StrCopy(szName, PChar(FName));
      GetFieldDescList(TmpFields, False, False);
      iFields          := Length(TmpFields);
      iRecBufSize      := FRecBufSize;
      iBookMarkSize    := SizeOf(DSBOOKMRK);
      bReadOnly        := FReadOnly or FStreamedReadOnly;
      iIndexes         := Length(FIndexes);
      iOptParams       := Length(FAttrInfos);
      bDelta           := FIsDelta;
      iLCID            := FLCID;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetFieldDescs(Fields: pDSFLDDesc): DBResult;
var
  TmpFields: DSFLDDescList;
begin
  try
    GetFieldDescList(TmpFields, False, False);
    Move(Pointer(TmpFields)^, Fields^, Length(TmpFields) * SizeOf(DSFLDDesc));
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.PutBlankParent(pRecBuf: PChar; iFldNo: DWord);
var
  I: Integer;
  iBlankValue: DWord;
begin
  with FFields[iFldNo - 1] do
  begin
    iBlankValue := BLANK_NULL;

    for I := iFldNo to iFldNo + DWord(iUnits1) do
      with FFields[I] do
        if (PChar(pRecBuf) + iNullOffsInRec)^ = Char(NOT_BLANK) then
        begin
          iBlankValue := NOT_BLANK;
          Break;
        end;

    PutBlankInternal(pRecBuf, iFldNo, iBlankValue);

    if (iBlankValue = NOT_BLANK) and (iFieldIDParent <> 0) then
      PutBlankParent(pRecBuf, iFieldIDParent);
  end;
end;

procedure TDSBase.PutBlankValues(pRecBuf: Pointer; iFldNo, iBlankValue: DWord);
var
  I: Integer;
begin
  with FFields[iFldNo - 1] do
  begin
    if (iFldType = fldBLOB) and (iBlankValue <> NOT_BLANK) then
      PutBlobInternal(pRecBuf, iFldNo, 0, nil, 0);

    PutBlankInternal(pRecBuf, iFldNo, iBlankValue);

    if (iFieldIDParent <> 0) then
      PutBlankParent(pRecBuf, iFieldIDParent);

    { Clear children }
    if (iFldType in [fldADT, fldARRAY]) and (iBlankValue <> NOT_BLANK) then
      for I := iFldNo to iFldNo + DWord(iUnits1) do
        PutBlankValues(pRecBuf, I + 1, iBlankValue);
  end;
end;

procedure TDSBase.PutBlankInternal(pRecBuf: Pointer; iFldNo, iBlankValue: DWord);
begin
  with FFields[iFldNo - 1] do
    (PChar(pRecBuf) + iNullOffsInRec)^ := Char(iBlankValue);
end;

function TDSBase.PutBlank(pRecBuf: Pointer; iRecNo, iFldNo, iBlankValue: DWord): DBResult;
begin
  try
    if not Assigned(pRecBuf) then
      pRecBuf := GetRecordPtr(iRecNo);
    PutBlankValues(pRecBuf, iFldNo, iBlankValue);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetResponses(DeltaDS, ErrorDS: TDSBase; FieldNums: TList;
  iClientData: DWord; pfReconcile_MD: pfDSReconcile_MD): Boolean;
var
  iRslt         : dsCBRType;  { Result of previous callback. If set, the previuos parameters are repeated. }
  iAction       : DSAttr;   { Update request Insert/Modify/Delete }

  pRecBuf       : PChar;

  pRecUpd       : PChar;    { Record that failed update }
  pRecOrg       : PChar;    { Original record, if any }
  pRecConflict  : PChar;    { Conflicting error, if any }

  iErrResponse  : dsCBRType; { Resolver response }
  iErrCode      : Integer;  { Native error-code, (BDE or ..) }
  pErrMessage   : array[0..MAXMSGLEN + 1] of Char;
  pErrContext   : array[0..MAXMSGLEN + 1] of Char;

  ErrorCur, DeltaCur: TDSCursor;

  I, DeltaRecNo: Integer;
  Child, DeltaD, ErrorD: TDSBase;

  HasConflicts: Boolean;

  bBlank: Bool;
begin
  Result := True;

  if not Assigned(FResponses) then
  begin
    FResponses := TList.Create;
    FResponses.Capacity := ErrorDS.FRecords.Count * 2;
  end;

  if Assigned(FParentDS) then
    FieldNums.Add(Pointer(FParentFieldNo));

  DeltaCur := CursorFromIDS(DeltaDS.FRCursor);
  ErrorCur := CursorFromIDS(ErrorDS.FRCursor);

  iRslt := 0;

  AllocRecord(pRecConflict);
  try
    while ErrorCur.MoveRelative(1) <> DBERR_EOF do
    begin
      with ErrorCur do
      begin
        GetCurrentRecordPtr(pRecBuf);
        GetField(pRecBuf, 1, @DeltaRecNo, bBlank);
      end;

      Check(DeltaCur.MoveToSeqNo(DeltaRecNo));

      // Process detail datasets
      for I := 0 to High(FChilds) do
      begin
        Child   := BaseFromIDS(FChilds[I]);
        DeltaD  := BaseFromIDS(DeltaDS.FChilds[I]);
        ErrorD  := BaseFromIDS(ErrorDs.FChilds[I]);
        Result  := Child.GetResponses(DeltaD, ErrorD, FieldNums, iClientData, pfReconcile_MD);
        if not Result then Exit;
      end;

      with ErrorCur do
      begin
        GetField(pRecBuf, 2, @iErrResponse, bBlank);
        // Only detail records have errors
        if bBlank then Continue;
        GetField(pRecBuf, 3, @pErrMessage,  bBlank);
        GetField(pRecBuf, 4, @pErrContext,  bBlank);
        GetField(pRecBuf, 6, @iErrCode,     bBlank);
      end;

      InitRecord(pRecConflict);

      with ErrorDS do
      begin
        HasConflicts := False;
        for I := 6 to High(FFields) do
          with FFields[I] do
            if (iFldType <> fldTABLE) and ((iFldAttr and fldAttrLink) = 0) and
              ((pRecBuf + iNullOffsInRec)^ <> Char(BLANK_NOTCHANGED)) then
            begin
              HasConflicts := True;
              CopyField(FFields[I], Self.FFields[I - 6], pRecBuf, pRecConflict);
            end;
      end;

      with DeltaCur do
      begin
        GetCurrentRecordPtr(pRecBuf);
        pRecUpd := pRecBuf;

        iAction := GetRecordAttr(pRecBuf) and dsIsModified;

        // Store number of record
        if (iAction in [dsRecNew, dsRecDeleted]) then
          FResponses.Add(FRecNums.List^[FRecPos]) else
          FResponses.Add(FRecNums.List^[FRecPos - 1]);

        if iAction = dsRecNew then
          pRecOrg := nil
        else if iAction = dsRecDeleted then
          pRecOrg := pRecUpd
        else begin
          MoveRelative(-1);
          GetCurrentRecordPtr(pRecOrg);
        end;
      end;

      if iErrCode <> 0 then
      begin
        if Assigned(pfReconcile_MD) then
          iRslt := pfReconcile_MD(
            iClientData, iRslt, iAction,
            iErrResponse, iErrCode, pErrMessage, pErrContext,
            PByte(pRecUpd), PByte(pRecOrg), PByte(ord(HasConflicts) * Integer(pRecConflict)),
            FieldNums.Count, PInteger(FieldNums.List))
        else if iAction <> dscbrSkip then
          iRslt := dscbrABORT;
      end else begin
        iRslt := dscbrCORRECT;
      end;

      if not HasConflicts and (iRslt in [dscbrMERGE, dscbrREFRESH]) then
        iRslt := dscbrSKIP;

      case iRslt of
        dscbrABORT:
          begin
            Result := False;
            Exit;
          end;
        dscbrMERGE:
          begin
            if iAction <> dsRecNew then
              AssignFields(pRecOrg, pRecConflict, True);
          end;
        dscbrCORRECT:
          { New values where placed in the pRecUpd buffer }
          begin
            if iErrCode = 0 then
              AssignFields(pRecUpd, pRecConflict, True);
          end;
        dscbrREFRESH:
          begin
            if iAction <> dsRecNew then
              AssignFields(pRecUpd, pRecConflict, True);
          end;
      end;

      // Store response for record
      FResponses.Add(Pointer(iRslt));
    end;
  finally
    FreeMem(pRecConflict);
  end;

  if Assigned(FParentDS) then
    FieldNums.Count := FieldNums.Count - 1;
end;

procedure TDSBase.MergeRecords(DeltaDS: TDSBase);

  // Returns a corrected resolver response for record of the Delta dataset
  function GetResponse(DeltaRecNo: Integer): dsCBRType;
  var
    I, J, Tmp: Integer;
  begin
    with FResponses do
      if Count > 0 then
        for I := 0 to (Count - 1) div 2 do
        begin
          J := I * 2;
          Tmp := Integer(List^[J]);
          if Tmp = DeltaRecNo then
          begin
            Result := Integer(List^[J + 1]);
            Exit;
          end;
        end;
    Result := dscbrMERGE;
  end;

var
  Response      : dsCBRType;
  iAction       : DSAttr;   { Update request Insert/Modify/Delete }

  Child, DeltaD: TDSBase;
  DeltaCur: IDSCursor;
  DeltaCurA: TDSCursor;

  I, J: Integer;
  P, OldP: PDeltaRec;
  DeltaRecs: TList;

  pRecBuf: PChar;
begin
  // Process detail datasets
  for I := 0 to High(FChilds) do
  begin
    Child   := BaseFromIDS(FChilds[I]);
    DeltaD  := BaseFromIDS(DeltaDS.FChilds[I]);
    Child.MergeRecords(DeltaD);
  end;

  DeltaRecs := TList.Create;
  try
    GetDeltaRecords(nil, DeltaRecs);

    CreateDSBaseCursor(DeltaDS as IDSBase, DeltaCur);
    DeltaCurA := CursorFromIDS(DeltaCur);

    DisableNotify;
    try
      J := 0;
      while J < DeltaRecs.Count do
      begin
        P := DeltaRecs.List^[J];
        if P.RecAttr = dsRecUnmodified then
        begin
          Inc(J);
          Continue;
        end;

        if P.RecAttr = dsRecModified then
          OldP := DeltaRecs.List^[J - 1] else
          OldP := P;

        iAction := P.RecAttr;

        Response := GetResponse(J);

        case dsCBRType(Response) of
          dscbrMERGE         :   { Merge the changes reconcile : update original. Keep change. }
            begin
              if iAction <> dsRecNew then
              begin
                with DeltaCurA do
                begin
                  Check(MoveToSeqNo(J + 1)); 
                  Check(GetCurrentRecordPtr(pRecBuf));
                  AssignFields(FRecords.List^[OldP.RecNo], pRecBuf, True, True);
                end
              end;
            end;
          dscbrCORRECT       :   { Overwrite change with new values. }
            begin
              if iAction <> dsRecDeleted then
                with DeltaCurA do
                begin
                  Check(MoveToSeqNo(J + 1)); // ord(iAction = dsRecModified) + /?
                  Check(GetCurrentRecordPtr(pRecBuf));
                  AssignFields(OldP.pRecBuf, pRecBuf, True, True);
                  ClearChanges(iAction, OldP.RecNo, nil, DeltaRecs, False);
                end;
            end;
          dscbrCANCEL        :   { Cancel change (remove from delta). }
            begin
              ClearChanges(iAction, OldP.RecNo, nil, DeltaRecs, True);
            end;
          dscbrREFRESH       :   { Update original record. Cancel change. }
            begin
              if iAction <> dsRecNew then
              begin
                with DeltaCurA do
                begin
                  Check(MoveToSeqNo(J + 1)); // ord(iAction = dsRecModified) +
                  Check(GetCurrentRecordPtr(pRecBuf));
                  MergeRecord(iAction, OldP.RecNo, pRecBuf, DeltaRecs);
                end
              end else begin
                ClearChanges(iAction, OldP.RecNo, nil, DeltaRecs, True);
              end;
            end;
        end;
        Inc(J);
      end;
      NotifyCursors(ceNotify, 0, 0, 0);
    finally
      EnableNotify;
    end;
  finally
    ListFreeMem(DeltaRecs, True);
  end;
end;


function TDSBase.Reconcile_MD(pDsRec: IDSBase; pDeltaPacket, pErrorPacket: PSafeArray;
  iClientData: DWord; pfReconcile_MD: pfDSReconcile_MD): DBResult;

  procedure CreateRCursor(Base: TDSBase);
  var
    I: Integer;
  begin
    with Base do
    begin
      CreateDSBaseCursor(Base as IDSBase, FRCursor);
      if Assigned(FParentDS) then
        FParentDS.FRCursor.LinkCursors(0, nil, nil, FRCursor);
      FRCursor.MoveToBOF;
      for I := 0 to High(FChilds) do
        CreateRCursor(BaseFromIDS(FChilds[I]));
    end;
  end;

  procedure FreeRCursor(Base: TDSBase);
  var
    I: Integer;
  begin
    with Base do
    begin
      FRCursor := nil;
      for I := 0 to High(FChilds) do
        FreeRCursor(BaseFromIDS(FChilds[I]));
    end;
  end;

  procedure FreeResponses(Base: TDSBase);
  var
    I: Integer;
  begin
    with Base do
    begin
      FreeAndNil(FResponses);
      for I := 0 to High(FChilds) do
        FreeResponses(BaseFromIDS(FChilds[I]));
    end;
  end;

var
  DeltaDS, ErrorDS: IDSBase;
  DeltaDSA, ErrorDSA: TDSBase;
  FieldNums: TList;           
begin
  try
    DeltaDS := CreateDSBase;
    Check(DeltaDS.AppendData(pDeltaPacket, True));
    DeltaDSA := BaseFromIDS(DeltaDS);

    ErrorDS := CreateDSBase;
    Check(ErrorDS.AppendData(pErrorPacket, True));
    ErrorDSA := BaseFromIDS(ErrorDS);

    Result := DBERR_NONE;

    try
      try
        CreateRCursor(DeltaDSA);
        CreateRCursor(ErrorDSA);
        FieldNums := TList.Create;
        try
          if not GetResponses(DeltaDSA, ErrorDSA,
            FieldNums, iClientData, pfReconcile_MD) then Exit;
        finally
          FieldNums.Free;
        end;
      finally
        FreeRCursor(DeltaDSA);
        FreeRCursor(ErrorDSA);
      end;

      ErrorDS := nil;
      MergeRecords(DeltaDSA);
    finally
      FreeResponses(Self);
    end;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.ReconcileCallback(iRslt: Integer; iUpdateKind: DSAttr;
  iResAction: dsCBRType; iErrCode: Integer; pErrMessage, pErrContext: PChar;
  pRecUpd, pRecOrg, pRecConflict: Pointer; iLevels: Integer;
  piFieldIDs: PInteger): dsCBRType; stdcall;
begin
  if Assigned(FpfReconcile) then
    Result := FpfReconcile(FRiClientData, iRslt, iUpdateKind, iResAction, iErrCode,
      pErrMessage, pErrContext, pRecUpd, pRecOrg, pRecConflict)
  else if iResAction <> dscbrSkip then
     Result := dscbrABORT else Result := dscbrSkip;
end;

function TDSBase.Reconcile(DeltaPacket, ErrorPacket: PSafeArray;
  iClientData: DWord; pfReconcile: pfDSReconcile): DBResult;
var
  DsBase: IDSBase;
begin
  try
    Check(Self.Clone(0, False, False, DSBase));
    FRiClientData := iClientData;
    FpfReconcile := pfReconcile;
    Result := Reconcile_MD(DSBase, DeltaPacket, ErrorPacket,
      DWord(Self), @TDSBase.ReconcileCallback);
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.Refresh(NewPacket: PSafeArray; iClientData: DWord; pfReconcile: pfDSReconcile): DBResult;
begin
  Result := RefreshRecords(NewPacket, 0, iClientData, pfReconcile);
end;

procedure TDSBase.AppendChildren(pRecBuf: PChar; Source: TDSBase);
var
  I, J, RecNo, NewRecNo: Integer;
  pRecTmp, pRecNew: PChar;
  Src, Dst: TDSBase;
begin
  RecNo := GetRecordRecNo(pRecBuf);
  CommitRecord(pRecBuf);
  for I := 0 to High(FChilds) do
  begin
    Dst := BaseFromIDS(FChilds[I]);
    Src := BaseFromIDS(Source.FChilds[I]);
    Dst.DisableNotify;
    try
      with Src, FRecords do
        for J := 0 to Count - 1 do
        begin
          pRecNew := List^[J];
          Dst.AllocRecord(pRecTmp);
          try
            Dst.CopyRecord(pRecNew, pRecTmp);
            Dst.PutFieldInternal(pRecTmp, Dst.FLinkFieldNo, @RecNo);
            Dst.InternalInsertRecord(pRecTmp, NewRecNo);
            pRecTmp := Dst.FRecords.List^[NewRecNo];
            Dst.AppendChildren(pRecTmp, Src);
          except
            Dst.FreeRecord(pRecTmp);
            raise;
          end;
          Dst.NotifyCursors(ceRecIns, NewRecNo, 0, 0);
        end;
    finally
      Dst.EnableNotify;
    end;
    Dst.NotifyCursors(ceRebuild, 0, 0, 0);
    Dst.NotifyCursors(ceNotify, 0, 0, 0);
  end;
end;

function TDSBase.RefreshRecords(NewDataPacket: PSafeArray; iRecNo,
  iClientData: DWord; pfReconcile: pfDSReconcile): DBResult;
var
  DSBase: IDSBase;
  DSBaseA: TDSBase;
  pRecOld: PChar;
begin
  try
    Check(Clone(1, True, False, DSBase));
    Check(DSBase.SetProp(dspropDSISPARTIAL, DWord(False)));
    Check(DSBase.AppendData(NewDataPacket, True));

    DSBaseA := BaseFromIDS(DSBase);
    if DSBaseA.FRecords.Count = 0 then
      HBError(DBERR_RECNOTFOUND);

    pRecOld := FRecords.List^[iRecNo - 1];
    AssignFields(pRecOld, DSBaseA.FRecords.List^[0], True);

    AppendChildren(pRecOld, DSBaseA);
    
    NotifyCursors(ceRebuild, 0, 0, 0);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.ReleaseBlobs(iBlobId: DWord): DBResult;
begin
  ListFreeMem(FBlobs, False);
  Result := DBERR_NONE;
end;

function TDSBase.RemoveIndex(pszName: PChar): DBResult;
var
  I: Integer;
  NewIndexes: DSIDXDescList;
begin
  I := FindIndex(pszName);
  if I >= 0 then
  try
    if Assigned(FCursor) and (StrIComp(CursorFromIDS(FCursor).FIndexDS.szName, pszName) = 0) then
      FCursor := nil;
    CheckInactiveIndex(pszName);
    SetLength(NewIndexes, Length(FIndexes) - 1);
    Move(FIndexes[0], NewIndexes[0], I * SizeOf(DSIDXDesc));
    if I < Length(FIndexes) - 1 then
      Move(FIndexes[I + 1], NewIndexes[I], (Length(FIndexes) - I - 1) * SizeOf(DSIDXDesc));
    FIndexes := NewIndexes;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end else
    Result := DBERR_NOSUCHINDEX;
end;

function TDSBase.Reserved1(Input: Pointer): DBResult;
begin
  Result := DBERR_NONE;
end;

function TDSBase.Reserved2(Input: Pointer): DBResult;
begin
  Result := DBERR_NONE;
end;

function TDSBase.Reserved3(Input: Pointer): DBResult;
begin
  Result := DBERR_NONE;
end;

function TDSBase.Reset: DBResult;
var
  I: Integer;
begin
  try
    FCursor := nil;
    FMDCursor := nil;
    FAutoIncCounter := 0;
    FRecNoCounter := 0;
    FDSIsPartial := False;

    for I := 0 to High(FChilds) do
      FChilds[I].Reset;

    with FRecords do
    begin
      for I := 0 to Count - 1 do
        FreeRecord(PChar(List^[I]));
      Clear;
    end;
    ClearChangeLog;
    NotifyCursors(ceRebuild, 0, 0, 0);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.RollBack(iSavePoint: SAVEPOINT): DBResult;
var
  RecNo: Integer;
begin
  if FLogChanges then
  begin
    if FChangesDS.Count > iSavePoint then
    begin
      DisableNotify;
      try
        while FChangesDS.Count > iSavePoint do
          UndoLastChange(RecNo, True);
      finally
        EnableNotify;
      end;
      NotifyCursors(ceNotify, 0, 0, 0);
    end;
    Result := DBERR_NONE;
  end else
    Result := DBERR_NOTHINGTOUNDO;
end;

function TDSBase.SetFieldCalculation(iClientData: DWord;
  pfCalc: pfDSCalcField): DBResult;
begin
  FCiClientData := iClientData;
  FCpfCalc      := pfCalc;
  Result := DBERR_NONE;
end;

function TDSBase.SetProp(eProp: DSProp; iPropValue: DWord): DBResult;
begin
  try
    case eProp of
      dspropLOGCHANGES:
        begin
          FLogChanges := Boolean(iPropValue);
          if not FLogChanges then AcceptChanges;
        end;
      dspropREADONLY:
        FReadOnly := Boolean(iPropValue);
      dspropNOOFCHANGES:
        HBError(DBERR_NOTSUPPORTED);
      dspropCONSTRAINTS_DISABLED:
        FConstrDisabled := Boolean(iPropValue);
      dspropDSISPARTIAL:
        FDSIsPartial := Boolean(iPropValue);
      dspropRECORDSINDS:
        HBError(DBERR_NOTSUPPORTED);
      dspropAUTOINC_DISABLED:
        FAutoIncDisabled := Boolean(iPropValue);
      dspropISDELTA:
        FIsDelta := Boolean(iPropValue);
      dspropDONTINCLMETADATA:
        FNoMetaData := Boolean(iPropValue);
      dspropINCLBLOBSINDELTA:
         FBlobsInDelta := iPropValue;
      dspropGETSAVEPOINT:
        HBError(DBERR_NOTSUPPORTED);
      dspropCOMPRESSARRAYS:
        ;
      dspropMD_SEMANTICS:
        FMDOptions := iPropValue;
      dspropFIELD_FULLNAME:
        HBError(DBERR_NIY);
      dspropFIELDID_FORNAME:
        HBError(DBERR_NIY);
      dspropFIELDID_FORPARENT:
        HBError(DBERR_NIY);
      dspropCHANGEINDEX_VIEW:
        HBError(DBERR_NIY);
      dspropGETUNIQUEINDEX:
        HBError(DBERR_NOTSUPPORTED);
      dspropREMOTE_UPDATEMODE:
        HBError(DBERR_NIY);
      dspropXML_STREAMMODE:
        FXMLStreamMode := iPropValue;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.BlankNotChanged(pRecBuf, pRecOrg: PChar; CheckOnly: Boolean): Boolean;
var
  I: Integer;
  IsEqual: Boolean;
  pFld1, pFld2: Pointer;
  Blob1, Blob2: TBlobData;
  bBlank1, bBlank2: Bool;
begin
  for I := 0 to High(FFields) do
    with FFields[I] do
    begin
      if iFldType in [fldADT, fldARRAY] then
        IsEqual := False
      else if iFldType = fldTABLE then
        IsEqual := CheckOnly
      else if iFldType <> fldBLOB then
      begin
        if (pRecBuf + iNullOffsInRec)^ = Char(NOT_BLANK) then
          pFld1 := pRecBuf + iFldOffsInRec else
          pFld1 := nil;

        if (pRecOrg + iNullOffsInRec)^ = Char(NOT_BLANK) then
          pFld2 := pRecOrg + iFldOffsInRec else
          pFld2 := nil;

        IsEqual := FldCmpEx(iFldType, pFld1, pFld2, iUnits1, iUnits1, False) = 0;
      end else begin
        CheckBlobFetched(pRecOrg, I + 1);
        GetField(pRecBuf, I + 1, @Blob1, bBlank1);
        GetField(pRecOrg, I + 1, @Blob2, bBlank2);
        IsEqual := (bBlank1 = bBlank2) and (Blob1.Length = Blob2.Length) and
          SysUtils.CompareMem(Blob1.Data, Blob2.Data, Blob1.Length);
      end;

      if CheckOnly then
      begin
        if not IsEqual then
        begin
          Result := False;
          Exit;
        end;
      end else
        if IsEqual then
          PutBlankValues(pRecBuf, I + 1, BLANK_NOTCHANGED);
    end;
  Result := True;
end;

procedure TDSBase.GetDeltaRecords(pRecBufM: PChar; Recs: TList);

  procedure AddDeltaRec(Attr: DSAttr; RecNo: DWord; pRec: PChar);
  var
    P: PDeltaRec;
  begin
    GetMem(P, SizeOf(TDeltaRec));
    try
      P.RecAttr := Attr;
      P.RecNo   := RecNo;
      P.pRecBuf := pRec;
      Recs.Add(P);
    except
      FreeMem(P);
      raise;
    end;
  end;

var
  I, J: Integer;
  RecNums: TList;
  pNewRec, pOldRec, pTmpRec: PChar;
  P: PChangeInfo;
  OldAttr, NewAttr: DSAttr;
  RecNo, RecNoM, OldSeqNo, NewSeqNo, TmpSeqNo: Integer;
  bBlank: Bool;
begin
  RecNums := TList.Create;
  try
    with FChanges do
      for I := 0 to Count - 1 do
      begin
        P := List^[I];
        // This is the first change entry for the record.
        OldSeqNo := P.OldSeqNo;
        pOldRec := FRecords.List^[OldSeqNo];

        if Assigned(pRecBufM) then
        begin
          GetField(pOldRec, FLinkFieldNo, @RecNoM, bBlank);
          if RecNoM <> GetRecordRecNo(pRecBufM) then Continue;
        end;

        RecNo := GetRecordRecNo(pOldRec);

        // Record is already processed
        if FindInteger(RecNo, RecNums.List^, RecNums.Count) >= 0 then Continue;

        OldAttr := P.RecAttr;

        NewAttr := OldAttr;
        NewSeqNo := OldSeqNo;
        pNewRec := pOldRec;

        // Search for last change entry
        if OldAttr <> dsRecDeleted then
        begin
          // We need to find the last one.
          J := Count - 1;
          while J >= I do
          begin
            P := List^[J];
            if Assigned(P) then
            begin
              TmpSeqNo := P.NewSeqNo;
              pTmpRec := FRecords.List^[TmpSeqNo];
              if (RecNo = GetRecordRecNo(pTmpRec)) then
              begin
                NewAttr := P.RecAttr;
                NewSeqNo := TmpSeqNo;
                pNewRec := pTmpRec;
                Break;
              end;
            end;
            Dec(J);
          end;
        end;

        // pOldRec - original record
        // pNewRec - latest record version

        // Skip inserted and deleted records
        if (OldAttr = dsRecNew) then
          if (NewAttr = dsRecDeleted) then
          begin
            RecNums.Add(Pointer(RecNo));
            Continue;
          end else
            NewAttr := dsRecNew;

        // For deletes and updates:
        // insert the original record version
        if NewAttr <> dsRecNew then
          AddDeltaRec(dsRecUnmodified or (NewAttr and dsRecDeleted), OldSeqNo, pOldRec);

        // For inserts and updates:
        // insert the new record version
        if NewAttr = dsRecNew then
          AddDeltaRec(NewAttr, NewSeqNo, pNewRec)
        else if NewAttr = dsRecModified then
        begin
          // If master record changed as a result of changing detail record then
          // Do not add new record version because all its fieds will keep old values
          //
          if not BlankNotChanged(pNewRec, pOldRec, True) then
            AddDeltaRec(NewAttr, NewSeqNo, pNewRec)
        end;

        RecNums.Add(Pointer(RecNo));
      end;

  finally
    RecNums.Free;
  end;
end;

procedure TDSBase.ClearChanges(Attr: DSAttr; RecNo: DWord; pRecOrg: PChar; DeltaRecs: TList; RollingBack: Boolean);
var
  I, SeqNo: Integer;
  pRecBuf: PChar;
  RecOrgNo: Integer;
  P: PChangeInfo;
begin
  if not Assigned(pRecOrg) then pRecOrg := FRecords.List^[RecNo];

  RecOrgNo := GetRecordRecNo(pRecOrg);
  with FChanges do
    for I := Count - 1 downto 0 do
    begin
      P := List^[I];
      SeqNo := P.NewSeqNo;
      pRecBuf := FRecords.List^[SeqNo];
      if ((Attr <> dsRecModified) or (pRecBuf <> pRecOrg)) and (GetRecordRecNo(pRecBuf) = RecOrgNo) then
      begin
        DelChange(P);
        DestroyChangeEntry(P, True, { RollingBack, }DeltaRecs);
        if (pRecBuf <> pRecOrg) then
          RemoveRecord(SeqNo, pRecBuf, DeltaRecs);
      end;
    end;

  if FIsDelta then
    SetRecordAttr(pRecOrg, GetRecordAttr(pRecOrg) and not dsRecModified2) else
    SetRecordAttr(pRecOrg, dsRecUnmodified);

  if Attr <> dsRecDeleted then
    NotifyCursors(ceRecIns, RecNo{SeqNoGetRecordRecNo(pRecOrg)}, 0, 0)
  else
    RemoveRecord(RecNo, pRecOrg, DeltaRecs);
end;

function TDSBase.GetRowRequestPacket(RecNo: DWord; IndexDS: pDSIDXDesc;
  bKeysOnly, bFetchAllBlobs, bFetchAllDetails, bInclMetaData: Bool;
  out Packet: PSafeArray): DBResult;
var
  I, NewRecNo: Integer;
  DSBase: IDSBase;
  DSBaseA: TDSBase;
  Recs: DWord;
  pRecBuf, pRecNew: PChar;
begin
  try
    Check(Clone(1, True, True, DSBase));
    DSBaseA := BaseFromIDS(DSBase);

    Check(DSBase.SetProp(dspropLOGCHANGES, DWord(False)));
    // Ignoring bInclMetaData cause TPacketDataSet requires MetaData always.
    // Check(DSBase.SetProp(dspropDONTINCLMETADATA, DWord(bInclMetaData)));

    DSBaseA.AllocRecord(pRecNew);
    try
      GetRecordOrg(RecNo, RecNo, nil);
      pRecBuf := FRecords.List^[RecNo];

      for I := 0 to High(FFields) do
        if (not (bKeysOnly and Assigned(IndexDS)) or IsKeyField(IndexDS^, I + 1)) then
        begin
          CopyField(FFields[I], FFields[I], pRecBuf, pRecNew);
          if bFetchAllBlobs and (FFields[I].iFldType = fldBLOB) then
            DSBaseA.PutBlobInternal(pRecNew, I + 1, 0, nil, 0)
          else if bFetchAllDetails and (FFields[I].iFldType = fldTABLE) then
          begin
            Recs := 0;
            DSBaseA.PutFieldInternal(pRecNew, I + 1, @Recs);
            DSBaseA.PutBlankInternal(pRecNew, I + 1, BLANK_NULL);
          end;
        end;
      DSBaseA.InsertRecord(pRecNew, NewRecNo, False);
    finally
      DSBaseA.FreeRecord(pRecNew);
    end;

    Result := DSBase.StreamDS(Packet);
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.MergeRecord(Attr: DSAttr; RecNo: DWord; pNewRec: PChar; DeltaRecs: TList);
var
  pOldRec: PChar;
  NewRecNo: DWord;
begin
  DisableNotify;
  with FRecords do
  try
    pOldRec := List^[RecNo];

    { Swapping the old and new records to keep the last record buffer valid }
    if Assigned(pNewRec) then
    begin
      NewRecNo := FindInteger(Integer(pNewRec), List^, Count);
      List^[RecNo] := pNewRec;
      List^[NewRecNo] := pOldRec;
      pOldRec := pNewRec;
    end;

    ClearChanges(Attr, RecNo, pOldRec, DeltaRecs, False);
  finally
    EnableNotify;
  end;
end;

procedure TDSBase.InitDelta(Delta: TDSBase; pRecBufM, pRecBufM2: PChar; BlobsInDelta: DWord);

  procedure ReleaseBlobs(pRecBuf: PChar);
  var
    I: Integer;
    FldBuf: TBlobData;
    bBlank: Bool;
  begin
    for I := 0 to High(Delta.FFields) do
      with FFields[I] do
        if iFldType = fldBLOB then
        begin
          Delta.GetField(pRecBuf, I + 1, @FldBuf, bBlank);
          if FldBuf.Length > BlobsInDelta then
            Delta.PutBlobInternal(pRecBuf, I + 1, 0, nil, 0);
        end;
    end;

var
  I, J, NewRecNo: Integer;
  P: PDeltaRec;
  Recs: TList;
  pRecBuf, pRecNew: PChar;
  Child, ChildD: TDSBase;
begin
  Delta.FIsDelta := True;
  Delta.SetProp(dspropLOGCHANGES, DWord(False));

  Recs := TList.Create;
  try
    GetDeltaRecords(pRecBufM, Recs);

    pRecNew := nil;

    GetMem(pRecBuf, FRecBufSize);
    try
      for I := 0 to Recs.Count - 1 do
      begin
        P := Recs.List^[I];
        System.Move(P.pRecBuf^, pRecBuf^, FRecBufSize);
        SetRecordAttr(pRecBuf, P.RecAttr);

        // pRecNew initialized in the previous cycle
        if P.RecAttr = dsRecModified then
          BlankNotChanged(pRecBuf, pRecNew, False);

        Check(Delta.InsertRecord(pRecBuf, NewRecNo));
        pRecNew := Delta.FRecords.List^[NewRecNo];

        if Assigned(pRecBufM2) then
        begin
          NewRecNo := GetRecordRecNo(pRecBufM2);
          Delta.PutFieldInternal(pRecNew, Delta.FLinkFieldNo, @NewRecNo);
        end;

        if (P.RecAttr = dsRecModified) and (BlobsInDelta > 0) then
          ReleaseBlobs(pRecNew);

        if P.RecAttr <> dsRecModified then
          for J := 0 to High(FChilds) do
          begin
            Child := BaseFromIDS(FChilds[J]);
            ChildD := BaseFromIDS(Delta.FChilds[J]);
            Child.InitDelta(ChildD, pRecBuf, pRecNew, BlobsInDelta);
          end;
      end;
    finally
      FreeMem(pRecBuf);
    end;
  finally
    ListFreeMem(Recs, True);
  end;
  Delta.SetProp(dspropLOGCHANGES, DWord(True));
end;

function TDSBase.GetDelta(out DsDelta: IDSBase): DBResult;
begin
  try
    if FChangesDS.Count = 0 then
      HBError(DBERR_DELTAISEMPTY);
    Check(Clone(1, True, True, DsDelta));
    InitDelta(BaseFromIDS(DsDelta), nil, nil, FBlobsInDelta);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.WriteColumn(const Writer: IDSWriter; var DescNo: Integer; DoWrite: Boolean);
var
  I: Integer;
  FldDes: pDSFldDesc;
  FldDesc: TDSDataPacketFldDesc;
  Child: TDSBase;
  AttrInfos: TAttrInfos;
begin
  FldDes := @FFields[DescNo];
  Inc(DescNo);

  // Skip link attribute
  if (FldDes.iFldAttr and fldAttrLink) <> 0 then
      Exit;

  if DoWrite then
  begin
    FldDescToPacketDS(FldDes^, DescNo, FldDesc, AttrInfos);
    Check(Writer.AddColumnDesc(FldDesc));

    for I := High(AttrInfos) downto 0 do
    begin
      if FindOptAttribute(FAttrInfos, AttrInfos[I].iFieldNo, AttrInfos[I].Attr) >= 0 then
        DeleteAttrInfo(AttrInfos, I);
    end;

    if Assigned(AttrInfos) then
      WriteAttributes(Writer, AttrInfos, DescNo);

    WriteAttributes(Writer, FAttrInfos, DescNo);
  end;

  with FldDes^ do
    case iFldType of
      fldADT:
        for I := 1 to iUnits1 do
          WriteColumn(Writer, DescNo, DoWrite);
      fldARRAY:
        for I := 1 to iUnits1 do
          WriteColumn(Writer, DescNo, (I = 1) and DoWrite);
      fldTABLE:
        if DoWrite then
        begin
          Child := GetEmbeddedDSBase(DescNo);
          Child.FXMLStreamMode := FXMLStreamMode;
          Child.WriteMetaData(Writer, False);
        end;
    end;
end;

procedure TDSBase.WriteMetaData(const Writer: IDSWriter; Root: Boolean);
var
  I: Integer;
begin
  I := 0;
  while I <= High(FFields) do
    WriteColumn(Writer, I, True);

  if FIsDelta then
    Check(Writer.AddAttribute(pcktAttrArea, szDATASET_DELTA, (dsfldBOOL shl dsSizeBitsLen) or 1, SizeOf(FIsDelta), @FIsDelta));

  if FStreamedReadOnly then
    Check(Writer.AddAttribute(pcktAttrArea, szREADONLY, (dsfldBOOL shl dsSizeBitsLen) or 1, SizeOf(FStreamedReadOnly), @FStreamedReadOnly));

  if FAutoIncCounter <> 0 then
    Check(Writer.AddAttribute(pcktAttrArea, szAUTOINCVALUE, (dsfldINT shl dsSizeBitsLen) or 4, SizeOf(FAutoIncCounter), @FAutoIncCounter));

  if FConstrDisabled then
    Check(Writer.AddAttribute(pcktAttrArea, szCONSTRAINTS, (dsfldBOOL shl dsSizeBitsLen) or 1, SizeOf(FConstrDisabled), @FConstrDisabled));

  if FMDOptions <> 0 then
    Check(Writer.AddAttribute(pcktAttrArea, szMD_SEMANTICS, (dsfldINT shl dsSizeBitsLen) or 4, SizeOf(FMDOptions), @FMDOptions));

  if FLCID <> 0 then
    Check(Writer.AddAttribute(pcktAttrArea, szLCID, (dsfldINT shl dsSizeBitsLen) or 4, SizeOf(FLCID), @FLCID));

  if FXMLStreamMode = 0 then
    WriteAttributes(Writer, FAttrInfos, 0);

  WriteChangeLog(Writer);

  if not Root then
    Check(Writer.AddAttribute(pcktAttrArea, nil, 0, 0, nil));
end;

procedure TDSBase.WriteField(const Writer: IDSWriter; pRecBuf: PChar; FieldNo: DWord);
const
  D: Integer = 0;
var
  pFldBuf: PChar;
  FldStat: TPcktFldStatus;
  BlobFld: TBlobData;
  Recs: DWord;
  bBlank: Bool;
  Child: TDSBase;
begin
  with FFields[FieldNo - 1] do
  begin
    // Skip link attribute
    if (iFldAttr and fldAttrLink) <> 0 then
      Exit;

    FldStat := FldStatusMap[(pRecBuf + iNullOffsInRec)^];

    if FldStat <> fldIsChanged then
      Check(Writer.PutField(FldStat, 0, nil))
    else begin
      pFldBuf := pRecBuf + iFldOffsInRec;
      case iFldType of
        fldZSTRING:
          begin
            Check(Writer.PutField(FldStat, StrLen(pFldBuf) + 1, pFldBuf));
          end;
        fldUNICODE:
          begin
            Check(Writer.PutField(FldStat, PWord(pFldBuf)^, PChar(pFldBuf) + 2));
          end;
        fldADT, fldARRAY:
          begin
            if ((pRecBuf + iNullOffsInRec)^ = Char(BLANK_NULL)) then
              Check(Writer.PutField(fldIsNull, 0, nil)) else
              Check(Writer.PutField(fldIsChanged, 0, nil));
          end;
        fldBLOB:
          begin
            Check(GetField(pRecBuf, FieldNo, @BlobFld, bBlank));
            if (BlobFld.Length and dsDELAYEDBIT) = dsDELAYEDBIT then
              Check(Writer.PutField(FldStat, BlobFld.Length, @D)) else
              Check(Writer.PutField(FldStat, BlobFld.Length, BlobFld.Data));
          end;
        fldTABLE:
          begin
            Check(GetField(pRecBuf, FieldNo, @Recs, bBlank));
            if bBlank then
              Check(Writer.PutField(fldIsNull, 0, nil))
            else begin
              Check(Writer.PutField(fldIsChanged, SizeOf(Recs), @Recs));
              if ((Recs = DWord(-1)) or ((Recs and dsDELAYEDBIT) = 0)) then
              begin
                Child := GetEmbeddedDSBase(FieldNo);
                Child.FXMLStreamMode := FXMLStreamMode;
                Child.WriteDataSet(Writer, pRecBuf);
                Check(Writer.EndOfNestedRows);
              end;
            end;
          end;
      else
        Check(Writer.PutField(FldStat, iFldLen, pFldBuf));
      end;
    end;
  end;
end;

procedure TDSBase.WriteRecord(const Writer: IDSWriter; pRecBuf: PChar);
var
  I: Integer;
  Attr: DSAttr;
begin
  Attr := GetRecordAttr(pRecBuf);
  Writer.PutField(fldIsChanged, SizeOf(Attr), @Attr);
  for I := 0 to High(FFields) do
    WriteField(Writer, pRecBuf, I + 1);
end;

procedure TDSBase.WriteChangeLog(const Writer: IDSWriter);
var
  I: Integer;
  P: PChangeInfo;
  Stream: TMemoryStream;
begin
  if FChanges.Count = 0 then Exit;
  Stream := TMemoryStream.Create;
  try
    with FChanges do
      for I := 0 to Count - 1 do
      begin
        P := List^[I];
        // Skip records that inserted by children
        if not ((P^.RecAttr = dsRecModified) and (P^.OldSeqNo = P^.NewSeqNo)) then
          Stream.WriteBuffer(P^, SizeOf(TChangeInfo));
      end;
    if Stream.Size > 0 then
      Check(Writer.AddAttribute(pcktAttrArea, szCHANGE_LOG, (dsfldBYTES shl dsSizeBitsLen) or dsVaryingFldType or SizeOf(Byte), Stream.Size, Stream.Memory));
  finally
    Stream.Free;
  end;
end;

procedure TDSBase.WriteDataSet(const Writer: IDSWriter; pRecBufM: PChar);

  procedure GetChildRecords(Recs: TList);
  var
    I, RecNo, Offs: Integer;
    pRecBuf: PChar;
  begin
    RecNo := GetRecordRecNo(pRecBufM);
    Offs := FFields[FLinkFieldNo - 1].iFldOffsInRec;
    with FRecords do
      for I := 0 to Count - 1 do
      begin
        pRecBuf := List^[I];
        if PInteger(pRecBuf + Offs)^ = RecNo then
          with Recs do
            if FindInteger(Integer(pRecBuf), List^, Count) < 0 then
              Add(pRecBuf);
      end;
  end;

  procedure WriteRecords(Records: TList);
  var
    I: Integer;
    pRecBuf: PChar;
  begin
    with Records do
      for I := 0 to Count - 1 do
      begin
        pRecBuf := List^[I];
        if Assigned(FRecsWritten) then
        begin
          if (FindInteger(Integer(pRecBuf), FRecsWritten.List^, Count) < 0) then
            FRecsWritten.Add(pRecBuf)
          else begin
            GetTickCount;
            Continue;
          end;
        end;
        WriteRecord(Writer, pRecBuf);
      end;
  end;

var
  Recs: TList;
begin
  if Assigned(pRecBufM) then
  begin
    if not Assigned(FRecsWritten) then
    begin
      FRecsWritten := TList.Create;
      FRecsWritten.Capacity := FRecords.Count;
    end;

    Recs := TList.Create;
    try
      GetChildRecords(Recs);
      WriteRecords(Recs);
    finally
      Recs.Free;
    end;
  end else
    WriteRecords(FRecords);
end;

function TDSBase.StreamDS(out Packet: PSafeArray): DBResult;

 procedure DoRecsWrittenFree(BaseDS: TDSBase);
 var
   I: Integer;
 begin
   with BaseDS do
   begin
     FreeAndNil(FRecsWritten);
     for I := 0 to High(FChilds) do
       DoRecsWrittenFree(BaseFromIDS(FChilds[I]));
   end;
 end;

var
  Writer: IDSWriter5;
begin
  try
    Writer := TDSWriter.Create as IDSWriter5;
    Check(Writer.SetXMLMode(FXMLStreamMode));
    Check(Writer.Init_Sequential(PACKETVERSION_1, Length(FFields)));

    if (FXMLStreamMode <> 0) or not FNoMetaData then
      WriteMetaData(Writer, True);

    try
      WriteDataSet(Writer, nil);
    finally
      DoRecsWrittenFree(Self);
    end;

    Check(Writer.GetDataPacket(Packet));
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.AppendData(Packet: PSafeArray; bEof: Bool): DBResult;

  procedure DoWriterStart(DataSet: TDSBase);
  var
    I: Integer;
  begin
    with DataSet do
    begin
      WriterStart;
      for I := 0 to High(FChilds) do
        DoWriterStart(BaseFromIDS(FChilds[I]));
    end;
  end;

  procedure DoWriterApply(DataSet: TDSBase);
  var
    I: Integer;
  begin
    with DataSet do
    begin
      // Initialize memory stream
      FWFldDatas := Self.FWFldDatas;
      WriterApply;
      for I := 0 to High(FChilds) do
        DoWriterApply(BaseFromIDS(FChilds[I]));
    end;
  end;


  procedure DoWriterFree(DataSet: TDSBase);
  var
    I: Integer;
  begin
    with DataSet do
    begin
      WriterFree;
      for I := 0 to High(FChilds) do
        DoWriterFree(BaseFromIDS(FChilds[I]));
    end;
  end;

var
  Reader: IDSReader;
begin
  try
    Reader := TDSReader.Create as IDSReader;
    try
      FWFldDatas := TMemoryStream.Create;
      DoWriterStart(Self);
      Check(Reader.ReadDataPacket(Self as IDSWriter, Packet));
      DoWriterApply(Self);
      FDSIsPartial := not bEof;
      Result := DBERR_NONE;
    finally
      FreeAndNil(FWFldDatas);
      DoWriterFree(Self);
    end;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.AddChange(Change: PChangeInfo);
begin
  AddChangeDS(Self, Change);
  FChanges.Add(Change);
end;

procedure TDSBase.DelChange(Change: PChangeInfo);
begin
  DelChangeDS(Self, Change);
  FChanges.Remove(Change);
end;

function TDSBase.HasChangeEntry(RecNo: Integer): Boolean;
var
  I: Integer;
  P: PChangeInfo;
begin
  with FChangesDS do
  begin
    for I := 0 to Count - 1 do
      if List^[I] = Self then
      begin
        P := FChangesVL.List^[I];
        if (P.NewSeqNo = RecNo) then
        begin
          Result := True;
          Exit;
        end;
      end;
   end;
  Result := False;
end;

procedure TDSBase.AddChangeDS(DataSet: TDSBase; Change: PChangeInfo);
begin
  if (DataSet = Self) and not HasChangeEntry(Change.NewSeqNo) then
    Inc(FChangeCount);

  FChangesDS.Add(DataSet);
  FChangesVL.Add(Change);

  if Assigned(FParentDS) then
    FParentDS.AddChangeDS(DataSet, Change);
end;

procedure TDSBase.DelChangeDS(DataSet: TDSBase; Change: PChangeInfo);
var
  I: Integer;
begin
  with FChangesVL do
    I := FindInteger(Integer(Change), List^, Count);
  FChangesDS.Delete(I);  // DataSet
  FChangesVL.Delete(I);  // Change
  if (DataSet = Self) and not HasChangeEntry(Change.NewSeqNo) then
    Dec(FChangeCount);
  if Assigned(FParentDS) then
    FParentDS.DelChangeDS(DataSet, Change);
end;

procedure TDSBase.LogChange(RecAttr: DSAttr; OldSeqNo, NewSeqNo: Integer);
var
  Change: PChangeInfo;
  ParentSeqNo: Integer;
  pRecBuf: PChar;
begin
  if FLogChanges then
  begin
    ParentSeqNo := 0;
    if Assigned(FParentDS) then
    begin
      pRecBuf := GetRecordPtr(NewSeqNo);
      ParentSeqNo := GetParentSeqNo(pRecBuf);
      FParentDS.LogChange(dsRecModified, ParentSeqNo, ParentSeqNo);
    end;

    GetMem(Change, SizeOf(TChangeInfo));
    try
      ZeroMem(Change, SizeOf(TChangeInfo));
      Change^.RecAttr := RecAttr and dsIsModified;
      Change^.OldSeqNo := OldSeqNo;
      Change^.NewSeqNo := NewSeqNo;
      Change^.ParentSeqNo := ParentSeqNo;
      AddChange(Change);
    except
      FreeMem(Change);
      raise;
    end;
  end;
end;

procedure TDSBase.ClearChangeLog;
var
  I: Integer;
  P: PChangeInfo;
begin
  for I := FChanges.Count - 1 downto 0 do
  begin
    P := FChanges.List^[I];
    DelChange(P);
    FreeMem(P);
  end;
end;

function TDSBase.FindChangeEntry(var Start: Integer; NewSeqNo: Integer): PChangeInfo;
begin
  while Start >= 0 do
  begin
    Result := FChanges.List^[Start];
    if (NewSeqNo < 0) or (PChangeInfo(Result)^.NewSeqNo = NewSeqNo) then Exit;
    Dec(Start);
  end;
  Result := nil;
end;

procedure TDSBase.DestroyChangeEntry(var Change: PChangeInfo; Decrement: Boolean; DeltaRecs: TList);

  procedure DecrementChanges(RecNo: Integer);
  var
    I: Integer;
    P: PChangeInfo;
    D: PDeltaRec;
  begin
    with FChanges do
      for I := 0 to Count - 1 do
      begin
        P := List^[I];
        with P^ do
        begin
          if (OldSeqNo > RecNo) then Dec(OldSeqNo);
          if (NewSeqNo > RecNo) then Dec(NewSeqNo);
          if (ParentSeqNo > RecNo) then Dec(ParentSeqNo);
        end;
      end;

    if Assigned(DeltaRecs) then
      with DeltaRecs do
        for I := 0 to Count - 1 do
        begin
          D := List^[I];
          if (D^.RecNo > DWord(RecNo)) then Dec(D^.RecNo);
        end;
  end;

begin
  ClearParentChangeEntry(Change.ParentSeqNo);
  if Decrement and not ((Change.NewSeqNo = Change.OldSeqNo) and (Change.RecAttr = dsRecModified)) then
    DecrementChanges(Change.NewSeqNo);
  FreeMem(Change);
end;

procedure TDSBase.ClearParentChangeEntry(ParentSeqNo: Integer);

  function FindParentEntry(ParentSeqNo: Integer): PChangeInfo;
  var
    I: Integer;
  begin
    with FParentDS.FChanges do
      for I := Count - 1 downto 0 do
      begin
        Result := List^[I];
        if Result^.NewSeqNo = ParentSeqNo then Exit;
      end;
    Result := nil;
  end;

var
  P: PChangeInfo;
  Start: Integer;
begin
  if Assigned(FParentDS) then
  begin
    Start := FParentDS.FChanges.Count - 1;
    P := FParentDS.FindChangeEntry(Start, ParentSeqNo);
    if Assigned(P) then
    begin
      FParentDS.DelChange(P);
      FParentDS.DestroyChangeEntry(P, True);
    end;
  end;
end;

procedure TDSBase.UndoChangeEntry(Change: PChangeInfo; var RecNo: Integer; Commited: Boolean);
var
  OldRec, NewRec: PChar;
begin
  if Commited then
    with Change^ do
    begin
      RecNo := OldSeqNo;
      case RecAttr of
        dsRecNew:
          RemoveRecord(NewSeqNo);

        dsRecModified:
          begin
            if NewSeqNo = OldSeqNo then
            begin
              ;
            end else
            begin
              NewRec := FRecords.List^[NewSeqNo];
              OldRec := FRecords.List^[OldSeqNo];
              SetRecordAttr(OldRec, GetRecordAttr(OldRec) and not dsRecOrg);
              SetRecordAttr(NewRec, dsUnused);
              NotifyCursors(ceRecMod, NewSeqNo, OldSeqNo, -1);
              FRecords.Delete(NewSeqNo);
              FreeRecord(NewRec);
            end;
          end;

        dsRecDeleted:
          begin
            OldRec := FRecords.List^[OldSeqNo];
            SetRecordAttr(OldRec, GetRecordAttr(OldRec) and not dsRecDeleted);
            NotifyCursors(ceRecIns, OldSeqNo, 0, 0);
          end;
      end;
    end;
  DestroyChangeEntry(Change, Change.RecAttr <> dsRecDeleted);
end;

function TDSBase.InternalUndoLastChange(var RecNo: Integer; Commited: Boolean): DBResult;
var
  I: Integer;
  P: PChangeInfo;
begin
  I := FChanges.Count - 1;

  P := FindChangeEntry(I, -1);
  if Assigned(P) then
  begin
    DelChange(P);
    UndoChangeEntry(P, RecNo, Commited);
    Result := DBERR_NONE;
  end else
    Result := DBERR_NOTHINGTOUNDO;
end;

function TDSBase.UndoLastChange(var RecNo: Integer; Commited: Boolean): DBResult;
var
  DataSet: TDSBase;
begin
  if FChangesDS.Count > 0 then
  begin
    with FChangesDS do
      DataSet := List^[Count - 1];
    Result := DataSet.InternalUndoLastChange(RecNo, Commited);
  end else
    Result := DBERR_NOTHINGTOUNDO;
end;

procedure TDSBase.InsertCursor(ACursor: TDSCursor);
begin
  FCursors.Add(ACursor);
  ACursor.FDSBase := Self;
end;

procedure TDSBase.RemoveCursor(ACursor: TDSCursor);
begin
  FCursors.Remove(ACursor);
  ACursor.FDSBase := nil;
end;

procedure TDSBase.DisableNotify;
begin
  Inc(FNotifyDisableCount);
end;

procedure TDSBase.EnableNotify;
begin
  Dec(FNotifyDisableCount);
end;

procedure TDSBase.NotifyCursors(Event: Integer; Data1, Data2, Data3: Integer);
var
  I: Integer;
begin
  for I := 0 to FCursors.Count - 1 do
    TDSCursor(FCursors.List^[I]).Notify(Event, Data1, Data2, Data3);
end;

function TDSBase.InitRecord(pRecBuf: PChar): DBResult;
var
  I: Integer;
begin
  ZeroMem(pRecBuf, FRecBufSize);
  try
    for I := 0 to High(FFields) do
    begin
      PutBlankInternal(pRecBuf, I + 1, BLANK_NOTCHANGED);
      with FFields[I] do
        case iFldType of
          fldBCD:
            SetZeroBcd(pFMTBcd(pRecBuf + iFldLen)^);
        end;
    end;
    SetRecordAttr(pRecBuf, dsRecUnmodified);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.AllocRecord(var P: PChar);
begin
  GetMem(P, FRecBufSize);
  try
    InitRecord(P);
  except
    FreeRecord(P);
    raise;
  end;
end;

procedure TDSBase.FreeRecord(var P: PChar);
var
  I: Integer;
  FldBuf: TBlobData;
  bBlank: Bool;
begin
  if Assigned(P) then
  begin
    for I := 0 to High(FFields) do
      if FFields[I].iFldType = fldBLOB then
      begin
        GetField(P, I + 1, @FldBuf, bBlank);
        FreeMem(FldBuf.Data);
      end;
    ReallocMem(P, 0);
  end;
end;

procedure TDSBase.RemoveRecord(RecNo: DWord; pRecBuf: PChar; DeltaRecs: TList);
var
  I: Integer;
  P: PDeltaRec;
begin
  with FRecords do
  begin
    if not Assigned(pRecBuf) then pRecBuf := List^[RecNo];
    SetRecordAttr(pRecBuf, dsUnused);
    NotifyCursors(ceRecDel, RecNo, 0, -1);
    Delete(RecNo);
    FreeRecord(pRecBuf);

    if Assigned(DeltaRecs) then
      with DeltaRecs do
        for I := 0 to Count - 1 do
        begin
          P := List^[I];
          if P.RecNo > RecNo then Dec(P.RecNo);
        end;
  end;
end;

procedure TDSBase.AssignFields(pRecDst, pRecSrc: PChar; ChangesOnly: Boolean; SkipLinks: Boolean);
var
  I: Integer;
begin
  for I := 0 to High(FFields) - ord(SkipLinks and Assigned(FParentDS)) do
    with FFields[I] do
      if not ChangesOnly or ((pRecSrc + iNullOffsInRec)^ <> Char(BLANK_NOTCHANGED)) then
        CopyField(FFields[I], FFields[I], pRecSrc, pRecDst);
end;

procedure TDSBase.CopyRecord(Source, Dest: PChar);
begin
  Move(Source^, Dest^, Reserved);
  AssignFields(Dest, Source);
end;

function TDSBase.GetRecord(RecNo: DWord; pRecBuf: PChar): DBResult;
begin
  try
    if Assigned(pRecBuf) then
      Move(GetRecordPtr(RecNo)^, pRecBuf^, FRecBufSize);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetRecordPtr(RecNo: Integer): PChar;
begin
  with FRecords do
  begin
    if (RecNo < 0) or (RecNo >= Count) then
      HBError(DBERR_RECNOTFOUND);
    Result := FRecords.List^[RecNo]
  end;
end;

procedure TDSBase.GetRecordOrg(RecNo: DWord; var OrgRecNo: DWord; pRecBuf: PChar = nil);
var
  I: Integer;
  RecCounter: Integer;
begin
  RecCounter := GetRecordRecNo(FRecords.List^[RecNo]);
  OrgRecNo := RecNo;
  for I := 0 to RecNo - 1 do
    if GetRecordRecNo(FRecords.List^[I]) = RecCounter then
    begin
      OrgRecNo := I;
      GetRecord(I, pRecBuf);
    end;
end;

function TDSBase.GetField(pRecBuf: Pointer; iFieldNo: DWord;
  pFldBuf: Pointer; var bBlank: Bool): DBResult;
begin
  with FFields[iFieldNo - 1] do
  begin
    bBlank := Bool((PChar(pRecBuf) + iNullOffsInRec)^);
    if Assigned(pFldBuf) then
      Move((PChar(pRecBuf) + iFldOffsInRec)^, pFldBuf^, iFldLen);
    Result := DBERR_NONE;
  end;
end;

procedure TDSBase.PutFieldBuff(pRecBuf: Pointer; iFieldNo: DWord; pFldBuf: Pointer);
var
  MinLen: Integer;
begin
  if Assigned(pFldBuf) then
    with FFields[iFieldNo - 1] do
      case iFldType of
        fldZSTRING:
          begin
            MinLen := MinInteger(StrLen(PChar(pFldBuf)), iUnits1);
            Inc(PChar(pRecBuf), iFldOffsInRec);
            Move(pFldBuf^, pRecBuf^, MinLen);
            (PChar(pRecBuf) + MinLen)^ := #0;
          end;
        fldUNICODE:
          begin
            MinLen := MinInteger(PWord(pFldBuf)^ div 2, iUnits1);
            Inc(PChar(pRecBuf), iFldOffsInRec);
            MinLen := MinLen * 2;
            Move(pFldBuf^, pRecBuf^, MinLen + 2);
          end;
      else
        Move(pFldBuf^, (PChar(pRecBuf) + iFldOffsInRec)^, iFldLen);
      end;
end;

function TDSBase.PutFieldInternal(pRecBuf: Pointer; iFieldNo: DWord;
  pFldBuf: Pointer): DBResult;
begin
  with FFields[iFieldNo - 1] do
  try
    PutFieldBuff(pRecBuf, iFieldNo, pFldBuf);

    if Assigned(pFldBuf) then
      PutBlankValues(pRecBuf, iFieldNo, NOT_BLANK) else
      PutBlankValues(pRecBuf, iFieldNo, BLANK_NULL);

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.PutField(pRecBuf: Pointer; iFieldNo: DWord; pFldBuf: Pointer): DBResult;
begin
  with FFields[iFieldNo - 1] do
  try
    if iFldType = fldBLOB then
      HBError(DBERR_INVALIDMODIFYREQUEST);

    if ((iFldAttr and fldAttrREADONLY) = fldAttrREADONLY) and (iFldSubType <> fldstAutoInc) then
      HBError(DBERR_READONLYFLD);

    Result := PutFieldInternal(pRecBuf, iFieldNo, pFldBuf);
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.CheckNotReadOnly;
begin
  if FReadOnly or FStreamedReadOnly then HBError(DBERR_TABLEREADONLY);
end;

function TDSBase.RevertRecord(RecNo: Integer): DBResult;
var
  I: Integer;
  P: PChangeInfo;
  Rec: PChar;
begin
  I := FChanges.Count - 1;
  try
    if I < 0 then
      HBError(DBERR_NOTHINGTOUNDO);

    Result := DBERR_NONE;

    DisableNotify;
    try
      while (I >= 0) and (RecNo >= 0) do
      begin
        Rec := FRecords.List^[RecNo];
        if (GetRecordAttr(Rec) and dsRecOrg) = dsRecOrg then Exit;
        P := FindChangeEntry(I, RecNo);
        if Assigned(P) then
        begin
          DelChange(P);
          UndoChangeEntry(P, RecNo, True);
        end;
      end;
    finally
      EnableNotify;
    end;
    NotifyCursors(ceNotify, 0, 0, 0);
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetMaxAutoInc(FieldNo: DWord): Integer;
var
  Desc: Bool;
  pRecBuf: PChar;
  bBlank: Bool;
begin
  if not Assigned(FAICursor) then
  begin
    CreateDSBaseCursor(Self as IDSBase, FAICursor);
    Desc := True;
    FAICursor.SortOnFields(1, @FieldNo, @Desc, nil);
  end;
  FAICursor.MoveToBOF;
  if FAICursor.MoveRelative(1) = DBERR_NONE then
  begin
    CursorFromIDS(FAICursor).GetCurrentRecordPtr(pRecBuf);
    GetField(pRecBuf, FieldNo, @Result, bBlank);
  end else
    Result := FAutoIncCounter;
end;

procedure TDSBase.CheckRequired(pRecBuf: PChar; AutoInc: Boolean);
var
  I: Integer;
  bBlank: Bool;
begin
  for I := 0 to High(FFields) do
    with FFields[I] do
    begin
      if ((pRecBuf + iNullOffsInRec) = char(BLANK_NULL)) or
         ((pRecBuf + iNullOffsInRec) = char(BLANK_NOTCHANGED)) then
      begin
        if AutoInc and (iFldSubType = fldstAUTOINC) then
        begin
          GetField(pRecBuf, I + 1, nil, bBlank);
          if bBlank then
          begin
            FAutoIncCounter := GetMaxAutoInc(I + 1);
            Inc(FAutoIncCounter);
            PutFieldInternal(pRecBuf, I + 1, @FAutoIncCounter);
          end;
        end else if Boolean(iFldAttr and fldAttrREQUIRED) then
          HBError(DBERR_REQDERR);
      end;
    end;
end;

procedure TDSBase.CheckConstraints(pRecBuf, ConstrType: PChar; Defaults: Boolean);
var
  ExprEval: TDSExprEvaluator;

  procedure DoCheckDefault(iFldNum: Integer);
  var
    Tmp: Variant;
    iFldType, iFldLen: Integer;
    ValPtr: Pointer;
    W: WideString;
    Dest: string;
  begin
    try
      Tmp := ExprEval.Evaluate(iFldType, iFldLen);
      if VarIsNull(Tmp) then Exit;
      ValPtr := GetValuePtr(Tmp, iFldType);
      if FFields[iFldNum - 1].iFldType = fldUNICODE then
      begin
        W := Tmp;
        SetLength(Dest, Length(W) * 2);
        Word(Pointer(Dest)^) := Length(W) * 2;
        Move(PWideChar(W)^, (PWideChar(Dest) + 1)^, Word(Pointer(Dest)^));
        PutFieldInternal(pRecBuf, iFldNum, Pointer(Dest));
      end else
        PutFieldInternal(pRecBuf, iFldNum, ValPtr);
    except
      on E: ENotSupported do ;
    end;
  end;

  procedure DoCheckCondition(ErrorMsg: PChar);

    procedure Error;
    begin
      FConstrErrorMsg := StrPas(ErrorMsg);
      HBError(DBERR_CONSTRAINTFAILED);
    end;

  var
    Tmp: Variant;
    iFldType, iFldLen: Integer;
  begin
    try
      Tmp := ExprEval.Evaluate(iFldType, iFldLen);
      if not Tmp then Error;
    except
      on EVariantError do Error;
    end;
  end;

var
  I: Integer;
  Expr: PChar;
begin
  ExprEval := TDSExprEvaluator.Create;
  try
    ExprEval.DSBase := Self;
    ExprEval.SetRecBuf(pRecBuf);
    for I := 0 to High(FAttrInfos) do
      with FAttrInfos[I] do
        if (iFieldNo = 0) and (AttrComp(Attr, ConstrType) = 0) then
        begin
          Expr := PChar(Value) + SizeOf(Integer);
          ExprEval.SetData(Pointer(Expr + SizeOf(TSQLExprInfo) +
            PSQLExprInfo(Expr).iErrStrLen));
          if Defaults then
            DoCheckDefault(PSQLExprInfo(Expr).iFldNum) else
            DoCheckCondition(Expr + SizeOf(TSQLExprInfo));
        end;
  finally
    ExprEval.Free;
  end;
end;

procedure TDSBase.CheckUnique(pRecBuf, pRecOrg: PChar);
var
  I, J: Integer;
  pRecFound: PChar;
  Cursor: TDSCursor;
begin
  I := 0;
  J := FLastUsedIndex;
  while I <= High(FIndexes) do
  begin
    J := J mod (High(FIndexes) + 1);
    if J > 1 then
      with FIndexes[J] do
        if bUnique then
        begin
          FLastUsedIndex := J;
          CursorNeeded(FCursor);
          Cursor := CursorFromIDS(FCursor);
          Cursor.UseIndexOrder(szName);
          if (Cursor.GetRecordForKey(iFields, 0, pRecBuf, nil) = DBERR_NONE) and
             (Cursor.GetCurrentRecordPtr(pRecFound) = DBERR_NONE) and
             (pRecFound <> pRecOrg) then
            HBError(DBERR_KEYVIOL);
        end;
    Inc(I);
    Inc(J);
  end;
end;

procedure TDSBase.CheckBlobFetched(pRecBuf: PChar; iFldNo: DWord);
var
  FldBuf: TBlobData;
  bBlank: Bool;
begin
  Check(GetField(pRecBuf, iFldNo, @FldBuf, bBlank));
  if (FldBuf.Length and dsDELAYEDBIT) = dsDELAYEDBIT then
    HBError(DBERR_BLOBNOTFETCHED);
end;

procedure TDSBase.CalculateFields(pRecBuf: PChar);
begin
  if Assigned(FCpfCalc) then FCpfCalc(FCiClientData, pRecBuf);
end;

procedure TDSBase.InternalInsertRecord(pRecBuf: PChar; var NewRecNo: Integer);
begin
  SetRecordRecNo(pRecBuf, FRecNoCounter);
  Inc(FRecNoCounter);
  NewRecNo := FRecords.Add(pRecBuf);
end;

function TDSBase.InsertRecord(pRecBuf: PChar; var NewRecNo: Integer; CommitTables: Boolean): DBResult;
var
  NewRec: PChar;
begin
  try
    AllocRecord(NewRec);
    try
      NewRecNo := -1;
      try
        CopyRecord(pRecBuf, NewRec);
        if not FIsDelta then
        begin
          CalculateFields(NewRec);
          CheckRequired(NewRec, not FAutoIncDisabled);
          CheckConstraints(NewRec, szBDEDOMX, False);
          CheckConstraints(NewRec, szBDERECX, False);
          CheckConstraints(NewRec, szBDEDOMCL, False);
          CheckConstraints(NewRec, szBDERECCL, False);
          CheckUnique(NewRec, nil);

          if FLogChanges then
            SetRecordAttr(NewRec, dsRecNew) else
            SetRecordAttr(NewRec, dsRecUnmodified);
        end;
        InternalInsertRecord(NewRec, NewRecNo);
        LogChange(dsRecNew, NewRecNo, NewRecNo);
      except
        NewRec := nil;
        raise;
      end;
    except
      FreeRecord(NewRec);
      raise;
    end;
    CommitRecord(NewRec, CommitTables);
    NotifyCursors(ceRecIns, NewRecNo, 0, 0);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.ModifyRecord(RecNo: DWord; pRecBuf: PChar; var NewRecNo: Integer; CommitTables: Boolean; Detail: Boolean): DBResult;
var
  OldRec: PChar;
  NewRec: PChar;
begin
  try
    OldRec := GetRecordPtr(RecNo);
    if BlankNotChanged(OldRec, pRecBuf, True) then
    begin
      NewRecNo := RecNo;
      Result := DBERR_NONE;
      Exit;
    end;

    AllocRecord(NewRec);
    try
      CopyRecord(pRecBuf, NewRec);
      if not FIsDelta then
      begin
        CalculateFields(NewRec);
        CheckRequired(NewRec, False);
        CheckConstraints(NewRec, szBDEDOMX, False);
        CheckConstraints(NewRec, szBDERECX, False);
        CheckConstraints(NewRec, szBDEDOMCL, False);
        CheckConstraints(NewRec, szBDERECCL, False);
        CheckUnique(NewRec, OldRec);
      end;

      SetRecordRecNo(NewRec, GetRecordRecNo(OldRec));

      if FLogChanges then
      begin
        NewRecNo := FRecords.Add(NewRec);
        if not FIsDelta then
        begin
          SetRecordAttr(NewRec, dsRecModified);
          SetRecordAttr(OldRec, GetRecordAttr(OldRec) or dsRecOrg);
        end else begin
          SetRecordAttr(NewRec, GetRecordAttr(OldRec) or dsRecModified2);
          SetRecordAttr(OldRec, GetRecordAttr(OldRec) or dsRecOrg);
        end;
        LogChange(dsRecModified, RecNo, NewRecNo);
      end else begin
        if not FIsDelta then
          SetRecordAttr(NewRec, dsRecUnmodified);
        FRecords.List^[RecNo] := NewRec;
        NewRecNo := RecNo;
        FreeRecord(OldRec);
      end;
    except
      FreeRecord(NewRec);
      raise;
    end;

    CommitRecord(NewRec, CommitTables);
    if FLogChanges then
      NotifyCursors(ceRecMod, RecNo, FRecords.Count - 1, 0) else
      NotifyCursors(ceRecMod, RecNo, -1, 0);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.DeleteRecord(RecNo: DWord): DBResult;

  procedure CascadeDelete(ChildDS: TDSBase);
  begin
    if ChildDS.FMDCursor = nil then
    begin
      ChildDS.CursorNeeded(ChildDS.FMDCursor);
      FMDCursor.LinkCursors(0, nil, nil, ChildDS.FMDCursor);
    end;

    repeat
      ChildDS.FMDCursor.MoveToBOF;
      ChildDS.FMDCursor.MoveRelative(1);
    until ChildDS.FMDCursor.DeleteRecord <> DBERR_NONE;
  end;

var
  I: Integer;
  OldRec: PChar;
begin
  try
    if Assigned(FChilds) and (dsCASCADEDELETES and FMDOptions <> 0) then
    begin
      CursorNeeded(FMDCursor);
      with CursorFromIDS(FMDCursor), FRecNums do
        for I := 0 to Count - 1 do
          if DWord(List^[I]) = RecNo then
          begin
            RecPos := I;
            Break;
          end;

      for I := 0 to High(FChilds) do
        CascadeDelete(BaseFromIDS(FChilds[I]));
    end;

    OldRec := GetRecordPtr(RecNo);

    if FLogChanges then
    begin
      SetRecordAttr(OldRec, dsRecDeleted);
      NotifyCursors(ceRecDel, RecNo, -1, 0);
      LogChange(dsRecDeleted, RecNo, RecNo);
    end else
    begin
      SetRecordAttr(OldRec, dsUnused);
      NotifyCursors(ceRecDel, RecNo, -1, -1);
      FRecords.Delete(RecNo);
      FreeRecord(OldRec);
    end;

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSBase.CommitRecord(pRecBuf: PChar; CommitTables: Boolean = True);
var
  I: Integer;
  FldBuf: TBlobData;
  bBlank: Bool;
begin
  for I := 0 to High(FFields) do
    with FFields[I] do
      case iFldType of
        fldBLOB:
          begin
            GetField(pRecBuf, I + 1, @FldBuf, bBlank);
            if FldBuf.IsNew then
            begin
              FreeMem(FldBuf.Data);
              FldBuf.Length    := FldBuf.NewLength;
              FldBuf.Data      := FldBuf.NewData;
              FldBuf.NewLength := 0;
              FldBuf.NewData   := nil;
              FldBuf.IsNew     := False;
              PutFieldBuff(pRecBuf, I + 1, @FldBuf);
            end;
          end;
        fldTABLE:
          begin
            if CommitTables then
              PutBlankInternal(pRecBuf, I + 1, NOT_BLANK);
          end;
      end;
  FBlobs.Clear;
end;

function TDSBase.GetBlob(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
  pBuf: Pointer; var iLength: DWord): DBResult;
var
  FldBuf: TBlobData;
  bBlank: Bool;
begin
  try
    CheckBlobFetched(pRecBuf, iFieldNo);

    GetField(pRecBuf, iFieldNo, @FldBuf, bBlank);

    if FldBuf.IsNew then
      Move(FldBuf.NewLength, FldBuf, SizeOf(Integer) * 2);

    if iOffset + iLength > FldBuf.Length then
      iLength := FldBuf.Length - iOffSet;

    Move((PChar(FldBuf.Data) + iOffSet)^, pBuf^, iLength);

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetBlobLen(pRecBuf: Pointer; iFieldNo: DWord;
  var iLength: DWord): DBResult;
var
  FldBuf: TBlobData;
  bBlank: Bool;
begin
  try
    CheckBlobFetched(pRecBuf, iFieldNo);

    GetField(pRecBuf, iFieldNo, @FldBuf, bBlank);

    if FldBuf.IsNew then
      iLength := FldBuf.NewLength else
      iLength := FldBuf.Length;

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.PutBlobInternal(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
  pBuf: Pointer; iLength: DWord): DBResult;
var
  FldBuf: TBlobData;
  Delayed: Boolean;
  bBlank: Bool;
  P: Pointer;
begin
  with FFields[iFieldNo - 1] do
  try
    GetField(pRecBuf, iFieldNo, @FldBuf, bBlank);

    Delayed := (iLength and dsDELAYEDBIT) = dsDELAYEDBIT;
    iLength := iLength and not dsDELAYEDBIT;

    if iLength <> FldBuf.NewLength then
    begin
      P := nil;
      ReallocMem(P, iLength);
      if Assigned(P) then FBlobs.Add(P);
      FBlobs.Remove(FldBuf.NewData);
      FreeMem(FldBuf.NewData);
      if Delayed then
        FldBuf.NewLength := dsDELAYEDBIT else
        FldBuf.NewLength := iLength;
      FldBuf.NewData := P;
    end;

    FldBuf.IsNew := True;

    Move(pBuf^, FldBuf.NewData^, iLength);

    PutFieldBuff(pRecBuf, iFieldNo, @FldBuf);

    if iLength > 0 then
      PutBlankValues(pRecBuf, iFieldNo, NOT_BLANK)
    else begin
      PutBlankInternal(pRecBuf, iFieldNo, BLANK_NULL);
      if iFieldIDParent > 0 then
        PutBlankParent(pRecBuf, iFieldIDParent);
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.PutBlob(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
  pBuf: Pointer; iLength: DWord): DBResult;
begin
  with FFields[iFieldNo - 1] do
  try
    if (iFldAttr and fldAttrREADONLY) = fldAttrREADONLY then
      HBError(DBERR_READONLYFLD);

    CheckBlobFetched(pRecBuf, iFieldNo);

    Result := PutBlobInternal(pRecBuf, iFieldNo, iOffSet, pBuf, iLength);
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.Init(Version: DWord; Columns: Word;
  FieldDesc: PDSDataPacketFldDesc): DBResult;
var
  I: Integer;
begin
  try
    Init_Sequential(Version, Columns);
    for I := 0 to Columns - 1 do
      Check(AddColumnDesc(TDSDataPacketFldDescList(FieldDesc)[I]));
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.Init_Sequential(Version: DWord; Columns: Word): DBResult;
begin
  try
    if Version > PACKETVERSION_2 then
      Result := DBERR_DATAPACKETMISMATCH else
      Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.GetDataPacket(var SA: PSafeArray): DBResult;
begin
  Result := DBERR_NA;
end;

function TDSBase.GetSize(var DataPacketSize: DWord): DBResult;
begin
  Result := DBERR_NA;
end;

function TDSBase.GetColumnCount(var Count: DWord): DBResult;
begin
  Result := DBERR_NA;
end;

function TDSBase.GetColumnDescs(FieldDes: PDSDataPacketFldDesc): DBResult;
begin
  Result := DBERR_NA;
end;

function TDSBase.AddColumnDesc(const AFldDes: TDSDataPacketFldDesc): DBResult;
begin
  try
    with FWDataSet do
    begin
      if not FWHasMetaData then
        AddFldInfo(FWFldDescs, AFldDes);

      // It is possible situation when FWFldDescs not filled
      // but children datasets are created

      if (PacketFieldType(AFldDes.iFieldType) = dsfldEMBEDDEDTBL) then
        if FWHasChildren then
          Self.FWDataSet := GetEmbeddedDSBaseByDescNo(FWDataSetFieldNo)
        else begin
          CreateEmbeddedDS(-1);
          Self.FWDataSet := BaseFromIDS(FChilds[High(FChilds)]);
          Self.FWDataSet.WriterStart;
        end;
      Inc(FWDataSetFieldNo);
    end;
    FWDataSetAttrInfos := @FWDataSet.FWAttrInfos;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.AddAttribute(AAttrArea: TPcktAttrArea; AAttr: PChar;
  AAttrType, ALen: DWord; AValue: Pointer): DBResult;
begin
  try
    if not Assigned(AAttr) or (StrLen(AAttr) = 0) then
    begin
      FWDataSet := FWDataSet.FParentDS;
      FWDataSetAttrInfos := @FWDataSet.FWAttrInfos;
    end else
      case AAttrArea of
        fldAttrArea:
          with FWDataSet do
            if not FWHasMetaData then
              AddAttrInfo(Self.FWDataSetAttrInfos^, FWDataSetFieldNo, AAttr, AAttrType, ALen, AValue);
        pcktAttrArea:
          if AttrComp(AAttr, szDATASET_DELTA) = 0 then
            FWDataSet.FIsDelta := Boolean(AValue^)
          else if AttrComp(AAttr, szCHANGE_LOG) = 0 then
          begin
            SetLength(FWDataSet.FWChanges, ALen div SizeOf(TChangeInfo));
            Move(AValue^, Pointer(FWDataSet.FWChanges)^, ALen);
          end else if AttrComp(AAttr, szREADONLY) = 0 then
            FWDataSet.FStreamedReadOnly := Boolean(AValue^)
          else if AttrComp(AAttr, szAUTOINCVALUE) = 0 then
            FWDataSet.FAutoIncCounter := Integer(AValue^)
          else if AttrComp(AAttr, szCONSTRAINTS) = 0 then
            FWDataSet.FConstrDisabled := Boolean(AValue^)
          else if AttrComp(AAttr, szMD_SEMANTICS) = 0 then
            FWDataSet.FMDOptions := DWord(AValue^)
          else if AttrComp(AAttr, szLCID) =  0 then
            FWDataSet.FLCID := Integer(AValue^)
          else
            AddAttrInfo(FWDataSetAttrInfos^, 0, AAttr, AAttrType, ALen, AValue);
      end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.WriterPutField(AFldStatus: TPcktFldStatus; ALen: DWord;
  ASrc: Pointer): DBResult;

  procedure AddFldDatas(FldInfo: PFldInfo; ALen: DWord; ASrc: Pointer);
  var
    H: Integer;
    FldType: DWord;
  const
    C: Char = #0;
  begin
    H := Length(FldInfo.FldDatas);
    SetLength(FldInfo.FldDatas, H + 1);
    with FldInfo.FldDatas[H] do
    begin
      FldStatus := AFldStatus;

      FldType := PacketFieldType(FldInfo.FldDes.iFieldType);
      if AFldStatus = fldIsChanged then
      begin
        with FWFldDatas do
          Offs := Position;
        case FldType of
          dsfldZSTRING:
            begin
              Len := ALen + 1;
              FWFldDatas.WriteBuffer(ASrc^, ALen);
              FWFldDatas.WriteBuffer(C, SizeOf(C));
            end;
          dsfldUNICODE:
            begin
              FWFldDatas.WriteBuffer(ALen, 2);
              FWFldDatas.WriteBuffer(ASrc^, ALen);
            end;
        else
          Len := ALen;
          FWFldDatas.WriteBuffer(ASrc^, Len and not dsDELAYEDBIT);
        end;
      end;
    end;
  end;

const
  I: Integer = 0;

var
  FldInfo: pFldInfo;
  FieldNo: DWord;
begin
  try
    FWDataSet.WriterCreateMetaData;

    if FWDataSet.FWStateReached then
    begin
      with FWDataSet do
        FldInfo := @FWFldDescs[FWDataSetFieldNo];

      Inc(I);
      AddFldDatas(FldInfo, ALen, ASrc);

      with FldInfo^ do
      begin
        if FldArrayElem then
          FldArrayElemCounter := (FldArrayElemCounter + 1) mod FldArrayElemCount;

        with FWDataSet do
        begin
          FieldNo := FWDataSetFieldNo;
          if FldArrayElemCounter = 0 then
          begin
            FWDataSetFieldNo := (FWDataSetFieldNo + 1) mod Length(FWFldDescs);
            if FWDataSetFieldNo = 0 then
              FWStateReached := False;
          end;
        end;

        if (AFldStatus <> fldIsNull) and (PacketFieldType(FldDes.iFieldType) = dsfldEMBEDDEDTBL) and
          ((PDWord(ASrc)^ = DWord(-1)) or ((PDWord(ASrc)^ and dsDELAYEDBIT) = 0)) then
        begin
          FWDataSet := FWDataSet.GetEmbeddedDSBase(FieldNo + 1);
          FWDataSet.FWStateReached := False;
        end;
      end;
    end else with FWDataSet do
    begin
      FWStates.Add(Pointer(DSAttr(ASrc^)));
      if FLinkFieldCreated then
        FWRecNos.Add(Pointer(FParentDS.FRecords.Count + FParentDS.FWStates.Count - 1));
      FWStateReached := True;
      FWDataSetFieldNo := 0;
    end;

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.EndOfNestedRows: DBResult;
begin
  try
    FWDataSet := FWDataSet.FParentDS;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSBase.WriterReset: DBResult;
begin
  Result := DBERR_NA;
end;

procedure TDSBase.WriterStart;
begin
  FWHasMetaData := Assigned(FWFldDescs);
  FWHasChildren := Length(FChilds) > 0;
  FWDataSet := Self;
  FWDataSetAttrInfos := @FWAttrInfos;
  FWDataSetFieldNo := 0;
  FWStateReached := False;
  FWStates := TList.Create;
  FWRecNos := TList.Create;
  ResetElemCounters(FWFldDescs);
end;

procedure TDSBase.WriterCreateRecord(Recs: TList; Index: Integer);

var
  pRecBuf: PChar;

  procedure PutFieldData(var FieldInfo, FieldNo: Integer);
  var
    Buff: Pointer;
    FldData: PFldData;
  begin
    with FWFldDescs[FieldInfo] do
    begin
      if FldArrayElem then
        FldData := @FldDatas[Index * FldArrayElemCount + FldArrayElemCounter] else
        FldData := @FldDatas[Index];

      if FldArrayElem then
        FldArrayElemCounter := (FldArrayElemCounter + 1) mod FldArrayElemCount;

      if FldData.FldStatus = fldIsUnChanged then
        PutBlankValues(pRecBuf, FieldNo + 1, BLANK_NOTCHANGED)
      else if FldData.FldStatus = fldIsNull then
        PutBlankValues(pRecBuf, FieldNo + 1, BLANK_NULL)
      else if FldData.FldStatus = fldIsChanged then
      begin
        Buff := PChar(FWFldDatas.Memory) + FldData.Offs;
        case FFields[FieldNo].iFldType of
          fldBLOB:
            PutBlobInternal(pRecBuf, FieldNo + 1, 0, Buff, FldData.Len);
        else
          PutFieldInternal(pRecBuf, FieldNo + 1, Buff);
        end;
      end;

      Inc(FieldNo);

      if FldArrayElem then
        FldArrayElemCounter := (FldArrayElemCounter + 1) mod FldArrayElemCount;

      if FldArrayElemCounter = 0 then
        Inc(FieldInfo);
    end;
  end;

var
  I, FieldNo: Integer;
begin
  AllocRecord(pRecBuf);
  try
    I := 0;
    FieldNo := 0;
    while FieldNo <= High(FWFldDescs) do
      PutFieldData(I, FieldNo);
    SetRecordAttr(pRecBuf, DSAttr(FWStates.List^[Index]));

    if FLinkFieldCreated then
    begin
      I := Integer(FWRecNos.List^[Index]);
      PutFieldInternal(pRecBuf, FLinkFieldNo, @I);
    end;

    CommitRecord(pRecBuf, False);
    Recs.Add(pRecBuf);
  except
    FreeRecord(pRecBuf);
    raise;
  end;
end;

procedure TDSBase.WriterCreateMetaData;

var
  TmpFields: DSFLDDescList;

  procedure DoCreateIndex(IndexName: PChar; AttrType, Len: DWord; Value: Pointer);
  var
    I: Integer;
    S: string;
    Index: DSIDXDesc;
    KeyFields: Variant;
  begin
    ZeroMem(@Index, SizeOf(DSIDXDesc));
    KeyFields := AttrToVariant(AttrType, Len, Value);

    Index.iFields := VarArrayHighBound(KeyFields, 1) + 1;
    for I := 0 to Index.iFields - 1 do
      Index.iKeyFields[I] := KeyFields[I];
    Index.bUnique := True;

    if Assigned(IndexName) then
      StrCopy(Index.szName, IndexName)
    else begin
      S := '';
      for I := 0 to Index.iFields - 1 do
        AddDelimeted(S, FFields[Index.iKeyFields[I] - 1].szName, ';');
      StrCopy(Index.szName, PChar(S));
    end;
    CreateIndex(Index);
  end;

  procedure AddAttributes(FieldNo: DWord);
  var
    I: Integer;
  begin
    if not Assigned(FWAttrInfos) then Exit;

    for I := High(FAttrInfos) downto 0 do
      with FAttrInfos[I] do
        if iFieldNo = FieldNo then
          DeleteAttrInfo(FAttrInfos, I);

    for I := 0 to High(FWAttrInfos) do
      with FWAttrInfos[I] do
        if iFieldNo = FieldNo then
        begin
          AddAttrInfo(FAttrInfos, FieldNo, Attr, AttrType, Len, Value);
          if (iFieldNo = 0) and ((StrIComp(Attr, szPRIMARY_KEY) = 0) or (StrIComp(Attr, szUNIQUE_KEY) = 0)) then
          begin
            if StrIComp(Attr, szPRIMARY_KEY) = 0 then
              DoCreateIndex(szPRIMARY_KEY, AttrType, Len, Value) else
              DoCreateIndex(nil, AttrType, Len, Value);
          end;
        end;
  end;

  procedure DoAddField(ParentID: DWord; var DescNo: Integer);
  var
    I, StartDescNo, Parent: DWord;
    FldDes: pDSFLDDesc;
  begin
    FldDes := @TmpFields[DescNo];
    AddFieldInternal(ParentID, @TmpFields[DescNo], False);
    Inc(DescNo);
    AddAttributes(Length(FFields));
    with FldDes^ do
      if iFldType = fldADT then
      begin
        Parent := High(FFields) + 1;
        for I := 1 to iUnits1 do
          DoAddField(Parent, DescNo)
      end else if iFldType = fldARRAY then
      begin
        Parent := High(FFields) + 1;
        StartDescNo := DescNo;
        for I := 1 to iUnits1 do
        begin
          DescNo := StartDescNo;
          DoAddField(Parent, DescNo);
        end;
      end
  end;

var
  I, J: Integer;
  Child: TDSBase;
  FldDesc: DSFLDDesc;
begin
  if Assigned(FFields) then Exit;

  if not Assigned(FWFldDescs) then
    HBError(DBERR_NOMETADATA);

  for I := 0 to High(FWFldDescs) do
  begin
    PacketDSToFldDesc(FWFldDescs[I].FldDes, I + 1, FWAttrInfos, FldDesc);
    AddFieldDesc(TmpFields, FldDesc);
  end;

  J := 0;
  for I := 0 to High(TmpFields) do
    with TmpFields[I] do
      if iFldType = fldTABLE then
      begin
        Child := BaseFromIDS(FChilds[J]);
        Child.FParentFieldNo := I + 1;
        Child.WriterCreateMetaData;
        iUnits1 := Length(Child.FFields);
        Inc(J);
      end;

  I := 0;
  while I <= High(TmpFields) do
    DoAddField(0, I);

  if Assigned(FParentDS) and (FLinkFieldNo = 0) then
  begin
    InitLinkField(FldDesc);
    AddFieldInternal(0, @FldDesc, True);
    FLinkFieldCreated := True;
  end;

  AddAttributes(0);
end;

procedure TDSBase.WriterApply;
var
  I, OldRecNo, NewRecNo, LastChange, StartRec: Integer;
  pRecBuf: PChar;
  Recs: TList;
begin
  WriterCreateMetaData;

  DisableNotify;
  try
    Recs := TList.Create;
    try
      try
        for I := 0 to FWStates.Count - 1 do
          WriterCreateRecord(Recs, I);
      except
        with Recs do
          for I := 0 to Count - 1 do
          begin
            pRecBuf := List^[I];
            FreeRecord(pRecBuf);
          end;
        raise;
      end;

      StartRec := FRecords.Count;
      LastChange := 0;

      // Append created records to the database
      with Recs do
      begin
        FRecords.Capacity := FRecords.Count + Count;

        for I := 0 to Count - 1 do
        begin
          pRecBuf := List^[I];
          InternalInsertRecord(pRecBuf, NewRecNo);
          if LastChange <= High(FWChanges) then
          begin
            with FWChanges[LastChange] do
              if NewSeqNo = I then
              begin
                if RecAttr = dsRecModified then
                begin
                  OldRecNo := GetRecordRecNo(FRecords.List^[OldSeqNo + StartRec]);
                  SetRecordRecNo(pRecBuf, OldRecNo);
                end;
                LogChange(RecAttr, OldSeqNo + StartRec, NewSeqNo + StartRec);
                Inc(LastChange);
              end;
          end;
          NotifyCursors(ceRecIns, NewRecNo, 0, 0);
        end;
      end;
    finally
      Recs.Free;
    end;
  finally
    EnableNotify;;
  end;
  NotifyCursors(ceNotifyAppend, 0, 0, 0);
end;

procedure TDSBase.WriterFree;
begin
  FreeAttrInfos(FWAttrInfos);
  FreeFieldInfos(FWFldDescs);
  FreeAndNil(FWStates);
  FreeAndNil(FWRecNos);
  FWChanges := nil;
end;

type
{ TDSAggExprEvaluator }
  TDSAggExprEvaluator = class(TDSExprEvaluator)
  private
    FDSCursor: TDSCursor;
    FiFields : Integer;
    FValid   : Boolean;
    FRecNo   : DWord;
    FValue   : Variant;
    FR1, FR2, FRec: DWord;
  protected
    procedure DoFirst; override;
    procedure DoNext; override;
    function DoGetEOF: Boolean; override;
    procedure SetRecordByRecNo(RecNo: DWord);
  public
    constructor Create(ACursor: TDSCursor; AFields: Integer);
    destructor Destroy; override;
    function EvaluteRecNo(ARecNo: DWord): Variant;
    procedure Invalidate;
    property DSCursor: TDSCursor read FDSCursor;
    property iFields: Integer read FiFields;
    property Valid: Boolean read FValid;
    property Value: Variant read FValue;
  end;

constructor TDSAggExprEvaluator.Create(ACursor: TDSCursor; AFields: Integer);
begin
  FDSCursor := ACursor;
  FiFields := AFields;
end;

destructor TDSAggExprEvaluator.Destroy;
begin
  Invalidate;
  inherited;
end;

function TDSAggExprEvaluator.EvaluteRecNo(ARecNo: DWord): Variant;
begin
  try
    if ARecNo = DWord(-1) then
    begin
      Invalidate;
      FR1 := DWord(-1);
      FR2 := DWord(-1);
      FValue := EvaluteRec(nil);
    end else begin
      FDSCursor.GetAggBounds(FiFields, ARecNo, FR1, FR2);
      if not (FValid and (FRecNo >= FR1) and (FRecNo <= FR2)) then
      begin
        FRecNo := ARecNo;
        FValue := Evaluate(FiFldType, FiFldLen);
        FDSCursor.FAggsValid.Add(Self);
        FValid := True;
      end;
    end;
  except
    FValue := Null;
  end;
  Result := FValue;
end;

procedure TDSAggExprEvaluator.Invalidate;
begin
  if FValid then
  begin
    FValue := Unassigned;
    FValid := False;
    FDSCursor.FAggsValid.Remove(Self);
  end;
end;

procedure TDSAggExprEvaluator.SetRecordByRecNo(RecNo: DWord);
begin
  if (RecNo <> DWord(-1)) and (RecNo >= FR1) and (RecNo <= FR2) then
    SetRecBuf(FDSBase.FRecords.List^[Integer(FDSCursor.FRecNums.List^[RecNo])]);
end;

procedure TDSAggExprEvaluator.DoFirst;
begin
  FRec := FR1;
  SetRecordByRecNo(FRec);
end;

procedure TDSAggExprEvaluator.DoNext;
begin
  if FRec <> DWord(-1) then
  begin
    Inc(FRec);
    SetRecordByRecNo(FRec);
  end;
end;

function TDSAggExprEvaluator.DoGetEOF: Boolean;
begin
  Result := (FRec = DWord(-1)) or (FRec > FR2);
end;

{ TDSCursor }
procedure TDSCursor.Initialize;
begin
  inherited;
  FRecNums := TList.Create;
  FAggs := TList.Create;
  FAggsValid := TList.Create;
  FDetails := TList.Create;
  FExprEval := TDSExprEvaluator.Create;
end;

destructor TDSCursor.Destroy;
begin
  InitCursor(nil);
  FAggs.Free;
  FAggsValid.Free;
  FRecNums.Free;
  FExprEval.Free;
  FDetails.Free;
  inherited;
end;

procedure TDSCursor.GetAggBounds(iFields: Integer; RecNo: DWord; var R1, R2: DWord);
var
  pRecBuf: PChar;
begin
  if FDefaultOrder or FChangeIndex or (iFields = 0) then
  begin
    R1 := 0;
    R2 := FRecNums.Count - 1;
  end else begin
    pRecBuf := FDSBase.FRecords.List^[Integer(FRecNums.List^[RecNo])];
    if InternalMoveToKey(keySEARCHEQ, iFields,
      0, pRecBuf, nil, Integer(R1)) = DBERR_NONE then
    begin
      if InternalMoveToKey(keySEARCHGT, iFields,
        0, pRecBuf, nil, Integer(R2)) <> DBERR_NONE then
          R2 := R1
      else
        Dec(R2);
    end else begin
      // This situation can't happen
      R1 := DWord(-1);
      R2 := DWord(-1);
    end;
  end
end;

procedure TDSCursor.InvalidateAggs;
var
  I: Integer;
begin
  with FAggsValid do
    for I := Count - 1 downto 0 do
      TDSAggExprEvaluator(FAggsValid.List^[I]).Invalidate;
end;

function TDSCursor.AddAggregate(iFlds, iCanLen: DWord; pCanExpr: Pointer;
  var hAgg: hDSAggregate): DBResult;
var
  A: TDSAggExprEvaluator;
begin
  try
    A := TDSAggExprEvaluator.Create(Self, iFlds);
    try
      A.DSBase := FDSBase;
      A.SetData(pCanExpr);
      FAggs.Add(A);
      hAgg := hDSAggregate(A);
      Result := DBERR_NONE;
    except
      A.Free;
      raise;
    end;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.DropAggregate(hAgg: hDSAggregate): DBResult;
var
  I: Integer;
begin
  try
    I := FAggs.IndexOf(Pointer(hAgg));
    if I >= 0 then
    begin
      TDSAggExprEvaluator(hAgg).Free;
      FAggs.Delete(I);
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetAggregateDesc(hAgg: hDSAggregate; var Desc: DSFLDDesc): DBResult;
begin
  try
    ZeroMem(@Desc, SizeOf(DSFLDDesc));
    with TDSAggExprEvaluator(hAgg) do
    begin
      EvaluteRecNo(DWord(-1));
      Desc.iFldType := iFldType;
      Desc.iFldLen  := iFldLen;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetAggregateValue(hAgg: hDSAggregate; pValue: Pointer;
  var bBlank: Bool): DBResult;
var
  P: Pointer;
  Tmp: Variant;
begin
  try
    UpdateCursorPos;
    with TDSAggExprEvaluator(hAgg) do
    begin
      Tmp := EvaluteRecNo(FRecPos);
      bBlank := VarIsNull(Tmp);
      if not bBlank then
      begin
        P := GetValuePtr(Tmp, iFldType);
        Move(P^, pValue^, iFldLen);
      end;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetSubGroupState(iFields: DWord;
  var iSubGroupState: GROUPSTATE): DBResult;
var
  RecNo, R1, R2: DWord;
begin
  try
    CheckCursorPos;
    RecNo := FRecPos;
    GetAggBounds(iFields, RecNo, R1, R2);
    if (R1 = R2) then
      iSubGroupState := grSTATEFIRSTLAST
    else if (RecNo = R1) then
      iSubGroupState := grSTATEFIRST
    else if (RecNo = R2) then
      iSubGroupState := grSTATELAST
    else
      iSubGroupState := grSTATEMIDDLE;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.MoveToNextSubGroup(iFields: DWord): DBResult;
begin
  try
    Result := DBERR_NIY;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.AddFilter(pcanExpr: Pointer; iLen: DWord;
  var hFilter: hDSFilter): DBResult;
begin
  try
    FExprEval.SetData(pcanExpr);
    hFilter := Pointer(1);
    Rebuild;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.AddFilterCallBack(iClientData: DWord;
  pfFilter: pfDSFilter; var hFilter: hDSFilter): DBResult;
begin
  try
    FFClientData := iClientData;
    FFFilter     := pfFilter;
    hFilter      := Pointer(2);
    Rebuild;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.DropFilter(hFilter: hDSFilter): DBResult;
begin
  try
    case Integer(hFilter) of
      1: FExprEval.SetData(nil);
      2: FFFilter     := nil;
    else
      Result := DBERR_NOSUCHFILTER;
      Exit;
    end;
    RangeDropped;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.CloneCursor(Cursor: IDSCursor): DBResult;
var
  Cur: TDSCursor;
  pKeyRec, pRecBufM: PChar;
begin
  try
    Cur := CursorFromIDS(Cursor);
    Result := InitCursor(Cur.FDSBase);
    Cursor.GetIndexDescs(True, FIndexDS);
    // Copy filters
    FExprEval.SetData(Cur.FExprEval.Data);
    FFClientData := Cur.FFClientData;
    FFFilter     := Cur.FFFilter;
    // Link to master
    if Assigned(Cur.FMaster) then
    begin
      Cur.FMaster.FDSBase.AllocRecord(pKeyRec);
      if Cur.FMaster.GetCurrentRecordPtr(pRecBufM) = DBERR_NONE then
        Cur.FMaster.FDSBase.CopyRecord(pRecBufM, pKeyRec);
      FMasterRecBuf := pKeyRec;
      Cur.FMaster.LinkCursors(
        0,
        @Cur.FFieldsM[0],
        @Cur.FFieldsD[0],
        Self as IDSCursor);
    end else
      Rebuild;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.CompareBookMarks(pBookMark1, pBookMark2: Pointer;
  var iCmp: Integer): DBResult;
var
  SeqNo1, SeqNo2, Bmk1, Bmk2: Integer;
begin
  SeqNo1 := pDSBOOKMRK(pBookmark1)^.iRecNo;
  SeqNo2 := pDSBOOKMRK(pBookmark2)^.iRecNo;
  with FRecNums do
  begin
    Bmk1 := FindInteger(SeqNo1, List^, Count);
    Bmk2 := FindInteger(SeqNo2, List^, Count);
  end;
  if Bmk1 > Bmk2 then iCmp :=  1 else
  if Bmk1 < Bmk2 then iCmp := -1 else iCmp := 0;
  Result := DBERR_NONE;
end;

procedure TDSCursor.CheckCursorPos;
begin
  if FCrack then
    HBError(DBERR_NOCURRREC)
  else if FRecPos >= FRecNums.Count then
    HBError(DBERR_EOF)
  else if FRecPos < -1 then
    HBError(DBERR_BOF)
  else if FRecPos < 0 then
    HBError(DBERR_EOF);
end;

procedure TDSCursor.UpdateCursorPos;
var
  Recs: Integer;
begin
  Recs := FRecNums.Count;

  if Recs = 0 then
    RecPos := -1;

  if FRecPos < -1 then
    RecPos := -1
  else if FRecPos >= Recs then
    RecPos := Recs - 1;
end;

function TDSCursor.SetRange(iFields: DWord; pKey1: Pointer;
  bKey1Incl: Bool; pKey2: Pointer; bKey2Incl: Bool): DBResult;
begin
  try
    CheckIndexed;

    if not Assigned(FKeyRec1) then
    begin
      FDSBase.AllocRecord(FKeyRec1);
      FDSBase.AllocRecord(FKeyRec2);
    end;

    FDSBase.CopyRecord(pKey1, FKeyRec1);
    FDSBase.CopyRecord(pKey2, FKeyRec2);
    FKeyFields := iFields;
    FKeyRec1Incl := bKey1Incl;
    FKeyRec2Incl := bKey2Incl;

    Rebuild;

    Result := MoveToBOF;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.InsertRecNo(RecNo: Integer): Integer;
var
  I: Integer;
  pRecBuf: PChar;
begin
  with FRecNums do
  begin
    Result := Count;
    I := FindInteger(RecNo, List^, Result);
    if I < 0 then
    begin
      pRecBuf := FDSBase.FRecords.List^[RecNo];
      InternalMoveToKey(keySEARCHGT, 0, 0, pRecBuf, nil, Result);
      Insert(Result, Pointer(RecNo));
      InvalidateAggs;
    end;
  end;
end;

procedure TDSCursor.RemoveRecNo(RecNo: Integer);
var
  I: Integer;
begin
  with FRecNums do
    I := FindInteger(RecNo, List^, Count);
  if I >= 0 then
  begin
    FRecNums.Delete(I);
    InvalidateAggs;
  end;
end;

procedure TDSCursor.RangeDropped;
var
  I: Integer;
  Rec, SaveRec: PChar;
begin
  with FDSBase.FRecords do
  begin
    if (FRecPos >= 0) and (FRecPos < FRecNums.Count) then
      SaveRec := List^[FRecPos] else SaveRec := nil;
    RecPos := -1;
    for I := 0 to Count - 1 do
    begin
      Rec := List^[I];
      if AcceptsRecord(Rec) then
      begin
        InsertRecNo(I);
        if Rec = SaveRec then RecPos := I;
      end;
    end;
  end;
end;

function TDSCursor.DropRange: DBResult;
begin
  try
    if Assigned(FKeyRec1) then
    begin
      FDSBase.FreeRecord(FKeyRec1);
      FDSBase.FreeRecord(FKeyRec2);
      RangeDropped;
      Result := DBERR_NONE;
    end else
      Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.ExtractKey(pRecBuf, pKeyBuf: Pointer): DBResult;
begin
  try
    CheckIndexed;
    Result := FDSBase.ExtractKey(pRecBuf, pKeyBuf, FIndexDS);
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetBlob(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
  pBuf: Pointer; var iLength: DWord): DBResult;
begin
  Result := FDSBase.GetBlob(pRecBuf, iFieldNo, iOffSet, pBuf, iLength);
end;

function TDSCursor.GetBlobLen(pRecBuf: Pointer; iFieldNo: DWord;
  var iLength: DWord): DBResult;
begin
  Result := FDSBase.GetBlobLen(pRecBuf, iFieldNo, iLength);
end;

function TDSCursor.GetCurrentBookMark(pBookMark: Pointer): DBResult;
var
  pRecBuf: PChar;
begin
  try
    CheckCursorPos;
    pRecBuf := FDSBase.FRecords.List^[CurrentSeqNo];
    with pDSBOOKMRK(pBookMark)^ do
    begin
      iPos     := FRecPos;
      iState   := GetRecordAttr(pRecBuf);
      iRecNo   := GetRecordRecNo(pRecBuf);
      iSeqNo   := FRecPos + 1;
      iOrderID := 0;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetCurrentRecord(pRecBuf: Pointer): DBResult;
begin
  try
    CheckCursorPos;
    Result := FDSBase.GetRecord(CurrentSeqNo, pRecBuf);
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetCursorProps(var p1: DSProps): DBResult;
begin
  Result := FDSBase.GetProps(p1);
end;

function TDSCursor.GetField(pRecBuf: Pointer; iFieldNo: DWord;
  pFldBuf: Pointer; var bBlank: Bool): DBResult;
begin
  Result := FDSBase.GetField(pRecBuf, iFieldNo, pFldBuf, bBlank);
end;

function TDSCursor.GetCurrentRecordPtr(var pRecBuf: PChar): DBResult;
begin
  try
    CheckCursorPos;
    with FDSBase do pRecBuf := FRecords.List^[CurrentSeqNo];
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetFieldDescs(p1: pDSFLDDesc): DBResult;
begin
  Result := FDSBase.GetFieldDescs(p1);
end;

function TDSCursor.GetIndexDescs(bCurrentOnly: Bool; var IdxDesc: DSIDXDesc): DBResult;
begin
  if bCurrentOnly then
  begin
    Move(FIndexDS, IdxDesc, SizeOf(DSIDXDesc));
    Result := DBERR_NONE;
  end else
    Result := FDSBase.GetIndexDescs(@IdxDesc);
end;

function TDSCursor.GetProp(eProp: CURProp; piPropValue: Pointer): DBResult;
var
  RecNo: DWord;
begin
  case eProp of
    curpropCONSTRAINT_ERROR_MESSAGE:{ r pCHAR,   Constraint Error Message }
      ;
    curpropDSBASE:                  { r pDSBASE, Underlying DSBASE) }
      IDSBase(piPropValue^) := FIDSBase;
    curpropSETCRACK:                { w Sets crack-value to supplied value (DBERR_NOCURRREC) }
      ;
    curpropGETORG_RECBUF:           { r returns recordbuffer for original record, error if none }
      begin
        CheckCursorPos;
        FDSBase.GetRecordOrg(CurrentSeqNo, RecNo, piPropValue);
      end;
    else if eProp = CURProp(-1) then
      TDSCursor(piPropValue^) := Self;
  end;
  Result := DBERR_NONE;
end;

function TDSCursor.GetRecordAttribute(var Attr: DSAttr): DBResult;
begin
  try
    CheckCursorPos;
    with FDSBase do
      Attr := GetRecordAttr(FRecords.List^[CurrentSeqNo]);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetRecordBlock(piRecs: PDWord; pRecBuf: Pointer): DBResult;
begin
  Result := DBERR_NIY;
end;

procedure TDSCursor.CheckFetched;
var
  pRecBufM: PChar;
  FldBuf: DWord;
  bBlank: Bool;
begin
  if Assigned(FMaster) and (FMaster.GetCurrentRecordPtr(pRecBufM) = DBERR_NONE) then
  begin
    FMaster.GetField(pRecBufM, FDSBase.FParentFieldNo, @FldBuf, bBlank);
    if (FldBuf <> DWord(-1)) and ((FldBuf and dsDELAYEDBIT) <> 0) then
      HBError(DBERR_DETAILSNOTFETCHED);
  end;
end;

function TDSCursor.GetRecordCount(var iRecs: Integer): DBResult;
begin
  try
    CheckFetched;
    iRecs := FRecNums.Count;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetRecordNumber(var iRecNo: DWord): DBResult;
begin
  try
    CheckCursorPos;
    iRecNo := FRecPos + 1;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.GetSequenceNumber(var iSeq: DWord): DBResult;
var
  Bmk: DSBOOKMRK;
begin
  Result := GetCurrentBookMark(@Bmk);
  if Result = DBERR_NONE then iSeq := Bmk.iSeqNo;
end;

function TDSCursor.InitCursor(DataSet: IDSBase): DBResult;
var
  I: Integer;
  TmpBase: TDSBase;
begin
  try
    if Assigned(FDSBase) then
    begin
      Clear;
      FDSBase.RemoveCursor(Self);
      TDSExprEvaluator(FExprEval).DSBase := nil;
      ListDestroyObjects(FAggs, False);
      FFields := nil;
      if Assigned(FMaster) then
      begin
        if Assigned(FMasterRecBuf) and Assigned(FMaster.FDSBase) then
          FMaster.FDSBase.FreeRecord(FMasterRecBuf);
        FMaster.RemoveCursor(Self);
      end;

      for I := FDetails.Count - 1 downto 0 do
        RemoveCursor(FDetails[I]);
    end;

    FIDSBase := nil;

    if Assigned(DataSet) then
    begin
      TmpBase := BaseFromIDS(DataSet);
      TmpBase.InsertCursor(Self);
      FIDSBase := DataSet;
      TDSExprEvaluator(FExprEval).DSBase := FDSBase;
      SetLength(FFields, Length(FDSBase.FFields));
      Move(Pointer(FDSBase.FFields)^, Pointer(FFields)^, SizeOf(DSFldDesc) * Length(FFields));
      FIsDelta := FDSBase.FIsDelta;
      ZeroMem(@FIndexDS, SizeOf(DSIDXDesc));
      UseIndexOrder(nil);
      Rebuild;
    end;

    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.InitRecord(pRecBuf: Pointer): DBResult;
begin
  try
    FDSBase.InitRecord(pRecBuf);
    FDSBase.CheckConstraints(pRecBuf, szBDEDEFX, True);
    FDSBase.CheckConstraints(pRecBuf, szBDEDEFCL, True);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.InsertRecord(pRecBuf: Pointer): DBResult;
var
  NewPos, NewRecNo, Tmp: Integer;
  pRecBufM: PChar;
begin
  try
    FDSBase.CheckNotReadOnly;
    if FDSBase.FIsDelta then
      HBError(DBERR_NA);
    CheckIndexWritable;
    if Assigned(FMaster) then
    begin
      Check(FMaster.GetCurrentRecordPtr(pRecBufM));
      if FFieldsM[0] = 0 then
      begin
        Tmp := GetRecordRecNo(pRecBufM);
        FDSBase.PutFieldInternal(pRecBuf, FFieldsD[0], @Tmp);
      end;
      Tmp := -1;
      FMaster.FDSBase.PutFieldInternal(pRecBufM, FDSBase.FParentFieldNo, @Tmp);
    end;
    Check(FDSBase.InsertRecord(pRecBuf, NewRecNo));
    with FRecNums do
      NewPos := FindInteger(NewRecNo, List^, Count);
    if NewPos >= 0 then RecPos := NewPos;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.ModifyRecord(pRecBuf: Pointer): DBResult;
var
  NewPos, NewRecNo: Integer;
begin
  try
    FDSBase.CheckNotReadOnly;
    CheckIndexWritable;
    CheckCursorPos;
    Check(FDSBase.ModifyRecord(CurrentSeqNo, pRecBuf, NewRecNo));
    with FRecNums do
      NewPos := FindInteger(NewRecNo, List^, Count);
    if NewPos >= 0 then RecPos := NewPos;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.DeleteRecord: DBResult;
begin
  try
    FDSBase.CheckNotReadOnly;
    if FDSBase.FIsDelta then
      HBError(DBERR_NA);
    CheckCursorPos;
    CheckIndexWritable;
    Check(FDSBase.DeleteRecord(CurrentSeqNo));
    UpdateCursorPos;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.LinkCursors(iFieldsM: DWord; piFieldsM,
  piFieldsD: PDWord; hCurDet: IDSCursor): DBResult;
var
  Cursor: TDSCursor;
begin
  try
    Cursor := CursorFromIDS(hCurDet);
    with Cursor do
    begin
      if iFieldsM > 0 then
      begin
        HBError(DBERR_NIY);
        SetLength(FFieldsM, iFieldsM);
        SetLength(FFieldsD, iFieldsM);
        Move(piFieldsM^, Pointer(FFieldsM)^, iFieldsM * SizeOf(DWord));
        Move(piFieldsD^, Pointer(FFieldsD)^, iFieldsM * SizeOf(DWord));
      end else
      begin
        if FDSBase.FLinkFieldNo = 0 then
          HBError(DBERR_INVALIDLINKEXPR);
        SetLength(FFieldsM, 1);
        SetLength(FFieldsD, 1);
        FFieldsM[0] := 0;
        FFieldsD[0] := FDSBase.FLinkFieldNo;
      end;
    end;
    InsertCursor(Cursor);
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.LocateWithFilter(pCanExpr: Pointer; iLen: DWord): DBResult;
var
  I: Integer;
  ExprEval: TDSExprEvaluator;
  Records: PPointerList;
begin
  ExprEval := TDSExprEvaluator.Create;
  with TDSExprEvaluator(ExprEval) do
  try
    DSBase := Self.FDSBase;
    SetData(pCanExpr);
    Records := FDSBase.FRecords.List;
    with FRecNums do
      for I := 0 to Count - 1 do
        if EvaluteRec(Records^[Integer(List^[I])]) then
        begin
          RecPos := I;
          Result := DBERR_NONE;
          Exit;
        end;
    Result := DBERR_RECNOTFOUND;
  finally
    ExprEval.Free;
  end;
end;

function TDSCursor.MoveRelative(i: Integer): DBResult;
var
  NewPos: Integer;
begin
  NewPos := FRecPos + I;
  if NewPos < -1 then
  begin
    RecPos := -1;
    Result := DBERR_BOF
  end else if NewPos >= FRecNums.Count then
  begin
    RecPos := FRecNums.Count;
    Result := DBERR_EOF
  end else begin
    RecPos := NewPos;
    Result := DBERR_NONE;
  end;
end;

function TDSCursor.MoveToBookMark(pBookMark: Pointer): DBResult;
var
  I, NewPos, RecNo: Integer;
  pRecBuf: PChar;
begin
  if Counter = 31 then
    Counter := 0;
  if Assigned(pBookMark) then
  begin
    RecNo := pDSBOOKMRK(pBookmark)^.iRecNo;
    NewPos := -1;
    with FRecNums do
      for I := 0 to Count - 1 do
      begin
        pRecBuf := FDSBase.FRecords.List^[Integer(List^[I])];
        if GetRecordRecNo(pRecBuf) = RecNo then
        begin
          NewPos := I;
          Break;
        end;
      end;

    if NewPos >= 0 then
    begin
      RecPos := NewPos;
      Result := DBERR_NONE;
    end else
    begin
    // The code below is provided for compability with Borland only.
    // It will not return an error if a bookmark is moved to a record that
    // is not actually in a dataset.
      Result := DBERR_NONE;
    (*
      if pDSBOOKMRK(pBookmark)^.iState and (dsRecNew or dsRecModified or dsRecDeleted) <> 0 then
      begin
        Result := DBERR_NONE;
        Exit;
      end else
        Result := DBERR_RECNOTFOUND;
    *)
    end;
  end else
    Result := DBERR_RECNOTFOUND;
  Inc(Counter);
end;

function TDSCursor.MoveToBOF: DBResult;
begin
  RecPos := -1;
  Result := DBERR_NONE;
end;

function TDSCursor.MoveToEOF: DBResult;
begin
  UpdateCursorPos;
  RecPos := FRecNums.Count;
  Result := DBERR_NONE;
end;

procedure TDSCursor.CheckIndexed;
begin
  if FDefaultOrder or FChangeIndex then
    HBError(DBERR_NOTINDEXED);
end;

procedure TDSCursor.CheckIndexWritable;
begin
  if FChangeIndex then
    HBError(DBERR_INDEXREADONLY);
end;

function TDSCursor.InternalMoveToKey(SearchCond: DBSearchCond; iFields,
  iPartLen: DWord; pKeyBuf, pRecBuf: Pointer; var RecNo: Integer): DBResult;
var
  Cmp, CmpP: Integer;
  L, H, K: Integer;

  Records, RecNums: PPointerList;
  IndexDS: pDSIDXDesc;
  IndexUnique: Boolean;

  function CmpPrev: Integer;
  begin
    if K > 0 then
      Result := FDSBase.CmpRecsEx(IndexDS, iFields, iPartLen, Records^[Integer(RecNums^[K - 1])], pKeyBuf) else
      Result := -1;
  end;

var
  Rec: PChar;

begin
{ iFields contains fields for non-partial compare }
  if iPartLen > 0 then Inc(iFields);

  L := 0;

  Records := FDSBase.FRecords.List;

  with FRecNums do
  begin
    RecNums := FRecNums.List;
    H := FRecNums.Count - 1;
  end;

  if FDefaultOrder then
  begin
    IndexDS := nil;
    IndexUnique := True;
  end else begin
    IndexDS := @FIndexDS;
    IndexUnique := FIndexDS.bUnique;
  end;

  while L <= H do
  begin
    K := (L + H) shr 1;

    Rec := Records^[Integer(RecNums^[K])];
    Cmp := FDSBase.CmpRecsEx(IndexDS, iFields, iPartLen, Rec, pKeyBuf);

    if (Cmp < 0) or (Cmp = 0) and (SearchCond = keySEARCHGT) then
      L := K + 1
    else begin
      if IndexUnique and ((SearchCond = keySEARCHEQ) or (Cmp = 0))then
        CmpP := -1 else CmpP := CmpPrev;

      if (SearchCond = keySEARCHGT ) and (Cmp >  0) and (CmpP <= 0) or
         (SearchCond = keySEARCHGEQ) and (Cmp >= 0) and (CmpP <  0) or
         (SearchCond = keySEARCHEQ ) and (Cmp =  0) and (IndexUnique or (CmpP < 0)) then
      begin
        RecNo := K;
        if Assigned(pRecBuf) then
          Move(Rec^, pRecBuf^, FDSBase.FRecBufSize);
        Result := DBERR_NONE;
        Exit;
      end;
      H := K - 1;
    end;
  end;
  Result := DBERR_RECNOTFOUND;
end;

function TDSCursor.GetRecordForKey(iFields, iPartLen: DWord; pKey,
  pRecBuf: Pointer): DBResult;
begin
  try
    CheckIndexed;
    Result := InternalMoveToKey(keySEARCHEQ, iFields, iPartLen, pKey, pRecBuf, FRecPos);
  except
    on EHDBError do
      Result := DBResult(False);
  end;
end;

function TDSCursor.MoveToKey(SearchCond: DBSearchCond; iFields,
  iPartLen: DWord; pRecBuf: Pointer): DBResult;
begin
  try
    CheckIndexed;
    Result := InternalMoveToKey(SearchCond, iFields, iPartLen, pRecBuf, nil, FRecPos);
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.MoveToSeqNo(i: DWord): DBResult;
begin
  try
    if (i > 0) and (i <= DWord(FRecNums.Count)) then
    begin
      RecPos := i - 1;
      Result := DBERR_NONE;
    end else
      Result := DBERR_RECNOTFOUND;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.PutBlob(pRecBuf: Pointer; iFieldNo, iOffSet: DWord;
  pBuf: Pointer; iLength: DWord): DBResult;
begin
  Result := FDSBase.PutBlob(pRecBuf, iFieldNo, iOffSet, pBuf, iLength);
end;

function TDSCursor.PutField(pRecBuf: Pointer; iFieldNo: DWord;
  pFldBuf: Pointer): DBResult;
begin
  Result := FDSBase.PutField(pRecBuf, iFieldNo, pFldBuf)
end;

function TDSCursor.GetRowRequestPacket(bKeysOnly, bFetchAllBlobs,
  bFetchAllDetails, bInclMetaData: Bool; out Packet: PSafeArray): DBResult;
var
  IndexDS: pDSIDXDesc;
begin
  try
    CheckCursorPos;
    if FDefaultOrder or FChangeIndex then
      IndexDS := nil else IndexDS := @FIndexDS;
    Result := FDSBase.GetRowRequestPacket(CurrentSeqNo, IndexDS, bKeysOnly, bFetchAllBlobs,
      bFetchAllDetails, bInclMetaData, Packet);
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.ResyncWithMaster: DBResult;
begin
  try
    if Assigned(FMaster) then
    begin
      Rebuild;
      ResyncDetails;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

procedure TDSCursor.InsertCursor(ACursor: TDSCursor);
begin
  FDetails.Add(ACursor);
  ACursor.FMaster := Self;
  ACursor.ResyncWithMaster;
end;

procedure TDSCursor.RemoveCursor(ACursor: TDSCursor);
begin
  FDetails.Remove(ACursor);
  ACursor.FMaster := nil;
  ACursor.FFieldsM := nil;
  ACursor.FFieldsD := nil;
  if Assigned(ACursor.FDSBase) then ACursor.Rebuild;
end;

procedure TDSCursor.ResyncDetails;
var
  I: Integer;
begin
  with FDetails do
    for I := 0 to Count - 1 do
      TDSCursor(List^[I]).ResyncWithMaster;
end;

function TDSCursor.RevertRecord: DBResult;
begin
  try
    UpdateCursorPos;
    if FRecPos >= 0 then
      Result := FDSBase.RevertRecord(CurrentSeqNo) else
      Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.RefreshRecord(Packet: PSafeArray): DBResult;
begin
  try
    CheckCursorPos;
    Result := FDSBase.RefreshRecords(Packet, CurrentSeqNo + 1, 0, nil)
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.SetNotifyCallBack(iClientData: DWord;
  pfCallBack: pfCHANGECallBack): DBResult;
begin
  FCNClientData := iClientData;
  FCNCallBack   := pfCallBack;
  Result        := DBERR_NONE;
end;

function TDSCursor.SetProp(eProp: CURProp; iPropValue: DWord): DBResult;
begin
  case eProp of
    curpropCONSTRAINT_ERROR_MESSAGE:{ r pCHAR,   Constraint Error Message }
      ;
    curpropDSBASE:                  { r pDSBASE, Underlying DSBASE) }
      ;
    curpropSETCRACK:                { w Sets crack-value to supplied value (DBERR_NOCURRREC) }
      begin
        FCrack := True;
        ResyncDetails;
      end;
    curpropGETORG_RECBUF:           { r returns recordbuffer for original record, error if none }
      ;
  end;
  Result := DBERR_NONE;
end;

function TDSCursor.SortOnFields(iFields: DWord; piFields: PDWord;
  pDescending, pCaseInsensitive: PBool): DBResult;
begin
  try
    FDSBase.FreeRecord(FKeyRec1);
    FDSBase.FreeRecord(FKeyRec2);
    ZeroMem(@FIndexDS, SizeOf(DSIDXDesc));
    FIndexDS.iFields := iFields;
    Move(piFields^, FIndexDS.iKeyFields, SizeOf(DWord) * iFields);
    FIndexDS.iFields := iFields;
    if Assigned(pDescending) then
      Move(pDescending^, FIndexDS.bDescending, SizeOf(Bool) * iFields);
    if Assigned(pCaseInsensitive) then
      Move(pCaseInsensitive^, FIndexDS.bCaseInsensitive, SizeOf(Bool) * iFields);

    FDSBase.CreateIndex(FIndexDS);

    FChangeIndex := False;
    FDefaultOrder := False;
    Sort;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.UndoLastChange(bFollowChange: Bool): DBResult;
var
  NewPos, RecNo: Integer;
begin
  try
    FInUndo := True;
    try
      Check(FDSBase.UndoLastChange(RecNo, True));
      if bFollowChange then
      begin
        with FRecNums do
          NewPos := FindInteger(RecNo, List^, Count);
        if NewPos >= 0 then RecPos := NewPos;
      end;
      NotifyChanged;
    finally
      FInUndo := False;
    end;
    Result := DBERR_NONE;
  except
    Result := HandleExceptions;
  end;
end;

function TDSCursor.UseIndexOrder(pszName: PChar): DBResult;
var
  I: Integer;
  NewChangeIndex: Boolean;
  DefIndexName: PChar;
begin
  if not Assigned(pszName) then
  begin
    if FIsDelta then
      DefIndexName := szDEFAULT_ORDER else
      DefIndexName := szPRIMARY_KEY;

    if FDSBase.FindIndex(DefIndexName) >= 0 then
      pszName := DefIndexName else
      pszName := szDEFAULT_ORDER;
  end;

  if (StrIComp(pszName, FIndexDS.szName) <> 0) then
  begin
    I := FDSBase.FindIndex(pszName);
    if (I >= 0) then
    begin
      FDSBase.FreeRecord(FKeyRec1);
      FDSBase.FreeRecord(FKeyRec2);

      FDSBase.GetIndexDesc(I, @FIndexDS);
      NewChangeIndex := StrIComp(FIndexDS.szName, szCHANGEINDEX) = 0;
      FDefaultOrder := not NewChangeIndex and (StrIComp(FIndexDS.szName, szDEFAULT_ORDER) = 0);

      if FChangeIndex <> NewChangeIndex then
      begin
        FChangeIndex := NewChangeIndex;
        Rebuild;
      end else
        Sort;
      Result := DBERR_NONE;
    end else
      Result := DBERR_NOSUCHINDEX;
  end else
    Result := DBERR_NONE;
end;

function TDSCursor.VerifyField(iFieldNo: DWord; pFldBuf: Pointer): DBResult;
begin
  Result := DBERR_NONE;
end;

procedure TDSCursor.NotifyChanged;
begin
  try
    if Assigned(FCNCallBack) then FCNCallBack(FCNClientData)
  except end;
end;

procedure TDSCursor.Notify(const Event, Data1, Data2, Data3: Integer);

  procedure UpdateRecNums;
  var
    I, RecNo: Integer;
    RecNums: PPointerList;
  begin
    RecNums := FRecNums.List;

    for I := 0 to FRecNums.Count - 1 do
    begin
      RecNo := Integer(RecNums^[I]);
      if RecNo >= Data1 then
      begin
        Inc(RecNo, Data3);
        Integer(RecNums^[I]) := RecNo;
      end;
    end;
  end;

var
  Rec1, Rec2: PChar;
begin
  case Event of
    ceRebuild:
      begin
        Rebuild;
      end;

    ceRecIns:
      begin
        Rec1 := FDSBase.FRecords.List^[Data1];
        if AcceptsRecord(Rec1) then
          InsertRecNo(Data1)
        else
          RemoveRecNo(Data1);
      end;

    ceRecMod:
      begin
        with FDSBase.FRecords do
        begin
          Rec1 := List^[Data1];
          if Data2 >= 0 then
            Rec2 := List^[Data2] else Rec2 := nil;
        end;

        if AcceptsRecord(Rec1) then
          InsertRecNo(Data1)
        else
          RemoveRecNo(Data1);

        if Data2 >= 0 then
        begin
          if AcceptsRecord(Rec2) then
            InsertRecNo(Data2)
          else
            RemoveRecNo(Data2);
        end;

        UpdateRecNums;
      end;

    ceRecDel:
      begin
        Rec1 := FDSBase.FRecords.List^[Data1];

        if AcceptsRecord(Rec1) then
          InsertRecNo(Data1)
        else
          RemoveRecNo(Data1);

        UpdateRecNums;
     end;

  ceNotify:
    NotifyChanged;

  ceNotifyAppend:
    begin
//      MoveRelative(1);
//      MoveToEOF;
      NotifyChanged;
    end;
  end;

  if (FDSBase.FNotifyDisableCount = 0) and not FInUndo then NotifyChanged;
end;

function TDSCursor.CurrentSeqNo: DWord;
begin
  Result := DWord(FRecNums.List^[FRecPos])
end;

procedure TDSCursor.SetRecPos(Value: Integer);
begin
  if FRecPos <> Value then
  begin
    FRecPos := Value;
    FCrack := False;
    ResyncDetails;
  end;
end;

procedure TDSCursor.Clear;
begin
  FRecNums.Clear;
  InvalidateAggs;
end;

procedure TDSCursor.Rebuild;
var
  I: Integer;
  pRecBuf: PChar;
begin
  Clear;

  with FDSBase.FRecords do
  begin
    for I := 0 to Count - 1 do
    begin
      pRecBuf := List^[I];
      if Assigned(pRecBuf) and AcceptsRecord(pRecBuf) then
        FRecNums.Add(Pointer(I));
    end;
  end;
  Sort;
end;

function TDSCursor.CmpRecsIndexEx(pRec1, pRec2: PChar; iFields: DWord): Integer;
begin
  Result := FDSBase.CmpRecsEx(@FIndexDS, iFields, 0, pRec1, pRec2);
end;

function TDSCursor.CmpRecsIndex(pRec1, pRec2: PChar): Integer;
begin
  Result := FDSBase.CmpRecsEx(@FIndexDS, 0, 0, pRec1, pRec2);
end;

function TDSCursor.CmpRecsIndexDef(pRec1, pRec2: PChar): Integer;
begin
  Result := FDSBase.CmpRecsEx(nil, 0, 0, pRec1, pRec2);
end;

procedure TDSCursor.Sort;

 var
   Records, RecNums: PPointerList;

  procedure QuickSort(L, R: Integer; Compare: TCompareRecords);
  var
    I, J: Integer;
    P, Tmp: Pointer;
  begin
    repeat
      I := L;
      J := R;
      P := Records^[Integer(RecNums^[(L + R) shr 1])];
      repeat
        if (L <> R) then
        begin
          while Compare(Records^[Integer(RecNums^[I])], P) < 0 do Inc(I);
          while Compare(Records^[Integer(RecNums^[J])], P) > 0 do Dec(J);
        end;
        if I <= J then
        begin
          if I < J then
          begin
            Tmp := RecNums^[I];
            RecNums^[I] := RecNums^[J];
            RecNums^[J] := Tmp;
          end;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then QuickSort(L, J, Compare);
      L := I;
    until I >= R;
  end;

  procedure SortChangeIndex;
  var
    I: Integer;
  begin
    with FRecNums do
    begin
      Count := FDSBase.FRecords.Count;
      for I := 0 to Count - 1 do
        Integer(RecNums^[I]) := I;
    end;
  end;

begin
  InvalidateAggs;

  if (FRecNums.Count < 2) then Exit;

  Records := FDSBase.FRecords.List;

  with FRecNums do
  begin
    RecNums := List;

    if FChangeIndex then
      SortChangeIndex
    else if FDefaultOrder then
      QuickSort(0, Count - 1, CmpRecsIndexDef)
    else
      QuickSort(0, Count - 1, CmpRecsIndex);
  end;
end;

function TDSCursor.AcceptsRecord(pRecBuf: PChar): Boolean;

  function RecordVisible: Boolean;
  var
    RecAttr: DSAttr;
  begin
    RecAttr := GetRecordAttr(pRecBuf);
    if (RecAttr and (dsUnused or dsRecOrg)) <> 0 then
      Result := False
    else if FChangeIndex then
      Result := FIsDelta and ((RecAttr and dsRecModified2) <> 0) or not FIsDelta and ((RecAttr and dsIsModified) <> 0) else
      Result := FIsDelta or ((RecAttr and dsIsNotVisible) = 0);
  end;

  function RecordInRange: Boolean;
  var
    I: Integer;
  begin
    if Assigned(FKeyRec1) then
    begin
      I := CmpRecsIndexEx(pRecBuf, FKeyRec1, FKeyFields);
      if (I > 0) or (FKeyRec1Incl and (I = 0)) then
      begin
        I := CmpRecsIndexEx(pRecBuf, FKeyRec2, FKeyFields);
        Result := (I < 0) or (FKeyRec2Incl and (I = 0))
      end else
        Result := False;
    end else
      Result := True;
  end;

  function RecordFilteredExpr: Boolean;
  begin
    if Assigned(FExprEval.Data) then
      Result := TDSExprEvaluator(FExprEval).EvaluteRec(pRecBuf) else
      Result := True;
  end;

  function RecordFilteredFunc: Boolean;
  begin
    if Assigned(FFFilter) then
      Result := FFFilter(FFClientData, pRecBuf) else
      Result := True;
  end;

  function RecordInDetails: Boolean;
  var
    pRecBufM: PChar;
    pFldD: PChar;
    Err: DBResult;
  begin
    Result := False;
    if Assigned(FMaster) then
    begin
      if Assigned(FMasterRecBuf) then
      begin
        pRecBufM := FMasterRecBuf;
        Err := DBERR_NONE;
      end else
      begin
        Err := FMaster.GetCurrentRecordPtr(pRecBufM);
      end;

      if Err = DBERR_NONE then
      begin
        if FFieldsM[0] = 0 then
        begin
          pFldD := pRecBuf + FFields[FFieldsD[0] - 1].iFldOffsInRec;
          Result := GetRecordRecNo(pRecBufM) = PInteger(pFldD)^;
        end else begin
          // Multifield links are not supported
          Result := True;
        end;
      end;
    end else
      Result := True;
  end;

begin
  Result := Assigned(pRecBuf) and RecordVisible and RecordInDetails and
    RecordInRange and RecordFilteredExpr and RecordFilteredFunc;
end;

initialization
  DecimalSeparator := '.';
  
{$IFDEF RELEASE_HB }
  {$I check.inc }
{$ENDIF}

end.

